
/* $Id: t2_work.h,v 1.9 2000/01/11 19:34:27 chialin Exp $ */

/* t2_work.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/20/98
 * Work list
 */


#ifndef _T2_WORK_H_
#define _T2_WORK_H_

#include	"t2_header.h"
#include	"t2_tqueue.h"
#include	"t2_qenv.h"
#include	"t2_comm.h"
#include	"t2_clusterreq.h"


extern int _g_mynode;

// forward declaration
class T2_WorkList;

////////////////////////////////////////////////////////
//
// T2_WorkEntry
//
// A base class for an entry of work in the work list
// defined below.
//
////////////////////////////////////////////////////////
class T2_WorkEntry {
friend class T2_WorkList;
public:
  T2_WorkEntry() {};
  virtual ~T2_WorkEntry() {};

  virtual void handler() { cout << "T2_WorkEntry::handler is called" << endl; };

protected:
  static T2_RTS *rts;			// a pointer to rts
  static T2_WorkList *wlist;		// a pointer to the work list
  static T2_CommManager *comm;		// a pointer to the comm. manager

private:
};


////////////////////////////////////////////////////////
//
// T2_ClusterWork
// A derived class from T2_WorkEntry for cluster to
// be fired.
//
////////////////////////////////////////////////////////
class T2_ClusterWork: public T2_WorkEntry {
public:
  T2_ClusterWork(T2_ClusterRequest* c): cluster(c) { t2_Assert (c != NULL); };
  ~T2_ClusterWork() {};

  virtual void handler();

protected:
  T2_ClusterRequest* cluster;

private:
};



////////////////////////////////////////////////////////
//
// T2_GlobalCombineWork
//
// A derived class from T2_WorkEntry for cluster to
// be fired.
//
////////////////////////////////////////////////////////
class T2_GlobalCombineWork: public T2_WorkEntry {
public:
  T2_GlobalCombineWork(T2_QEnv* qp) { t2_Assert(qp != NULL); qenv = qp; };
  ~T2_GlobalCombineWork() { t2_Assert((qenv = NULL) == NULL); };

  virtual void handler();

//  static T2_CommRecvAcc& getCommRecvAcc()
//	{ t2_Assert (commRecvAcc != NULL); return *commRecvAcc; };
//  static T2_CommRecvAcc* setCommRecvAcc(T2_CommRecvAcc* p)
//	{ return (commRecvAcc = p); };
//  static T2_CommSendAcc& getCommSendAcc()
//	{ t2_Assert (commSendAcc != NULL); return *commSendAcc; };
//  static T2_CommSendAcc* setCommSendAcc(T2_CommSendAcc* p)
//	{ return (commSendAcc = p); };

protected:
  T2_QEnv *qenv;

//  static T2_CommRecvAcc* commRecvAcc;
//  static T2_CommSendAcc* commSendAcc;
private:
};



////////////////////////////////////////////////////////
//
// T2_AccMsgWork
//
// A derived class from T2_WorkEntry for handling an
// received accumulator message
//
////////////////////////////////////////////////////////
class T2_AccMsgWork: public T2_WorkEntry {
public:
  T2_AccMsgWork(T2_ProcID p, T2_AccMsgRecvBuffer* a, T2_QEnv* q);
  ~T2_AccMsgWork() {};

  virtual void handler();

protected:
  T2_ProcID src;
  T2_AccMsgRecvBuffer* msg;
  T2_QEnv* qenv;

private:
};


inline
T2_AccMsgWork::T2_AccMsgWork(T2_ProcID p, T2_AccMsgRecvBuffer* a, T2_QEnv* q):
				src(p), msg(a), qenv(q)
{
}


////////////////////////////////////////////////////////
//
// T2_OutputHandleWork
//
// A derived class from T2_WorkEntry for handling the
// final output.
//
////////////////////////////////////////////////////////
class T2_OutputHandleWork: public T2_WorkEntry {
public:
  T2_OutputHandleWork(T2_QEnv* qp) { t2_Assert(qp != NULL); qenv = qp; };
  ~T2_OutputHandleWork() { t2_Assert((qenv = NULL) == NULL); };

  virtual void handler();

protected:
  T2_QEnv* qenv;

private:
};



////////////////////////////////////////////////////////
//
// T2_EndIterationWork
//
// A derived class from T2_WorkEntry for marking end
// of an iteration.
//
////////////////////////////////////////////////////////
class T2_EndIterationWork: public T2_WorkEntry {
public:
  T2_EndIterationWork(T2_QEnv* qp) { t2_Assert (qp != NULL); qenv = qp; };
  ~T2_EndIterationWork() { t2_Assert ((qenv = NULL) == NULL); };

  virtual void handler();

protected:
  T2_QEnv* qenv;

private:
};



////////////////////////////////////////////////////////
//
// T2_WorkList
//
// A work list of pointers to T2_WorkEntry objects.
//
////////////////////////////////////////////////////////
class T2_WorkList {
public:
  T2_WorkList(u_int l): wlist(l) {};
  ~T2_WorkList();

  // query methods
  bool isEmpty() const { return wlist.isEmpty(); };
  bool isFull() const { return wlist.isFull(); };
  u_int getNumberElements() const { return wlist.getNumberElements(); };
  u_int getMaxNumberElements() const { return wlist.getMaxQueueSize(); };

  // initialize
  void init(T2_RTS* r, T2_CommManager* c);

  // enqueue a cluster that is ready to be processed
  bool enqueueCluster(T2_ClusterRequest* c);

  // enqueue a query env which is ready to start its global combine phase
  bool enqueueGlobalCombine(T2_QEnv* qp);

  // enqueue an acc msg and the target query env for global combine
  bool enqueueAccMsg(T2_ProcID src, T2_AccMsgRecvBuffer* a, T2_QEnv* qp);

  // enqueue a qenv that is ready to advance to output handling phase
  bool enqueueOutputHandle(T2_QEnv* qp);

  // enqueue a qenv that has finished an iteration
  bool enqueueEndIteration(T2_QEnv* qp);

  // dequeue a work entry
  T2_WorkEntry* dequeue();

protected:
  T2_TQueue<T2_WorkEntry*> wlist;

private:
};


inline
T2_WorkList::~T2_WorkList()
{
  t2_Assert(isEmpty());

  T2_WorkEntry::rts = NULL;
  T2_WorkEntry::wlist = NULL;
  T2_WorkEntry::comm = NULL;
}

inline bool
T2_WorkList::enqueueCluster(T2_ClusterRequest* c)
{
  t2_Assert (c != NULL);

assert (!isFull());

  if (isFull() == true)
     return false;

# ifdef T2_ASSERTION
     bool r = 
# endif
  wlist.enqueue(new T2_ClusterWork(c));
  t2_Assert (r == true);
  return true;
}

inline bool
T2_WorkList::enqueueGlobalCombine(T2_QEnv* qp)
{
  t2_Assert (qp != NULL);

assert (!isFull());

  if (isFull() == true)
     return false;

# ifdef T2_ASSERTION
     bool r = 
# endif
  wlist.enqueue(new T2_GlobalCombineWork(qp));
  t2_Assert (r == true);
  return true;
}

inline  bool
T2_WorkList::enqueueAccMsg(T2_ProcID src, T2_AccMsgRecvBuffer* a, T2_QEnv* qenv)
{
  t2_Assert (qenv != NULL);

assert (!isFull());

  if (isFull() == true)
     return false;

# ifdef T2_ASSERTION
     bool r = 
# endif
  wlist.enqueue(new T2_AccMsgWork(src, a, qenv));
  t2_Assert (r == true);
  return true;
}

inline bool
T2_WorkList::enqueueOutputHandle(T2_QEnv* qenv)
{
  t2_Assert (qenv != NULL);

assert (!isFull());

  if (isFull() == true)
     return false;

# ifdef T2_ASSERTION
     bool r = 
# endif
  wlist.enqueue(new T2_OutputHandleWork(qenv));
  t2_Assert (r == true);
  return true;
}

inline bool
T2_WorkList::enqueueEndIteration(T2_QEnv* qenv)
{
  t2_Assert (qenv != NULL);

assert (!isFull());

  if (isFull() == true)
     return false;

# ifdef T2_ASSERTION
     bool r = 
# endif
  wlist.enqueue(new T2_EndIterationWork(qenv));
  t2_Assert (r == true);
  return true;
}

inline  T2_WorkEntry*
T2_WorkList::dequeue()
{
  return (wlist.isEmpty())?
	 (T2_WorkEntry *) NULL : (T2_WorkEntry *) wlist.dequeue();
}



////////////////////////////////////////////////////////
//
// allOutputWritesDone
//
// The function allOutputWritesDone should be called by
// whichever function that detects all the output writes
// of a qenv for an iteration have been completed.
// allOutputWritesDone would check to see if the output
// handling phase is complete; if it is, allOutputWritesDone
// adds qenv to the work list, otherwise, it does nothing
// and the qenv would eventually been put to the work list
// when all its appending sends complete.
//
////////////////////////////////////////////////////////
inline void
allOutputWritesDone(T2_QEnv* qenv, T2_WorkList& wlist)
{
  t2_Assert (qenv != NULL);

  if (qenv->endOutputHandle() == true) {	// end of output hanlding

     t2_say3(cout,
	    "allOutputWritesDone ends output-handling phase"
	    << ", enqueue EndIterationWork" << endl);

     t2_Assert (wlist.isFull() == false);

#    ifdef T2_ASSERTION
        bool r = 
#    endif
     wlist.enqueueEndIteration(qenv);
     t2_Assert (r == true);
  }
}



#endif	/* ! _T2_WORK_H_ */
