
/* $Id: comm_mpi.C,v 1.12 2000/12/02 02:10:50 chialin Exp $ */

/* comm_mpi.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/5/98
 * Communication functions implemented over MPI.
 */

#include	<limits.h>


//////////////////////////////////////////////////////////////////////////
//
// global variable
//
//////////////////////////////////////////////////////////////////////////
static MPI_Comm t2_comm_backend;

//////////////////////////////////////////////////////////////////////////
//
// initialization and finalization
//
//////////////////////////////////////////////////////////////////////////
void
t2_comm_setup(int& argc, char**& argv, char**& envp, int& numnodes, int& mynode)
{
  MPI_Init(&argc, &argv);

  // create a communicator for all the back-end nodes
//  MPI_Group t2_group_backend;
//  MPI_Comm_group(MPI_COMM_WORLD, &t2_group_backend);
//  MPI_Comm_create(MPI_COMM_WORLD, t2_group_backend, &t2_comm_backend);
  MPI_Comm_dup(MPI_COMM_WORLD, &t2_comm_backend);

  int r;
  r = MPI_Comm_size(t2_comm_backend, &numnodes);
  t2_Assert (r == MPI_SUCCESS);
  r = MPI_Comm_rank(t2_comm_backend, &mynode);
  t2_Assert (r == MPI_SUCCESS);
}


void
t2_comm_finalize()
{
  MPI_Comm_free(&t2_comm_backend);
  MPI_Finalize();
}



//////////////////////////////////////////////////////////////////////////
//
// communication routines among the backend nodes
//
//////////////////////////////////////////////////////////////////////////
// global barrier
void t2_be_gsync()
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Barrier(t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}


// a set of reduction functions - sum
u_int t2_be_gsum(u_int i)
{
  u_int s;

# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce(&i, &s, 1, MPI_UNSIGNED, MPI_SUM, t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
  return s;
}

int t2_be_gsum(int i)
{
  int s;

# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce(&i, &s, 1, MPI_INT, MPI_SUM, t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
  return s;
}

long t2_be_gsum(long i)
{
  long s;

# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce(&i, &s, 1, MPI_LONG, MPI_SUM, t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
  return s;
}

u_long t2_be_gsum(u_long i)
{
  u_long s;
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce(&i, &s, 1, MPI_UNSIGNED_LONG, MPI_SUM, t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
  return s;
}

void t2_be_gsum(u_int* in, u_int* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_UNSIGNED, MPI_SUM,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gsum(int* in, int* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_INT, MPI_SUM,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gsum(long* in, long* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_LONG, MPI_SUM,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gsum(u_long* in, u_long* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_UNSIGNED_LONG, MPI_SUM,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gmax(u_int* in, u_int* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_UNSIGNED, MPI_MAX,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gmax(int* in, int* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_INT, MPI_MAX,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gmax(long* in, long* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_LONG, MPI_MAX,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gmax(u_long* in, u_long* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_UNSIGNED_LONG, MPI_MAX,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gmin(u_int* in, u_int* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_UNSIGNED, MPI_MIN,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gmin(int* in, int* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_INT, MPI_MIN,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gmin(long* in, long* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_LONG, MPI_MIN,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}

void t2_be_gmin(u_long* in, u_long* out, const u_int cnt)
{
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Allreduce((void*) in, (void*) out, cnt, MPI_UNSIGNED_LONG, MPI_MIN,
		t2_comm_backend);
  t2_Assert (r == MPI_SUCCESS);
}


// blocking sends; return true when succeeded, false otherwise
bool t2_be_send(const char* buf, size_t nbytes, T2_ProcID dest, int tag)
{
  int r = MPI_Send((void *) buf, (int) nbytes, MPI_CHAR, (int) dest, tag,
		   t2_comm_backend);
  if (r == MPI_SUCCESS)			// send succeeded !
     return true;
  return false;
}

bool t2_be_send(const int* buf, size_t count, T2_ProcID dest, int tag)
{
  int r = MPI_Send((void *) buf, (int) count, MPI_INT, (int) dest, tag,
		   t2_comm_backend);
  if (r == MPI_SUCCESS)			// send succeeded !
     return true;
  return false;
}

bool t2_be_send(const u_int* buf, size_t count, T2_ProcID dest, int tag)
{
  int r = MPI_Send((void *) buf, (int) count, MPI_UNSIGNED, (int) dest, tag,
		   t2_comm_backend);
  if (r == MPI_SUCCESS)			// send succeeded !
     return true;
  return false;
}

bool t2_be_send(const long* buf, size_t count, T2_ProcID dest, int tag)
{
  int r = MPI_Send((void *) buf, (int) count, MPI_LONG, (int) dest, tag,
		   t2_comm_backend);
  if (r == MPI_SUCCESS)			// send succeeded !
     return true;
  return false;
}

bool t2_be_send(const u_long* buf, size_t count, T2_ProcID dest, int tag)
{
  int r = MPI_Send((void *) buf, (int) count, MPI_UNSIGNED_LONG, (int)dest, tag,
		   t2_comm_backend);
  if (r == MPI_SUCCESS)			// send succeeded !
     return true;
  return false;
}


// a function to access MPI return status
// ** (cannot be used for obtaining the msg size for a non-blocking send,
// **  eg MPI-ch would not return the correct msg size)
inline void
t2_MPI_get_return_status(MPI_Status& status, MPI_Datatype datatype,
			 int& source, int& tag, size_t& count)
{
  int msg_count, r;
  r = MPI_Get_count(&status, datatype, &msg_count);
  t2_Assert (r == MPI_SUCCESS);

  source = (T2_ProcID) status.MPI_SOURCE;
  tag = (int) status.MPI_TAG;
  count = (size_t) msg_count;
}


// blocking recvs; return true when succeeded, false otherwise
bool t2_be_recv(char* buf, size_t buflen, int& source, int& tag,
		size_t& nbytes)
{
  MPI_Status status;
  int r = MPI_Recv((void *) buf, (int) buflen, MPI_CHAR, source, tag,
		   t2_comm_backend, &status);
  if (r != MPI_SUCCESS)			// recv failed !
     return false;

  // check return status
  t2_MPI_get_return_status(status, MPI_CHAR, source, tag, nbytes);
  return true;
}

bool t2_be_recv(int* buf, size_t bufcount, int& source, int& tag,
		size_t& count)
{
  MPI_Status status;
  int r = MPI_Recv((void *) buf, (int) bufcount, MPI_INT, source, tag,
		   t2_comm_backend, &status);
  if (r != MPI_SUCCESS)			// recv failed !
     return false;

  // check return status
  t2_MPI_get_return_status(status, MPI_INT, source, tag, count);
  return true;
}

bool t2_be_recv(u_int* buf, size_t bufcount, int& source, int& tag,
		size_t& count)
{
  MPI_Status status;
  int r = MPI_Recv((void *) buf, (int) bufcount, MPI_UNSIGNED, source, tag,
		   t2_comm_backend, &status);
  if (r != MPI_SUCCESS)			// recv failed !
     return false;

  // check return status
  t2_MPI_get_return_status(status, MPI_UNSIGNED, source, tag, count);
  return true;
}

bool t2_be_recv(long* buf, size_t bufcount, int& source, int& tag,
		size_t& count)
{
  MPI_Status status;
  int r = MPI_Recv((void *) buf, (int) bufcount, MPI_LONG, source, tag,
		   t2_comm_backend, &status);
  if (r != MPI_SUCCESS)			// recv failed !
     return false;

  // check return status
  t2_MPI_get_return_status(status, MPI_LONG, source, tag, count);
  return true;
}

bool t2_be_recv(u_long* buf, size_t bufcount, int& source, int& tag,
		size_t& count)
{
  MPI_Status status;
  int r = MPI_Recv((void *) buf, (int) bufcount, MPI_UNSIGNED_LONG, source, tag,
		   t2_comm_backend, &status);
  if (r != MPI_SUCCESS)			// recv failed !
     return false;

  // check return status
  t2_MPI_get_return_status(status, MPI_UNSIGNED_LONG, source, tag, count);
  return true;
}


//////////////////////////////////////////////////////////////////////////
//
// a blocking method for probing message size
// returns faluse if error occurs, otherwise return true
//
//////////////////////////////////////////////////////////////////////////
bool t2_be_probe(int& src, int& tag, size_t& msgsize)
{
  MPI_Status status;
  int r = MPI_Probe(src, tag, t2_comm_backend, &status);
  if (r != MPI_SUCCESS)			// probe failed !
     return false;

  t2_MPI_get_return_status(status, MPI_CHAR, src, tag, msgsize);
  return true;
}


//////////////////////////////////////////////////////////////////////////
//
// methods for T2_CommControlBlock
//
//////////////////////////////////////////////////////////////////////////
T2_AsyncResult
t2_issueISend(T2_CommControlBlock& cb,
	      char* buf, size_t nbytes, int dest, int tag)
{
  // since MPI_Get_count does not return the correct message size for
  // non-blocking sends, we need to remember the msg size ourselves
  cb.nbytes = nbytes;
  int r = MPI_Isend(buf, nbytes, MPI_PACKED, dest, tag,
		    t2_comm_backend, &(cb.request));
  if (r == MPI_SUCCESS)
     return aSUCCEED;
  return aFAIL_SYS;
}


T2_AsyncResult
t2_issueIRecv(T2_CommControlBlock& cb,
	      char* buf, size_t nbytes, int src, int tag)
{
  // mark cb.nbytes with an invalid value
  cb.nbytes = UINT_MAX;
  int r = MPI_Irecv(buf, nbytes, MPI_PACKED, src, tag,
		    t2_comm_backend, &(cb.request));
  if (r == MPI_SUCCESS)
     return aSUCCEED;
  return aFAIL_SYS;
}


bool
t2_issueIProbe(T2_CommControlBlock& cb, int src, int tag)
{
  int flag;
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Iprobe(src, tag, t2_comm_backend, &flag, &(cb.status));
  t2_Assert (r == MPI_SUCCESS);
  if (flag == 0)
     return false;
  return true;
}


bool
t2_commTestCompletion(T2_CommControlBlock& cb)
{
  int flag;
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Test(&(cb.request), &flag, &(cb.status));
  t2_Assert (r == MPI_SUCCESS);
  if (flag)
     return true;
  return false;
}


int
t2_getMsgSize(T2_CommControlBlock& req)
{
  if (req.nbytes < UINT_MAX) 		// this req is a non-blocking send
     return req.nbytes;

  // this req is a non-blocking recv or probe
  int nbytes;
# ifdef T2_ASSERTION
     int r = 
# endif
  MPI_Get_count(&req.status, MPI_PACKED, &nbytes);
  t2_Assert (r == MPI_SUCCESS);
  return nbytes;
}


int
t2_getSource(T2_CommControlBlock& req)
{
  return req.status.MPI_SOURCE;
}


int
t2_getMsgTag(T2_CommControlBlock& req)
{
  return req.status.MPI_TAG;
}


int
t2_setSource(T2_CommControlBlock& req, int s)
{
  return (req.status.MPI_SOURCE = s);
}


int
t2_setMsgTag(T2_CommControlBlock& req, int t)
{
  return (req.status.MPI_TAG = t);
}


//////////////////////////////////////////////////////////////////////////
//
// methods for T2_Comm
//
//////////////////////////////////////////////////////////////////////////
