
/* lio.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/2/98
 * Method implementation for asynchronous reads.
 */

#include	<strstream.h>
#include	<unistd.h>
#include	<string.h>
#include	"t2_lio.h"


/////////////////////////////////////////////////////////////////////////
//
// include interface-dependent implementation
//
/////////////////////////////////////////////////////////////////////////
#ifdef POSIX_AIO
	#include "lio_posix.C"
#endif

#ifdef AIX_AIO
	#include "lio_aix.C"
#endif

#ifdef SUNOS_AIO
	#include "lio_sunos.C"
#endif


/////////////////////////////////////////////////////////////////////////
//
// methods for T2_LIO
//
/////////////////////////////////////////////////////////////////////////
T2_LIO::T2_LIO(const u_int n) : nlcb(n),freel(n),pendingl(n),readyl(n),donel(n)
{
  t2_Alloc(lcb, n, T2_LIOCB);
  t2_Alloc(key, n, T2_LIOKey);
  for (u_int i=0; i<n; i++)
     freel.in(lcb+i);

  memset(&lio_sigevent, 0, sizeof(lio_sigevent));	// init sigevent

  initLIOCB(lcb, n);		// initialize T2_LIOCB control blocks
  initAux();			// initialize interface-dependent aux data

#if (T2_TIME_EXECUTE > 1)
  exetimer = NULL;
#endif

}


T2_LIO::~T2_LIO()
{
  delete [] lcb;
  delete [] key;

  deleteAux();
}


bool
T2_LIO::getDoneKey(T2_LIOKey& k, int &sz)
{
  if (donel.isEmpty())
     return false;

  T2_LIOCB *p = donel.out();
  t2_Assert(p != NULL && p >= lcb);

  t2_Assert(!freel.isFull());
  freel += p;
  int i = p - lcb;
  k = key[i];
  sz = getReturnValue(*p);
  return true;
}


bool
T2_LIO::registerReq(const T2_LIOKey& k,
		    int op, int fd, int whence, off_t off,
		    char* buf, size_t size)
{
  t2_Assert (buf != NULL);
//  t2_Assert (size != 0);

  if (noMoreFreeCB())
     return false;

  T2_LIOCB *cb = freel.out();
  t2_Assert (cb != NULL);

  setValues(*cb, op, fd, whence, off, buf, size);
  t2_Assert (!readyl.isFull());
  readyl += cb;
  t2_Assert(cb >= lcb);
  int i = cb - lcb;
  key[i] = k;
  return true;
}

