
/* $Id: lio_sunos.C,v 1.12 2000/02/14 19:33:39 chialin Exp $ */

/* lio_sunos.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/2/98
 * Method implementation for asynchronous I/O under SunOS interface.
 */

#include	<errno.h>

/////////////////////////////////////
//
// T2_LIOCB methods
//
/////////////////////////////////////
void
initLIOCB(T2_LIOCB* lcb, u_int n)
{
//  t2_Assert (lcb != NULL);
//  for (u_int i=0; i<n; i++) {
//     T2_LIOCB& cb = lcb[i];
//  }
}


void
setMiscValues(T2_LIOCB& cb, int whence)
{
  cb.aio_whence = whence;
  cb.resultp.aio_return = AIO_INPROGRESS;
}


int
getReturnValue(T2_LIOCB& cb)
{
  return cb.resultp.aio_return;
}


// method to check if a control block completes
//bool
//opCompletes(T2_LIOCB& cb)
//{
//  return (cb.resultp.aio_return == AIO_INPROGRESS)? false : true;
//}


///////////////////////////////////////////////////////////////////////////
//
// T2_LIO methods
//
// aux.back_diff is a constant that allows one to find the owning
// T2_LIOCB from the address of a given resultp field.
//
// Since aiowait() is the only way to retire a aio_result_t pointer (ie
// changing its status from pending to complete), the implementation of
// issueReadyRequests and poll agrees on the protocol that when a ready
// request is issued, its control block (T2_LIOCB) is removed from readyl,
// but NOT put into pendingl. Rather it is left in no list until aiowait
// confirms its completion, at which time its control block is put into
// donel. However, to keep a count of the number of pending requests,
// a NULL pointer will be added to pendingl when a ready request is issued,
// and removed from pendingl when some pending request completes.
//
///////////////////////////////////////////////////////////////////////////
void
T2_LIO::initAux()
{
  T2_LIOCB a;
//  void *p1 = (void *) &a,
//       *p2 = (void *) &(a.resultp);
// well, ANSI C++ forbids using pointer of type `void *' in arithmetic ....
  int p1 = (int) &a,
      p2 = (int) &(a.resultp);
  aux.back_diff = p1 - p2;
}


void
T2_LIO::deleteAux()
{
}


T2_AsyncResult
T2_LIO::issueReadyRequests()
{
  const u_int nreqs = readyl.getNumberElements();

  t2_say4(cout, "T2_LIO::issueReadyRequests nreqs = " << nreqs << endl);

# if (T2_TIME_EXECUTE > 1)
{
     double tsnap = t2_dclock();
# endif

  for (u_int i=0; i<nreqs; i++) {
     T2_LIOCB* p = readyl[i];
     t2_Assert (p != NULL);
     t2_Assert (p->resultp.aio_return == AIO_INPROGRESS);

     int r;
     switch (p->lio_opcode) {
	case LIO_READ:	// read request
			r = aioread(p->lio_fildes,
				    p->aio_buf,
				    p->aio_nbytes,
				    p->aio_offset,
				    p->aio_whence,
				    &(p->resultp));

			t2_say4(cout,
//				"..... read p = " << (int) p << ", fd = "
				"..... read fd = "
				<< p->lio_fildes
		  		<< ", buf = " << (int) p->aio_buf
		  		<< ", nbytes = " << p->aio_nbytes
		  		<< ", offset = " << p->aio_offset
		  		<< ", whence = " << p->aio_whence
		  		<< " (nreqs = " << nreqs << ", i = " << i << ")"
		  		<< ", aio_return = "
				<< (int) p->resultp.aio_return
				<< endl);

			if (r != 0) {
			   t2_err_sys(__FILE__, __LINE__,
				     "aioread returns error, errno = %d", r);
			   return aFAIL_SYS;
			}
			break;

	case LIO_WRITE:	// write request
			r = aiowrite(p->lio_fildes,
				    p->aio_buf,
				    p->aio_nbytes,
				    p->aio_offset,
				    p->aio_whence,
				    &(p->resultp));
			t2_say4(cout,
				"..... write fd = " << p->lio_fildes
				<< ", buf = " << (int) p->aio_buf
				<< ", nbytes = " << p->aio_nbytes
				<< ", offset = " << p->aio_offset
				<< ", whence = " << p->aio_whence
				<< " (nreqs = " << nreqs << ", i = " << i << ")"
				<< ", aio_return = "
				<< (int) p->resultp.aio_return
				<< endl);

			if (r != 0) {
			   t2_err_sys(__FILE__, __LINE__,
				     "aiowrite returns error, errno = %d", r);
			   return aFAIL_SYS;
			}
			break;

	default:	// unrecognized opcode
			t2_err_quit(__FILE__, __LINE__,
				"unrecognized lio_opcode = %d", p->lio_opcode);
    }			// switch (p->lio_opcode)
  }

# if (T2_TIME_EXECUTE > 1)
     t2_Assert (exetimer != NULL);
     exetimer->incLiolistioTime(t2_dclock() - tsnap);
}
# endif


  // all requests are issued successfully
  // deposit a NULL pointer into pendingl for each request just issued
  for (u_int j=0; j<nreqs; j++) {
     t2_Assert (!pendingl.isFull());
     pendingl << (T2_LIOCB *) NULL;
  }
  readyl.emptySet();

  return aSUCCEED;
}


u_int
T2_LIO::poll()
{
  timeval t;
  t.tv_sec = 0;
  t.tv_usec = 0;

  int count = 0;
  aio_result_t *rp = NULL;

  while ((rp = aiowait(&t)) != NULL) {
     int eee = (int) rp;
     if (eee == -1) {			// some error occurred
	if (errno == EINVAL) {		// there is no outstanding aio requests
	   t2_Assert (pendingl.getNumberElements() == 0);
	   break;
	}

	t2_err_sys(__FILE__, __LINE__, "aiowait returns -1");
     }

     // find the owning T2_LIOCB from the given aio_result_t*
     // (I know this is ugly ....)
     T2_LIOCB *cb = (T2_LIOCB *) (((int) rp) + aux.back_diff);
     t2_Assert (((int)cb - (int) lcb) % sizeof(T2_LIOCB) == 0);

     // make sure that cb is not in any list
     const u_int nready = readyl.getNumberElements();
     for (u_int i=0; i<nready; i++) {
	t2_Assert (cb != readyl[i]);
     }
//     const u_int ndone = donel.getNumberElements();
     for (u_int j=0; j<nready; j++) {
	t2_Assert (cb != donel[j]);
     }

     // add cb to donel
     t2_Assert (!donel.isFull());
     donel.in(cb);
     count++;

     // remove a dummy NULL entry from pendingl
     pendingl.out();
  }

  return count;
}

