
/* $Id: ohmc.C,v 1.7 2000/02/04 00:20:51 chialin Exp $ */

/* ohmc.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 8/14/98
 * methods for T2_OHMetaChaos
 */



////////////////////////////////////////////////////////////////////////////
//
// static variable
//
////////////////////////////////////////////////////////////////////////////
MC_Server T2_OHMetaChaos::mc_server;


////////////////////////////////////////////////////////////////////////////
//
// T2_OHMetaChaos methods
//
////////////////////////////////////////////////////////////////////////////
T2_OHMetaChaos::T2_OHMetaChaos(T2_QEnv* p, T2_RTS& rts, T2_QSpecOutput& qspec):
		   T2_OutputHandle(p, true, qspec.useT2Protocol()),
		   mc_client((char *) qspec.getClient(),
			     (int) qspec.getNumberClients()),
		   ofbtag(-1), datatag(-1),
		   mcu(NULL), sched(NULL),
		   obuf(), ptrs(), cur_ptr(0)
{
  t2_Assert (qspec.getOutputHandleType() == t2_oMC);
  t2_Assert (p != NULL);

  if (!mc_server ||
     qspec.getClient() == NULL || qspec.getNumberClients() == 0)
  {
     invalidate();
     return;
  }

  ofbtag = qspec.getOffBandTag();
  datatag = qspec.getDataTag();
}


T2_OHMetaChaos::~T2_OHMetaChaos()
{
  t2_Assert (mcu == NULL);
  t2_Assert (sched == NULL);
}


bool
T2_OHMetaChaos::initBeforeQueryExecution(T2_RTS& rts)
{
  t2_Assert (isValid() == true);
  t2_Assert (qenv != NULL);

  // for now, just use a blocking wait ...
  mc_client.waitClient();

  T2_QQPlan *qqplan = qenv->getQQPlan();
  t2_Assert (qqplan != NULL);
  T2_Iteration niter = qqplan->getNumberSubplans();
  t2_Assert (niter > 0);

  {
     MC_OffBandComm offband;
     T2_MCHelloPacket hello;
     if (hello.bcast(offband, mc_client,
		     rts.getSystem().getMCProgramName(),
		     rts.getSystem().getNumberNodes(), niter,
		     ofbtag) == false)
     {
	// bcast fails
	return false;
     }
  }

  return true;
}


bool
T2_OHMetaChaos::handleOutput(const T2_QEnv& qenv, T2_Output& output,
			     T2_Accumulator& accumulator, T2_Iteration i)
{
  int kk = 0;
  t2_Assert (mcu == NULL);
  T2_MetaChaosUserConstructorPtr mcu_cons =
			T2_RTS::getMetaChaosUserConstructorPtr(kk);
  t2_Assert (mcu_const != NULL);
# ifdef T2_ASSERTION
     T2_UDFRet rv =
# endif
  mcu_cons(qenv.getQueryInfo().getSystem(),
	   output, accumulator, mc_server, mc_client, mcu);
  t2_Assert (rv == T2_UDFRet_OK);

  t2_Assert (sched == NULL);
  sched = new MC_Schedule(*mcu, mc_server, mc_client,
			  mcu->getOptions(), mcu->otherProgData());

  size_t buf_size = mcu->requiredMCBufferSpace();
  obuf.allocBuffer(buf_size);
  {
    T2_OutputBuffer outbuf(obuf);
    mcu->flushMCBuffer(outbuf, ptrs);
  }

  if (ptrs.getNumberPointers() == 0) {
     cleanupAfterAllWritesDoneForIteration();
     qenv->initOutputHandle(0);
     return false;				// no need of output handling
  }

  // issue 1st write
  if (issueOneWrite() == false) {
     cleanupAfterAllWritesDoneForIteration();
     qenv->initOutputHandle(0);
     return false;				// no need of output handling
  }

  // update query status
  qenv->initOutputHandle(1);

  return true;
}


bool
T2_OHMetaChaos::issueNextOutputWrite()
{
  if (issueOneWrite() == false) {		// all writes are issued
     cleanupAfterAllWritesDoneForIteration();

     // update query status
     t2_Assert (qenv != NULL);
     qenv->decNumberOutputWritesYetComplete();
     t2_Assert(qenv->getQueryStatus().getNumberOutputWritesYetComplete() == 0);
     return false;				// no need of output handling
  }

  return true;					// more writes may be needed
}


void
T2_OHMetaChaos::finalize()
{
}


bool
T2_OHMetaChaos::issueOneWrite()
{
  t2_Assert (cur_ptr <= ptrs.getNumberPointers());
  if (cur_ptr == ptrs.getNumberPointers())
     return false;


/////// send all data for now
  while (cur_ptr<ptrs.getNumberPointers()) {
      switch (ptrs.getCommType(cur_ptr)) {
	case T2_OUT_MC_COMM:
		sched->sendData((void*) ptrs.getDataPointer(cur_ptr),
				ptrs.getDataType(cur_ptr), datatag);
		break;

	case T2_OUT_MC_OFFBAND:
		{
		  MC_OffBandComm offband;
		  offband.bcast(mc_client,
				(void*) ptrs.getDataPointer(cur_ptr),
				ptrs.getDataSize(cur_ptr),
				ptrs.getDataType(cur_ptr),
				ofbtag);
		}
		break;

	case T2_OUT_SOCKET:
	case T2_OUT_FILE:
	case T2_OUT_UNKNOWN:
		t2_err_quit(__FILE__, __LINE__,
			"unexpected comm type %d",
			(int) ptrs.getCommType(cur_ptr));
		break;

	default:
		t2_err_quit(__FILE__, __LINE__,
			"unrecognized comm type %d",
			(int) ptrs.getCommType(cur_ptr));
      }

      cur_ptr++;
  }
  return false;
///////



  t2_Assert (sched != NULL);
  sched->sendData((void*) ptrs.getDataPointer(cur_ptr),
		  ptrs.getDataType(cur_ptr), datatag);
  cur_ptr++;
  return true;
}


void
T2_OHMetaChaos::cleanupAfterAllWritesDoneForIteration()
{
  obuf.releaseBuffer();
  ptrs.removeAllPointers();
  cur_ptr = 0;

  delete mcu;		t2_Assert ((mcu = NULL) == NULL);
  delete sched;		t2_Assert ((sched = NULL) == NULL);
}


bool
T2_OHMetaChaos::initMCServer(const char* mc_pname, int numnodes)
{
  t2_Assert (mc_pname != NULL);
  t2_Assert (numnodes != 0);

  return mc_server.initServer((char *) mc_pname, numnodes);
}


bool
T2_OHMetaChaos::initMCServer(const char* mc_pname, int numnodes, int me)
{
  t2_Assert (mc_pname != NULL);
  t2_Assert (numnodes != 0);

  return mc_server.initServer((char *) mc_pname, numnodes, me);
}


void
T2_OHMetaChaos::exitMCServer()
{
  mc_server.exitServer();
}
