
/* t2_bstream.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 4/3/98
 * A base class defined for a binary stream object.
 */

#ifndef _T2_BSTREAM_H_
#define _T2_BSTREAM_H_


#include	<unistd.h>
#include	"t2_header.h"


/////////////////////////////////////////////////////////////
//
// T2_BStream
//
// A base class for both an input and an output binary
// stream.
//
/////////////////////////////////////////////////////////////
class T2_BStream {
private:
protected:
  virtual void setErr(int e) = 0;

public:
  T2_BStream() {};
  virtual ~T2_BStream() {};

  virtual operator const void*() const = 0;
  virtual int berr() const = 0;
};



/////////////////////////////////////////////////////////////
//
// T2_IBStream
//
// A base class for an input binary stream.
//
/////////////////////////////////////////////////////////////
class T2_IBStream: public virtual T2_BStream {
private:
protected:
  bool convertEndianFlag;	// need to swap bytes for correct endianess?

public:
  T2_IBStream(bool cvtEnd): T2_BStream(), convertEndianFlag(cvtEnd) {};
  virtual ~T2_IBStream() {};

  // need to swap bytes for corrent endianess?
  bool convertEndian() const { return convertEndianFlag; };
  bool setConvertEndian(bool t) { return (convertEndianFlag = t); };

  virtual bool eof() const = 0;

  virtual ssize_t bread(void *buf, int count, size_t size) = 0;

  virtual T2_IBStream & operator >> (bool &b)  = 0;

  virtual T2_IBStream & operator >> (char &c)  = 0;
  virtual T2_IBStream & operator >> (signed char &c)  = 0;
  virtual T2_IBStream & operator >> (unsigned char &c)  = 0;

  virtual T2_IBStream & operator >> (short &s)  = 0;
  virtual T2_IBStream & operator >> (unsigned short &s)  = 0;

  virtual T2_IBStream & operator >> (int &i) = 0;
  virtual T2_IBStream & operator >> (unsigned int &i) = 0;

  virtual T2_IBStream & operator >> (long &l) = 0;
  virtual T2_IBStream & operator >> (unsigned long &l) = 0;

  virtual T2_IBStream & operator >> (float &f) = 0;
  virtual T2_IBStream & operator >> (double &d) = 0;

  // this method reads in a word ended by by blank, tab, newline, '\0', ...
  // it does not read in the entire char[] ended by '\0'
  virtual T2_IBStream & operator >> (char* s);
};



/////////////////////////////////////////////////////////////
//
// T2_OBStream
//
// A base class for an output binary stream.
//
/////////////////////////////////////////////////////////////
class T2_OBStream: public virtual T2_BStream {
private:
protected:
  bool convertEndianFlag;	// need to swap bytes for correct endianess?

public:
  T2_OBStream(bool cvtEnd): T2_BStream(), convertEndianFlag(cvtEnd) {};
  virtual ~T2_OBStream() {};

  // need to swap bytes for corrent endianess?
  bool convertEndian() const { return convertEndianFlag; };
  bool setConvertEndian(bool t) { return (convertEndianFlag = t); };

  virtual ssize_t bwrite(const void *buf, int count, size_t size) = 0;

  virtual T2_OBStream & operator << (bool b) = 0;

  virtual T2_OBStream & operator << (char c) = 0;
  virtual T2_OBStream & operator << (signed char c) = 0;
  virtual T2_OBStream & operator << (unsigned char c)  = 0;

  virtual T2_OBStream & operator << (short s) = 0;
  virtual T2_OBStream & operator << (unsigned short s) = 0;

  virtual T2_OBStream & operator << (int i) = 0;
  virtual T2_OBStream & operator << (unsigned int i) = 0;

  virtual T2_OBStream & operator << (long l) = 0;
  virtual T2_OBStream & operator << (unsigned long l) = 0;

  virtual T2_OBStream & operator << (float f) = 0;
  virtual T2_OBStream & operator << (double d) = 0;

  // this writes the entire char[], including '\0'
  virtual T2_OBStream & operator << (char* s);
};



#endif	/* ! _T2_BSTREAM_H_ */
