
/* t2_header.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/22/97
 * The header file that defines necessary macros for both the front-end and
 * the back-end.
 * This header file should be included by all the T2 programs.
 */

#ifndef _T2_HEADER_H_
#define _T2_HEADER_H_

/* All the T2 header files will check if _ANSI_PROTO_ is defined to decide
 * which style of prototypes to use
 */
#ifndef _ANSI_PROTO_
#   if (defined __cplusplus || defined __STDC__ || defined T2_XLC)
#	define _ANSI_PROTO_
#	undef _NO_PROTO
#   else	/* this is the case where neither _ANSI_PROTO nor_NO_PROTO */
		/* is defined ..... for now, I'll assume _NO_PROTO is used, */
		/* or equivalently, leave _ANSI_PROTO_ undefined */
#   endif
#else		/* _ANSI_PROTO_ */
#   undef _NO_PROTO
#endif		/* _ANSI_PROTO_ */


/* Set macro T2_EXTERN_C to extern "C" if __cplusplus isset, to empty string
 * otherwise.
 */
#ifdef __cplusplus
#  define T2_EXTERN_C		extern "C"
#else
#  define T2_EXTERN_C
#endif


/*
 * One of the following must be defined to decide which asynchronous I/O
 * interface to use:
 *	POSIX_AIO	use POSIX lio_listio, aio_read, aio_write, aio_error ..
 *	AIX_AIO		use AIX lio_listio, aio_read, aio_write, aio_error, ...
 *	SUNOS_AIO	use SunOS aioread, aiowrite, aiowait, ...
 * By default, POSIX style asynchronous I/O functions are used.
 */
#if (defined POSIX_AIO && defined SUNOS_AIO)
#   error Cannot have POSIX_AIO and SUNOS_AIO both defined.
#endif
#if (defined POSIX_AIO && defined AIX_AIO)
#   error Cannot have POSIX_AIO and AIX_AIO both defined.
#endif
#if (defined SUNOS_AIO && defined AIX_AIO)
#   error Cannot have SUNOS_AIO and AIX_AIO both defined.
#endif
#if (!defined POSIX_AIO && !defined SUNOS_AIO && !defined AIX_AIO)
#   define POSIX_AIO
#endif


/*
 * One of the following must be defined to decide which communication library
 * to use for inter-processor communication.
 *	T2_MPI_COMM	use MPI
 *	T2_PVM_COMM	use PVM
 * By default, MPI is used.
 *
 * Actually, PVM is not implemented now, so T2_PVM_COMM cannot be defined.
 */
#ifdef T2_PVM_COMM
#   error PVM cannot be used for the back-end communication.
#endif
#if (defined T2_MPI_COMM && defined T2_PVM_COMM)
#   error Cannot have T2_MPI_COMM and T2_PVM_COMM both defined.
#endif
#if (!defined T2_MPI_COMM && !defined T2_PVM_COMM)
#   define T2_MPI_COMM
#endif



/*
 * T2_ProcBit needs to know the maximum number of back-end processors to
 * decide the length of the bit vector used for the back-end processors.
 * This is given by the macro T2_MAX_NUM_BE_PROCS. If not defined, the
 * default value to be used will be 32.
 */
#ifndef T2_MAX_NUM_BE_PROCS
#   define T2_MAX_NUM_BE_PROCS	256
#endif


/* include stantdard header files */
#include	<sys/types.h>


/* definition for bool */
#if (!defined T2_SYS_SUPPORT_BOOL || !defined __cplusplus)
#  define T2_DEFINE_BOOL
/*
   typedef boolean_t	bool;
#  define false		FALSE
#  define true		TRUE
*/
   enum	bool { false=0, true };
#endif	/* !T2_SYS_SUPPORT_BOOL || !__cplusplus */


/* include T2 header files */
#include	"t2_utility.h"
#include	"t2_defs.h"
#include	"t2_class.h"
#include	"t2_endian.h"


#ifndef NULL
#   define NULL		0
#endif



#endif	/* _T2_HEADER_H_ */
