
/* t2_shape.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 4/16/98
 * Classes for points, boxes, regions, etc. in multi-dimensional spaces
 */


#ifndef _T2_SHAPE_H_
#define _T2_SHAPE_H_


#include	<limits.h>
#include        <values.h>
#include	<math.h>
#include	"t2_header.h"
#include	"t2_bstream.h"
#include	"t2_bstr.h"
#include	"t2_tset.h"
#include	"t2_tsetbstream.h"
#include	"t2_tsetbstr.h"
#include	"t2_tshape.h"


///////////////////////////////////////////////////////////////////
//
// some typedef
//
///////////////////////////////////////////////////////////////////
typedef float			T2_ShapeCoord_t;
typedef T2_TPoint<float>	T2_Point;
typedef T2_TBox<float>		T2_Box;
typedef T2_TRegion<float>	T2_Region;


///////////////////////////////////////////////////////////////////
//
// some operators
//
///////////////////////////////////////////////////////////////////
inline bool
operator ^(const T2_Point& p, const T2_Box& b)
{
  return (b ^ p)? true:false;
}

inline bool
operator ^(const T2_Point& p, const T2_Region& r)
{
  return (r ^ p)? true:false;
}

inline bool
operator ^(const T2_Box& b, const T2_Region& r)
{
  return (r ^ b)? true:false;
}



///////////////////////////////////////////////////////////////////
//
// some functions
//
///////////////////////////////////////////////////////////////////
inline void
t2_setMaxPoint(T2_TPoint<float>& pt)
{
  for (u_int i=0; i<pt.getNumberDimensions(); i++)
     pt[i] = FLT_MAX;
}

inline void
t2_setMinPoint(T2_TPoint<float>& pt)
{
  for (u_int i=0; i<pt.getNumberDimensions(); i++)
     pt[i] = -FLT_MAX;
//     pt[i] = FLT_MIN;
}

inline void
t2_setFullBox(T2_TBox<float>& box)
{
  t2_setMinPoint(box.getLow());
  t2_setMaxPoint(box.getHigh());
}

inline void
t2_setMaxPoint(T2_TPoint<float>& pt, u_int ndim)
{
  pt.setNumberDimensions(ndim);
  for (u_int i=0; i<pt.getNumberDimensions(); i++)
     pt[i] = FLT_MAX;
}

inline void
t2_setMinPoint(T2_TPoint<float>& pt, u_int ndim)
{
  pt.setNumberDimensions(ndim);
  for (u_int i=0; i<pt.getNumberDimensions(); i++)
     pt[i] = -FLT_MAX;
//     pt[i] = FLT_MIN;
}

inline void
t2_setFullBox(T2_TBox<float>& box, u_int ndim)
{
  box.setNumberDimensions(ndim);
  t2_setMinPoint(box.getLow());
  t2_setMaxPoint(box.getHigh());
}



#endif	/* !_T2_SHAPE_H_ */
