
/* t2_tsetbstream.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/7/98
 * inputing/outputing a tset into/from a binary file
 */


#ifndef _T2_TSETBSTREAM_H_
#define _T2_TSETBSTREAM_H_


#include	"t2_header.h"
#include	"t2_bstream.h"
#include	"t2_tset.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_TSetBStreamWriter and T2_TSetBStreamReader
//
// A template for outputing and inputing a T2_TSet. These sets work
// together. The class OutType is assumed to have a bunch of operator <<
// and the class InType is assumed to have a bunch of opertor >> .
// The element class is assumed to have the appropriate operator << and
// operator >> as well.
//
/////////////////////////////////////////////////////////////////////////
template <class Type>
class T2_TSetBStreamWriter {
public:
  T2_TSetBStreamWriter(const T2_TSet<Type>& s): tset(s) {};

  // output method
  T2_OBStream& bwrite(T2_OBStream& os) const;

protected:
  const T2_TSet<Type>& tset;

private:
};

template <class Type>
class T2_TSetBStreamReader {
public:
  T2_TSetBStreamReader(T2_TSet<Type>& s): tset(s) {};

  // input methods
  // bread creates exactly the same set that was written into T2_OBStream
  // (ie with the original length, which could be larger than the number of
  // elements in the set), while breadCompact only allocates space
  // enough to hold the set (ie tset.ndata always equal to tset.len after
  // breadCompact invoked)
  T2_IBStream& bread(T2_IBStream& is);
  T2_IBStream& breadCompact(T2_IBStream& is);

protected:
  T2_TSet<Type>& tset;

private:
};


template <class Type>
inline T2_OBStream&
T2_TSetBStreamWriter<Type>::bwrite(T2_OBStream& os) const
{
  os << tset.len << tset.ndata;
  for (u_int i=0; i<tset.ndata; i++)
     os << tset[i];
  return os;
}

template <class Type>
inline T2_IBStream&
T2_TSetBStreamReader<Type>::bread(T2_IBStream& is)
{
  t2_Assert(is);
  u_int l;
  is >> l;
  tset.emptySet();
  tset.growMaxSetSize(l);
  is >> tset.ndata;
  for (u_int i=0; i<tset.ndata; i++)
     is >> tset.data[i];
  return is;
}

template <class Type>
inline T2_IBStream&
T2_TSetBStreamReader<Type>::breadCompact(T2_IBStream& is)
{
  t2_Assert(is);
  u_int l;
  is >> l >> l;				// throw away the array length
  tset.emptySet();
  tset.growMaxSetSize(l);
  tset.ndata = l;
  for (u_int i=0; i<tset.ndata; i++)
     is >> tset.data[i];
  return is;
}



#endif	/* _T2_TSETBSTREAM_H_ */
