
/* $Id: t2_utility.h,v 1.10 2000/01/21 00:56:09 chialin Exp $ */

/* t2_utility.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/22/97
 * The header file for T2 utility functions/classes.
 */

#ifndef _T2_UTILITY_H_
#define _T2_UTILITY_H_


/* include standard header files */
#include	<stdio.h>
#include	<sys/resource.h>

#ifdef __cplusplus
  #include	<iostream.h>
#endif	/* __cplusplus */


/* include T2 error handlers */
#include	"t2_error.h"


/* some C++ function definitions */
#ifdef __cplusplus
#endif


/* declaration of timing functions */
#ifdef T2_USRDCLOCK_CPP
   /* currently expects user-supplied timing function to be defined in C 
    * (ie no need to demangle the name) so that it can be called in both C and
    * C++ functions.
    */
   #error Currently does not allow T2_USRDCLOCK_CPP to be defined
#endif
#if (defined T2_USRDCLOCK_C && defined T2_USRDCLOCK_CPP)
   #error Cannot have T2_USRDCLOCK_C and T2_USRDCLOCK_CPP both defined.
#endif
#ifdef T2_USRDCLOCK_C
   T2_EXTERN_C double T2_USRDCLOCK_C();
   #define t2_dclock()	T2_USRDCLOCK_C()
#endif
#ifndef t2_dclock
   T2_EXTERN_C double t2_dclock();
#endif



/* t2 system print out */
#ifdef T2_LABEL_IO
#  define t2_say_label(a)		do { \
					   extern int _g_mynode; \
					   a << _g_mynode << ":"; \
					} while (0)

   /* to be backward compatible */
#  define T2_LABEL_ERRMSG

#else /* T2_LABEL_IO */

#  define t2_say_label(a)

#endif /* T2_LABEL_IO */


#ifdef __cplusplus
   /* only valid for C++ code */

#   if (T2_VERBOSE > 0)

	/* example: t2_say1(cerr, "integer i = " << i << endl) */
#	define t2_say1(a, b)		\
					do {\
					   t2_say_label(a); \
					   a << b; \
					   a.flush(); \
					} while(0)

	/* example: t2_cond_say(i < 0, cerr, "integer i = " << i << endl) */
#	define t2_cond_say1(c, a, b)	do {\
					   if (c) { \
						t2_say_label(a); \
						a << b; \
						a.flush(); }\
					} while(0)

#   else					/* T2_VERBOSE > 0 */

#	define t2_say1(a, b)
#	define t2_cond_say1(c, a, b)

#   endif 					/* T2_VERBOSE > 0 */


#   if (T2_VERBOSE > 1)
	/* example: t2_say2(cerr, "integer i = " << i << endl) */
#	define t2_say2(a, b)			t2_say1(a, b)
#	define t2_cond_say2(c, a, b)		t2_cond_say1(c, a, b)

#   else					/* T2_VERBOSE > 1 */
#	define t2_say2(a, b)
#	define t2_cond_say2(c, a, b)
#   endif 					/* T2_VERBOSE > 1 */

#   if (T2_VERBOSE > 2)
	/* example: t2_say3(cerr, "integer i = " << i << endl) */
#	define t2_say3(a, b)			t2_say1(a, b)
#	define t2_cond_say3(c, a, b)		t2_cond_say1(c, a, b)

#   else					/* T2_VERBOSE > 2 */
#	define t2_say3(a, b)
#	define t2_cond_say3(c, a, b)
#   endif 					/* T2_VERBOSE > 2 */

#   if (T2_VERBOSE > 3)
	/* example: t2_say4(cerr, "integer i = " << i << endl) */
#	define t2_say4(a, b)			t2_say1(a, b)
#	define t2_cond_say4(c, a, b)		t2_cond_say1(c, a, b)

#   else					/* T2_VERBOSE > 3 */
#	define t2_say4(a, b)
#	define t2_cond_say4(c, a, b)
#   endif					/* T2_VERBOSE > 3 */

#   if (T2_VERBOSE > 4)
	/* example: t2_say5(cerr, "integer i = " << i << endl) */
#	define t2_say5(a, b)			t2_say1(a, b)
#	define t2_cond_say5(c, a, b)		t2_cond_say1(c, a, b)

#   else					/* T2_VERBOSE > 4 */
#	define t2_say5(a, b)
#	define t2_cond_say5(c, a, b)
#   endif					/* T2_VERBOSE > 4 */

/* to be backward compatible */
#define t2_say(a, b)				t2_say1(a, b)
#define t2_cond_say(c, a, b)			t2_cond_say1(c, a, b)

#endif	/* __cplusplus */



/* define getrusage if it is not defined in sys/resource.h */
#ifndef T2_SYS_DEFINE_GETRUSAGE
   #ifdef _ANSI_PROTO_
      T2_EXTERN_C int getrusage(int, struct rusage*);
   #else
      T2_EXTERN_C int getrusage();
   #endif
#endif	/* ! T2_SYS_DEFINE_GETRUSAGE */


/* prototype for searchString */
#ifdef _ANSI_PROTO_
   T2_EXTERN_C 
	const char *searchString(const char* line, u_int start, u_int* next);
#else
   T2_EXTERN_C 
	const char *searchString();
#endif


/* prototype for T2 string wrappers */
#ifdef _ANSI_PROTO_
   T2_EXTERN_C 
	const char *t2_stringNullWrap(const char* p),
		   *t2_stringEmptyWrap(const char* p);
#else
   T2_EXTERN_C 
	const char *t2_stringNullWrap(), *t2_stringEmptyWrap();
#endif



/* prototype for T2 atoi wrapper */
#ifdef _ANSI_PROTO_
   T2_EXTERN_C 
	int t2_atoi(const char* s, int *flag);
#else
   T2_EXTERN_C
	int t2_atoi();
#endif


#endif	/* _T2_UTILITY_H_ */
