
/* $Id: signal.c,v 1.1 2000/02/24 15:58:29 chialin Exp $ */

/* signal.C */

#include	"t2_signal.h"

/*
 * A wrapper function for the function signal.
 * Adapted from Stevens's UNIX Network Programming.
 * Argument signo should be a signal name and argument func should
 * be either a pointer to a function or one of the constants SIG_IGN 
 * (for ignoring the signal) or SIG_DFL (for resetting back to the default
 * handler).
 * The function returns the original signal handler.
 */

T2_Sigfunc*
t2_signal(int signo, T2_Sigfunc *func)
{
  struct sigaction act, oact;

  act.sa_handler = func;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if (signo == SIGALRM) {
#    ifdef SA_INTERRUPT
	act.sa_flags |= SA_INTERRUPT;	/* SunOS 4.x */
#    endif
  }
  else {
#    ifdef SA_RESTART
	act.sa_flags |= SA_RESTART;	/* SVR4, 4.4BSD */
#    endif
  }

  if (sigaction(signo, &act, &oact) < 0)
     return SIG_ERR;
  return oact.sa_handler;
}
