
/* $Id: socket.C,v 1.13 2000/03/16 00:48:11 chialin Exp $ */

/* socket.C */

#include	<string.h>
#include	"t2_socket.h"


////////////////////////////////////////////////////////////////////////
//
// blocking write and read functions
//
////////////////////////////////////////////////////////////////////////
ssize_t
T2_Sock::bWrite(const void *buf, int count, size_t size, ssize_t& nbytes_wrote)
{
  nbytes_wrote = 0;

  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     return -1;
  }

  size_t nbytes_left = count * size;
  if (nbytes_left > 0 && buf == NULL) {
     errval = EFAULT;
     return -1;
  }


# if T2_VERBOSE > 4
  {
     cout << "gonna bWrite(," << count << "," << size << ",):";
     char *bbp = (char*) buf;
     for (int xx=0; xx<count; xx++) {
        for (u_int yy=0; yy<size; yy++) {
           cout << " " << (int) *bbp;
           bbp++;
        }
        cout << " |";
     }
     cout << endl;
     cout.flush();
  }
# endif

# ifdef T2_SHOW_SOCKET_DATA
    cout << "#### bWrite:";
# endif

  errno = 0;
  while (nbytes_left > 0) {
     ssize_t s;
     if ((s = write(sockfd, buf, nbytes_left)) < 0) {	// write error
	errval = errno;
	if (nbytes_wrote == 0)		// write never succeeded
	   return -1;

#	ifdef T2_SHOW_SOCKET_DATA
	{
	   const char* pppp=(const char*) buf;
	   for (int kk=0; kk<nbytes_wrote; kk++)
	      cout << " " << (int) pppp[kk];
	   cout << endl;
	}
#	endif

	return nbytes_wrote;
     }
     t2_Assert ((u_int) s <= nbytes_left);

     if (s == 0 && errno != EINTR) { 	// cannot write anymore ?
	errval = errno;
	break;
     }

#    ifdef T2_SHOW_SOCKET_DATA
     {
	const char* pppp=(const char*) buf;
	for (int kk=0; kk<s; kk++)
	   cout << " " << (int) pppp[kk];
     }
#    endif

     nbytes_wrote += s;
     nbytes_left -= s;
     buf = (char *) buf + s;

     // not sure if one should sleep here ??
     // for now, sleep for 100000 usec (100ms)
     usleep(10000);
  }

# ifdef T2_SHOW_SOCKET_DATA
     cout << endl;
# endif

  errval = 0;				// write succeeded
  return nbytes_wrote;
}


ssize_t
T2_Sock::bRead(void *buf, int count, size_t size, ssize_t& nbytes_read)
{
  nbytes_read = 0;
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     return -1;
  }

  size_t nbytes_left = count * size;
  if (nbytes_left > 0 && buf == NULL) {
     errval = EFAULT;
     return -1;
  }


# if T2_VERBOSE > 4
     char *bbp = (char*) buf;
# endif

# ifdef T2_SHOW_SOCKET_DATA
     cout << "#### bRead:";
# endif

  errno = 0;
  while (nbytes_left > 0) {
     ssize_t s;
     if ((s = read(sockfd, buf, nbytes_left)) < 0) {	// read error
	errval = errno;
	if (nbytes_read == 0) 		// read never succeeded !!
	   return -1;

#	ifdef T2_SHOW_SOCKET_DATA
	{
	   const char* pppp=(const char*) buf;
	   for (int kk=0; kk<nbytes_read; kk++)
	       cout << " " << (int) pppp[kk];
	   cout << endl;
	}
#	endif

	return nbytes_read;
     }

     t2_AssertOrDo ((u_int) s <= nbytes_left,
		    cout << "s = " << s << " nbytes_left = " << nbytes_left
			 << endl);
     if (s == 0 && errno != EINTR) { 	// eof ??
	errval = errno;
	closeSocket();
	break;
     }

#    ifdef T2_SHOW_SOCKET_DATA
	const char* pppp=(const char*) buf;
	for (int kk=0; kk<s; kk++)
	   cout << " " << (int) pppp[kk];
#    endif

     nbytes_read += s;
     nbytes_left -= s;
     buf = (char *) buf + s;

     // not sure if one should sleep here ??
     // for now, sleep for 100000 usec (100ms)
     usleep(10000);
  }

# ifdef T2_SHOW_SOCKET_DATA
     cout << endl;
# endif

  t2_say4(cout, "bread intends to read " << count*size << " bytes, returns "
		<< nbytes_read << endl);


# if T2_VERBOSE > 4
  {
     cout << "just bRead(," << count << "," << size << ",):";
     for (int xx=0; xx<count; xx++) {
        for (u_int yy=0; yy<size; yy++) {
           cout << " " << (int) *bbp;
           bbp++;
        }
        cout << " |";
     }
     cout << endl;
     cout.flush();
  }
# endif

  errval = 0;				// read succeeded
  return nbytes_read;
}


////////////////////////////////////////////////////////////////////////
//
// T2_Sock methods
//
////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
//
// T2_Socket methods
//
////////////////////////////////////////////////////////////////////////
void
T2_Socket::transfer(T2_Socket& s)
{
  closeSocket();

  // get data from s
  memcpy((void *) &remote, (const void *) &s.remote, sizeof(remote));
  nbytes_read = s.nbytes_read;
  nbytes_wrote = s.nbytes_wrote;
  setInputConvertEndian(s.inputConvertEndian());
  setOutputConvertEndian(s.outputConvertEndian());
  T2_Sock::operator = (s);

  // clear s
  bzero((void *) &s.remote, sizeof(s.remote));
  s.nbytes_read = 0;
  s.nbytes_wrote = 0;
  s.sockfd = -1;
  s.errval = 0;
  s.isconnect = false;
}


bool
T2_Socket::connectByName(const char* hname, short port)
{
  closeSocket();
  if (hname == NULL) {
     errval = EADDRNOTAVAIL;		// Can't assign requested address
     return false;
  }

  struct hostent *host;
  if ((host = gethostbyname(hname)) == NULL) {
     errval = h_errno;
     t2_err_ret(__FILE__, __LINE__,
		    "gethostbyname error for host %s (%s)",
		    hname, hstrerror(h_errno));
     return false;
  }
  t2_Assert (host->h_addr_list != NULL);
  char **addr = host->h_addr_list;

  bzero((void *) &remote, sizeof(remote));
  remote.sin_family = AF_INET;
  remote.sin_port   = htons(port);
  memcpy(&remote.sin_addr, *addr, sizeof(struct in_addr));

  if (connectBySockAddr(remote) == false) {
     t2_err_ret(__FILE__, __LINE__, "cannot connect to server %s", hname);
     return false;
  }

  return true;
}


bool
T2_Socket::connectByAddress(const char* ipaddr, short port)
{
  closeSocket();
  if (ipaddr == NULL) {
     errval = EADDRNOTAVAIL;		// Can't assign requested address
     return false;
  }

  bzero((void *) &remote, sizeof(remote));
  remote.sin_family = AF_INET;
  remote.sin_port   = htons(port);
  if (!inet_pton(AF_INET, ipaddr, &remote.sin_addr)) {
     errval = errno;
     return false;
  }

  if (connectBySockAddr(remote) == false) {
     t2_err_ret(__FILE__, __LINE__, "cannot connect to server %s", ipaddr);
     return false;
  }

  return true;
}


bool
T2_Socket::connectBySockAddr(struct sockaddr_in& serv_addr)
{
  t2_Assert (connected() == false);
  if (openSocket() == false)
     return false;

  const int num_attempts = 10;
  int na;
  for (na=0; na<num_attempts; na++) {
     if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
     {
        errval = errno;

	// do I need to check for EISCONN?

	if (errval != EINTR) {
           t2_err_ret(__FILE__, __LINE__, "cannot connect to server");
           closeSocket();
           return false;
	}
	usleep(100000);
     }
     else
	break;				// connect succeeded !
  }

  if (na == num_attempts) {		// too many failures
     t2_err_ret(__FILE__, __LINE__, "cannot connect to server");
     closeSocket();
     return false;
  }

  isconnect = true;
  setErr(0);
 
  return true;
}


bool
T2_Socket::getPeerAddress(char* buf, size_t len)
{
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     return false;
  }

  if (buf == NULL) {
     errval = EADDRNOTAVAIL;		// Can't assign requested address
     return false;
  }

  if (inet_ntop(AF_INET, &(remote.sin_addr), buf, len) == NULL) {
     errval = errno;
     return false;
  }

  return true;
}


short
T2_Socket::getPeerPort()
{
  if (connected() == false) {
     errval = ENOTCONN;
     return -1;
  }
  return ntohs(remote.sin_port);
}


//ssize_t
//T2_Socket::bwrite(const void *buf, int count, size_t size)
//{
//  nbytes_wrote = 0;
//
//  if (connected() == false) {
//     errval = ENOTCONN;			// Socket is not connected
//     return -1;
//  }
//
//  size_t nbytes_left = count * size;
//  if (nbytes_left > 0 && buf == NULL) {
//     errval = EFAULT;
//     return -1;
//  }
//
//  while (nbytes_left > 0) {
//     ssize_t s;
//     if ((s = write(sockfd, buf, nbytes_left)) < 0) {	// write error
//	errval = errno;
//	if (nbytes_wrote == 0)		// write never succeeded
//	   return -1;
//	return nbytes_wrote;
//     }
//     t2_Assert (s <= nbytes_left);
//
//     if (s == 0)			// cannot write anymore ?
//	break;
//
//     nbytes_wrote += s;
//     nbytes_left -= s;
//     buf = (char *) buf + s;
//
//     // not sure if one should sleep here ??
//     // for now, sleep for 100000 usec (100ms)
//     usleep(10000);
//  }
//  errval = 0;				// write succeeded
//  return nbytes_wrote;
//}


//ssize_t
//T2_Socket::bread(void *buf, int count, size_t size)
//{
//  nbytes_read = 0;
//  if (connected() == false) {
//     errval = ENOTCONN;			// Socket is not connected
//     return -1;
//  }
//
//  size_t nbytes_left = count * size;
//  if (nbytes_left > 0 && buf == NULL) {
//     errval = EFAULT;
//     return -1;
//  }
//
//  while (nbytes_left > 0) {
//     ssize_t s;
//     if ((s = read(sockfd, buf, nbytes_left)) < 0) {	// read error
//	errval = errno;
//	if (nbytes_read == 0) 		// read never succeeded !!
//	   return -1;
//	return nbytes_read;
//     }
//
//     t2_AssertOrDo (s <= nbytes_left,
//		    cout << "s = " << s << " nbytes_left = " << nbytes_left
//			 << endl);
//     if (s == 0) { 			// eof
//	closeSocket();
//	break;
//     }
//
//     nbytes_read += s;
//     nbytes_left -= s;
//     buf = (char *) buf + s;
//
//     // not sure if one should sleep here ??
//     // for now, sleep for 100000 usec (100ms)
//     usleep(10000);
//  }
//
////cout << "bread intends to read " << count*size << ", returns "
////<< nbytes_read << endl;
//
//  errval = 0;				// read succeeded
//  return nbytes_read;
//}


ssize_t
T2_Socket::nbread(void *buf, int count, size_t size)
{
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     nbytes_read = 0;
     return -1;
  }

  if ((nbytes_read = read(sockfd, buf, count*size)) < 0) {	// read error
     errval = errno;
     nbytes_read = 0;
     return -1;
  }

  if (nbytes_read == 0)			// eof
     closeSocket();

  errval = 0;				// read succeeded
  return nbytes_read;
}



////////////////////////////////////////////////////////////////////////
//
// T2_SocketServer methods
//
////////////////////////////////////////////////////////////////////////
short
T2_SocketServer::getLocalPort() const
{
  if (connected() == false)
     return -1;

  return ntohs(servaddr.sin_port);
}


bool
T2_SocketServer::startServer(int p, int backlog)
{
  closeSocket();
  if (openSocket() == false)
     return false;

  bzero((void *) &servaddr, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_port = htons(p);
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(sockfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0) {
     errval = errno;
     closeSocket();
     t2_err_ret(__FILE__, __LINE__, "bind failed");
     return false;
  }


# if T2_VERBOSE > 1
  { // this is just for debugging
    char ipaddr[INET_ADDRSTRLEN];
    cout << "serverSocket just binds to address "
         << inet_ntop(AF_INET, &servaddr.sin_addr, ipaddr, sizeof(ipaddr))
         << endl;
    cout.flush();
  }
# endif

  if (listen(sockfd, backlog) < 0) {
     errval = errno;
     closeSocket();
     t2_err_ret(__FILE__, __LINE__, "listen failed");
     return false;
  }

  socklen_t sizeofSockAddr = sizeof(servaddr);
  getsockname(sockfd, (struct sockaddr*)&servaddr, &sizeofSockAddr);

  // turn off nonblocking explicitly
  if (unsetFcntlFlag(sockfd, O_NONBLOCK) < 0) {
     errval = errno;
     t2_err_ret(__FILE__, __LINE__, "setFcntlFlag error");
  }
  
  isconnect = true;
  errval = 0;
  return true;
}

bool
T2_SocketServer::pollNewConnection(T2_Socket& s, struct timeval* timeout)
{
  s.isconnect = false;
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     return false;
  }

  fd_set rset;
  FD_ZERO(&rset);
  FD_SET(sockfd, &rset);

  int r;
  if ((r = select(sockfd+1, &rset, NULL, NULL, timeout)) < 0) {
     errval = errno;
     t2_err_ret(__FILE__, __LINE__, "select returns error %d", errval);
     return false;
  }
  if (r == 0) {				// time out
     errval = ETIMEDOUT;
     return false;
  }

  if (FD_ISSET(sockfd, &rset)) {	// a new connection arrives
     s.closeSocket();
     struct sockaddr_in& newconn = s.remote;
     int& connfd = s.sockfd;
     socklen_t sizeofSockAddr = sizeof(newconn);
     if ((connfd = accept(sockfd, (struct sockaddr *)&newconn, &sizeofSockAddr))
        < 0)
     {
	switch (errno) {
	  case EWOULDBLOCK:
	  case ECONNABORTED:
	  case EPROTO:
	  case EINTR:
			// ignore these errors for reasons stated in
			// Steven's Unix Network Programming, Section 15.6,
			// and no new connection is made
			errval = ETIMEDOUT;
			break;
	  default:
			errval = errno;
			t2_err_ret(__FILE__, __LINE__,
				   "accept returns error %d", errval);
	}
	return false;
     }

     s.isconnect = true;
     errval = 0;
     return true;			// a new connection arrives
  }

  errval = 0;
  return false;				// no new connection
}


bool
T2_SocketServer::acceptConnection(T2_Socket& s)
{
  s.isconnect = false;
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     return false;
  }

  struct sockaddr_in& newconn = s.remote;
  int& connfd = s.sockfd;
  socklen_t sizeofSockAddr = sizeof(newconn);
  if ((connfd = accept(sockfd, (struct sockaddr *)&newconn, &sizeofSockAddr))
     < 0)
  {
     errval = errno;
     t2_err_ret(__FILE__, __LINE__, "accept returns error %d", errval);
     return false;
  }

  s.isconnect = true;
  errval = 0;
  return true;
}
