
/* $Id: txtfile.C,v 1.4 2000/01/25 20:52:33 chialin Exp $ */

/* txtfile.C */


#include	<stdlib.h>
#include	"t2_txtfile.h"

////////////////////////////////////////////////////////////////////////////
//
// T2_TxtFile methods
//
////////////////////////////////////////////////////////////////////////////
T2_TxtFile::T2_TxtFile(char m, size_t s): ifs(), mark(m), filename(NULL),
					  maxlinelen(s), lineno(0), linepos(0),
					  valid(false)
{
  line = new char[maxlinelen];
  cur = NULL;
  eof_f = true;
}


void
T2_TxtFile::getNextLine()
{
  if (eof() == true || valid == false)
     return;

  linepos = ifs.tellg();
  while (ifs.getline(line, maxlinelen) || ifs.gcount() > 0) {
     lineno++;
     if (line[0] != mark) {		 		// not a comment !
	linelen = strlen(line);
	cur = line;
	return;
     }

     // this is a comment line, so get another line
     linepos = ifs.tellg();
  }

  eof_f = true;						// eof
  cur = NULL;
}


void
T2_TxtFile::getNextNonemptyLine()
{
  if (eof() == true || !ifs)
     return;

  linepos = ifs.tellg();
  while (ifs.getline(line, maxlinelen) || ifs.gcount() > 0) {
     lineno++;
     u_int next;
     cur = (char *) searchString(line, 0, &next);
     if (cur != NULL && cur[0] != mark) { 		// found !
	linelen = strlen(line);
	return;
     }

     // empty line or comment, so get another line
     linepos = ifs.tellg();
  }

  eof_f = true;						// eof
  cur = NULL;
}


const char*
T2_TxtFile::getNextUncommentString()
{
  if (cur == NULL) {
     getNextNonemptyLine();
     cur = line;
  }

  while (eof() == false) {
     u_int nn;
     const char* string = searchString(line, cur-line, &nn);
     if (string == NULL || string[0] == mark) {		// empty line or comment
	getNextNonemptyLine();
	cur = line;
	continue;
     }
     line[nn-1] = '\0';

     // advance cur
     if (nn > linelen) 		// string is the last string of the line
	cur = NULL;
     else
        cur = line+nn;

     return string;
  }

  return NULL;			// eof
}


const char*
T2_TxtFile::getStartOfNextUncommentString(const char* buf) const
{
  if (buf == NULL)
     return NULL;

  while (isspace(*buf) && *buf != mark && *buf != '\0')
     buf++;

  if (*buf == '\0' || *buf == mark)	// eof of buf or start of comment
     return NULL;

  return buf;
}


char*
T2_TxtFile::copyNextUncommentString()
{
  const char* string = getNextUncommentString();
  if (string == NULL)
     return NULL;

  char* ret_string;
  size_t string_len = strlen(string)+1;
  t2_Alloc(ret_string, string_len, char);
  memcpy(ret_string, string, string_len);
  return ret_string;
}


bool
T2_TxtFile::copyNextUncommentString(char* buf)
{
  if (buf == NULL)
     return false;

  const char* string = getNextUncommentString();
  if (string == NULL)
     return false;

  strcpy(buf, string);
  buf[strlen(string)] = '\0';
  return true;
}


T2_TxtFile&
T2_TxtFile::operator >> (char& c)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        c = string[0];
  }

  return *this;
}


T2_TxtFile&
T2_TxtFile::operator >> (short& s)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        s = (short) atoi(string);
  }
  return *this;
}


T2_TxtFile&
T2_TxtFile::operator >> (unsigned short& s)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        s = (unsigned short) atoi(string);
  }
  return *this;
}


T2_TxtFile&
T2_TxtFile::operator >> (int& s)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        s = (int) atoi(string);
  }
  return *this;
}


T2_TxtFile&
T2_TxtFile::operator >> (unsigned int& s)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        s = (unsigned int) atoi(string);
  }
  return *this;
}


T2_TxtFile&
T2_TxtFile::operator >> (long& s)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        s = atol(string);
  }
  return *this;
}


T2_TxtFile&
T2_TxtFile::operator >> (unsigned long& s)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        s = (unsigned long) atol(string);
  }
  return *this;
}


T2_TxtFile&
T2_TxtFile::operator >> (float& s)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        s = (float) atof(string);
  }
  return *this;
}


T2_TxtFile&
T2_TxtFile::operator >> (double& s)
{
  if (eof()==false && ifs) {
     const char* string = getNextUncommentString();
     if (string != NULL)
        s = (double) strtod(string, (char**) NULL);
  }
  return *this;
}

