
/* $Id: svm_front_end.h,v 1.1 2000/02/13 17:23:50 chialin Exp $ */

/* svm_front_end.h */

#ifndef _SVM_FRONT_END_H
#define _SVM_FRONT_END_H

#include	<sys/types.h>


////////////////////////////////////////////////////////////////////////
//
// class svm_ImageInfoEntry
//
// A class to store the information regarding a svm dataset obtained
// bying inquiring the T2 front-end.
//
////////////////////////////////////////////////////////////////////////
class svm_ImageInfoEntry {
public:
  svm_ImageInfoEntry();
  ~svm_ImageInfoEntry() { delete [] dataset_name; delete [] thumbnail_file; };

  // const access methods
  const char* getDatasetName() const { return dataset_name; };
  u_int getDatasetID() const { return did; };
  u_int getIndexID() const { return iid; };
  u_int getProjectionID() const { return pid; };
  u_int getAccMetaID() const { return accid; };
  u_int getAggregationID() const { return aggid; };
  u_int getImageWidth() const { return x_res; };
  u_int getImageHeight() const { return y_res; };
  const char* getThumbnailFilename() const { return thumbnail_file; };
  size_t getThumbnailFileSize() const { return thumbnail_nbytes; };

  // non-const access methods
  char*& getDatasetName()  { return dataset_name; };
  u_int& getDatasetID()  { return did; };
  u_int& getIndexID()  { return iid; };
  u_int& getProjectionID()  { return pid; };
  u_int& getAccMetaID()  { return accid; };
  u_int& getAggregationID()  { return aggid; };
  u_int& getImageWidth()  { return x_res; };
  u_int& getImageHeight()  { return y_res; };
  char*& getThumbnailFilename()  { return thumbnail_file; };
  size_t& getThumbnailFileSize()  { return thumbnail_nbytes; };

  // write output
  ostream& write(ostream& os) const;

protected:
  char* dataset_name;			// dataset name
  u_int did,				// dataset id
	iid,				// index id
	pid,				// projection function id
	accid,				// accumulator meta object function id
	aggid;				// aggregation function id
  u_int x_res, y_res;			// width and height of image

  char* thumbnail_file;			// filename that stores the thumbnail
  size_t thumbnail_nbytes;		// size of thumbmail in bytes

private:
};


inline
svm_ImageInfoEntry::svm_ImageInfoEntry()
	: dataset_name(NULL), did(0), iid(0), pid(0), accid(0), aggid(0),
	   x_res(0), y_res(0), thumbnail_file(NULL), thumbnail_nbytes(0)
{
}

inline ostream&
svm_ImageInfoEntry::write(ostream& os) const
{
  os << "dataset name= " << t2_stringNullWrap(dataset_name) << endl
     << "dataset id= " << did << " index id= " << iid
     << " project id= " << pid << " acc id= " << accid
     << " aggr id= " << aggid << endl
     << "image width= " << x_res << " height= " << y_res << endl
     << "thumbnail filename= " << t2_stringNullWrap(thumbnail_file)
     << " size= " << thumbnail_nbytes << endl;
  return os;
}

inline ostream&
operator << (ostream& os, const svm_ImageInfoEntry& e)
{
  return e.write(os);
}



////////////////////////////////////////////////////////////////////////
//
// class svm_ImageInfoCollection
//
////////////////////////////////////////////////////////////////////////
class svm_ImageInfoCollection {
public:
  svm_ImageInfoCollection();
  ~svm_ImageInfoCollection();

  // query methods
  u_int getNumberEntries() const { return nentries; };

  // access operators
  const svm_ImageInfoEntry* operator [] (u_int i) const
				{ return (i>=nentries)? NULL : entries[i]; };
  svm_ImageInfoEntry*& operator [] (u_int i) { return entries[i]; };

  // add n more entries to the array
  void growSize(u_int n);

  // remove all entries
  void cleanup();

protected:
  u_int nentries;			// number of dataset entries
  svm_ImageInfoEntry** entries;		// an array of entries
					// (note that some of the entries
					// may be NULL)

private:
};


inline 
svm_ImageInfoCollection::svm_ImageInfoCollection()
			: nentries(0), entries(NULL)
{
}

inline
svm_ImageInfoCollection::~svm_ImageInfoCollection()
{
  for (u_int i=0; i<nentries; i++)
    delete entries[i];
  delete [] entries;
}

inline void
svm_ImageInfoCollection::growSize(u_int n)
{
  svm_ImageInfoEntry** tmp = entries;
  entries = new svm_ImageInfoEntry*[nentries+n];
  for (u_int i=0; i<nentries; i++)
    entries[i] = tmp[i];
  for (u_int j=nentries; j<nentries+n; j++)
    entries[j] = NULL;
  nentries += n;
  delete [] tmp;
}

inline void
svm_ImageInfoCollection::cleanup()
{
  for (u_int i=0; i<nentries; i++)
    delete entries[i];
  delete [] entries;

  nentries = 0;
  entries = NULL;
}

inline ostream&
operator << (ostream& os, const svm_ImageInfoCollection& c)
{
  for (u_int i=0; i<c.getNumberEntries(); i++) {
    if (c[i] == NULL)
	os << "dataset entry " << i << ": NULL" << endl;
    else
	os << "dataset entry " << i << ":" << endl << *(c[i]);
  }
  return os;
}



#endif	/* ! _SVM_FRONT_END_H */
