
/* $Id: svm_acc.h,v 1.2 2000/03/09 01:31:05 chialin Exp $ */

/* svm_acc.h */


#ifndef _SVM_ACC_H_
#define _SVM_ACC_H_

#include	"t2_acc.h"
#include	"svm_image.h"


/////////////////////////////////////////////////////////////////////////
//
// svm_ImageAcc
//
/////////////////////////////////////////////////////////////////////////
class svm_ImageAcc: public svm_Image, public T2_Accumulator {
public:
  svm_ImageAcc(u_int tw, u_int th,
	       u_int xp, u_int yp, u_int w, u_int h, u_int s, u_int v);
  ~svm_ImageAcc() {};

  // virtual methods
  virtual T2_UDFRet navigateAll(const T2_QueryInfo& qinfo,
				T2_AccIterator*& aitp);
  virtual T2_UDFRet navigate(const T2_QueryInfo& qinfo, T2_IteratorID i,
			     const T2_Region& r, T2_AccIterator*& aitp);

  // query methods for info on total output image
  u_int getTotalOutputImageWidth() const { return x_res; };
  u_int getTotalOutputImageHeight() const { return y_res; };

  // query methods
  u_int getXPosition() const { return x_pos; };
  u_int getYPosition() const { return y_pos; };

protected:
  // the following is the information regarding the entire output image
  u_int x_res, y_res;		// width and height of the entire output image

  // the following is the information regarding the tile image;
  // the information regarding the entire output image is kept with
  // svm_ImageAccMeta (some are also stored in this class)
  u_int x_pos, y_pos;		// coordinate in the output space for the
				// starting (ie top-left) pixel

private:
};



/////////////////////////////////////////////////////////////////////////
//
// svm_ImageAccIterator
//
/////////////////////////////////////////////////////////////////////////
class svm_ImageAccIterator: public svm_ImageIterator, public T2_AccIterator {
public:
  svm_ImageAccIterator(u_int w, u_int h, u_int psize, u_char* b,
		       u_int x_s, u_int y_s, u_int x_e, u_int y_e);
  ~svm_ImageAccIterator() {};

  // virtual methods
  virtual T2_UDFRet getNextElement(const T2_QueryInfo& qinfo,
				   T2_Accumulator& acc, bool& eod, void*& elem);

protected:
  svm_ImagePixel pixel;		// an object to host the pixel to be returned

private:
};


inline
svm_ImageAccIterator::
svm_ImageAccIterator(u_int w, u_int h, u_int psize, u_char* b,
		     u_int x_s, u_int y_s, u_int x_e, u_int y_e)
		: svm_ImageIterator(w, h, psize, b, x_s, y_s, x_e, y_e),
		  T2_AccIterator(), pixel()
{
}



/////////////////////////////////////////////////////////////////////////
//
// app_ImageAccMeta
//
/////////////////////////////////////////////////////////////////////////
class svm_ImageAccMeta: public T2_AccMetaObj {
public:
  svm_ImageAccMeta(u_int w, u_int h, u_int ps)
				: width(w), height(h), pixelsize(ps) {};
  ~svm_ImageAccMeta() {};

  // virtual functions
  virtual T2_UDFRet allocAcc(const T2_QueryInfo& qinfo,
			     T2_Iteration i, const T2_Region& r,
			     T2_Accumulator*& accp);
  virtual T2_UDFRet stripMine(const T2_QueryInfo& qinfo, size_t mem,
			      const T2_ClusterInfoList& infolist,
			      T2_VArray<T2_Region>& oqw);

protected:
  u_int width, height,		// size of the entire output image (in # pixels)
	pixelsize;		// number of bytes per pixel

  T2_UDFRet partitionByRows(size_t mem, T2_VArray<T2_Region>& oqw) const;
  T2_UDFRet partitionByCols(size_t mem, T2_VArray<T2_Region>& oqw) const;

private:
};



#endif	/* _SVM_ACC_H_ */
