
/* $Id: svm_accavg.C,v 1.2 2000/03/09 01:33:05 chialin Exp $ */

/* svm_accavg.C */


#include	<stdio.h>
#include	<iostream.h>
#include	<stdlib.h>
#include	<errno.h>
#include	"svm_accavg.h"
#include	"svm_utility.h"


/////////////////////////////////////////////////////////////////////
//
// svm_ImageAccAVG methods
//
/////////////////////////////////////////////////////////////////////
svm_ImageAccAVG::svm_ImageAccAVG(u_int tw, u_int th,
				 u_int xp, u_int yp, u_int w, u_int h,
				 u_int s, u_int v)
	: T2_Accumulator(), x_res(tw), y_res(th),
	  x_pos(xp), y_pos(yp), width(w), height(h), pixelsize(s), maxval(v)
{
  u_int tot_npixels = w * h;
  n_sumcounts = tot_npixels * (pixelsize +1);
  t2_Alloc (sum_count, n_sumcounts, u_int);
  if (sum_count == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to allocate %d bytes for sum_count[].", 
    n_sumcounts *sizeof(u_int));
  }

  // since there should be only one reasonable way to initialize the sums
  // and counts, so I'll just initialize them here;
  // if the aggregation function decides to initialize them some other way,
  // it can still do it during the initialization phase
  for (u_int i=0; i<n_sumcounts; i++)
      sum_count[i] = 0;
}


T2_UDFRet
svm_ImageAccAVG::navigateAll(const T2_QueryInfo& qinfo, T2_AccIterator*& aitp)
{
  aitp = (T2_AccIterator *)
	new svm_ImageAccAVGIterator(width, height, pixelsize,
				    (u_char *) sum_count,
				    0, 0, getWidth()-1, getHeight()-1);
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_ImageAccAVG::navigate(const T2_QueryInfo& qinfo, T2_IteratorID i,
		          const T2_Region& r, T2_AccIterator*& aitp)
{
  // ignore iterator id, since there is only one iterator

  if (r.getNumberBoxes() != 1) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: expecting a region with 1 box, but got one with %d boxes",
	r.getNumberBoxes());
     return T2_UDFRet_ERROR;
  }

  const u_int ndims = 2;
  const T2_Box& box = r[0];
  if (box.getNumberDimensions() != ndims) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: expecting a %d-dim region, but got one with %d dims",
	ndims, box.getNumberDimensions());
     return T2_UDFRet_ERROR;
  }
  u_int x_low = (u_int) box.getLow()[0],
	y_low = (u_int) box.getLow()[1],
	x_high = (u_int) box.getHigh()[0],
	y_high = (u_int) box.getHigh()[1];

  t2_Assert (x_high >= x_pos);
  t2_Assert (y_high >= y_pos);

  // need to translate the area between (x_low, y_low) and (x_high, y_high)
  // into local array index inside the chunk for fast iterating;
  u_int x_start, y_start, x_end, y_end;
  if (findIntersectionInLocalOffsets(
		x_pos, y_pos, x_pos+width-1, y_pos+height-1,
		(u_int) box.getLow()[0], (u_int) box.getLow()[1],
		(u_int) box.getHigh()[0], (u_int) box.getHigh()[1],
		x_start, y_start, x_end, y_end)
      == false)						// no intersection
  {
     aitp = (T2_AccIterator *)
		new svm_ImageAccAVGIterator(0, 0, 0, (u_char *) NULL,
					    1, 1, 0, 0);
  }
  else {
     aitp = (T2_AccIterator *)
		new svm_ImageAccAVGIterator(width, height, pixelsize,
					    (u_char *) sum_count,
					    x_start, y_start, x_end, y_end);
  }

//  u_int x_start = (x_low > x_pos)? (x_low - x_pos) : 0,
//        y_start = (y_low > y_pos)? (y_low - y_pos) : 0,
//        x_end = x_high - x_pos,
//        y_end = y_high - y_pos;
//
//  aitp = (T2_AccIterator *)
//	new svm_ImageAccAVGIterator(width, height, pixelsize,
//				    (u_char *) sum_count,
//				    x_start, y_start, x_end, y_end);

  return T2_UDFRet_OK;
}



/////////////////////////////////////////////////////////////////////
//
// svm_ImageAccIterator methods
//
/////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageAccAVGIterator::getNextElement(const T2_QueryInfo& qinfo,
				     T2_Accumulator& acc,
				     bool& eod, void*& elem)
{
  u_int x_off, y_off;			// offsets for pixel to return

  if (svm_ImageIterator::getNextElement(x_off, y_off, pixel) == false) {
     // end-of-data is true
     eod = true;
     return T2_UDFRet_OK;
  }

  elem = (void *) &pixel;
  eod = false;
  return T2_UDFRet_OK;
}



/////////////////////////////////////////////////////////////////////
//
// svm_ImageAccMetaAVG methods
//
/////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageAccMetaAVG::allocAcc(const T2_QueryInfo& qinfo,
			   T2_Iteration i, const T2_Region& tile_mbr,
			   T2_Accumulator*& accp)
{
  if (tile_mbr.getNumberBoxes() != 1) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: expecting a single box for the tile's mbr, but got %d boxes instead.",
	tile_mbr.getNumberBoxes());
     return T2_UDFRet_ERROR;
  }

  const u_int ndims = 2;		// can only handle 2-dim boxes
  const T2_Box& mbr = tile_mbr[0];
  if (mbr.getNumberDimensions() != ndims) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: expecting a %d-dim tile mbr, but got a %d-dim box instead.",
	ndims, mbr.getNumberDimensions());
     return T2_UDFRet_ERROR;
  }

  const T2_Point& low = mbr.getLow();
  const T2_Point& high = mbr.getHigh();
   
  u_int tile_width = (u_int) high[0] - (u_int) low[0] + 1,
        tile_height = (u_int) high[1] - (u_int) low[1] + 1;

  // for now, assume that the max value per color component is always 255
  const u_int maxval = 255;
  accp = (T2_Accumulator *)
		new svm_ImageAccAVG(width, height, (u_int) low[0], (u_int) low[1],
				 tile_width, tile_height, pixelsize, maxval);

  return T2_UDFRet_OK;
}


T2_UDFRet
svm_ImageAccMetaAVG::stripMine(const T2_QueryInfo& qinfo, size_t mem,
			    const T2_ClusterInfoList& infolist,
			    T2_VArray<T2_Region>& tile_mbrs)
{
  if (mem < sizeof(svm_ImageAccAVG)) {	// this shouldn't happen
     t2_err_ret(__FILE__,__LINE__,
	"ERROR:not enough memory %d to even store an svm_ImageAccAVG of %d bytes",
	mem, sizeof(svm_ImageAcc));
     return T2_UDFRet_ERROR;
  }

  mem -= sizeof(svm_ImageAccAVG);	    // reserve space for an svm_ImageAcc obj

  // since the accumulator, which stores a sum and a count per pixel, cannot be
  // used as the final output, when we get to the output handling phase, we will
  // have to keep both the accumulator and the output image in memory at the
  // same time; so this should be taken into account when stripmining the output
  // image;
  // what this means is that for the accumulator, each pixel effectively has, 
  // instead of just pixelsize bytes, pixelsize + (pixelsize * 2 + 1) * sizeof(u_int)
  // bytes; so temporarily change the pixel size when partitioning the output image
  u_int orig_pixelsize = pixelsize;
  pixelsize += (1 + pixelsize) * sizeof(u_int);

  // for simplicity, we are going to partition the image either by rows or
  // by columns, depending on whether the image is tall and skinny or short
  // and fat
  T2_UDFRet rv;
  if (width >= height)
     rv = partitionByCols(mem, tile_mbrs);
  else
     rv = partitionByRows(mem, tile_mbrs);

  // reset pixelsize
  pixelsize = orig_pixelsize;

  return rv;
}



/////////////////////////////////////////////////////////////////////
//
// svm_ImageAccMetaAVG constructor
//
// Expects the following values in binary format:
//	w, h		number of pixels along each dimension
//	psize		number of bytes per pixel
//
/////////////////////////////////////////////////////////////////////
T2_UDFRet
svmImageAccMetaAVGConstructor(const T2_QueryInfo& qinfo,
			      const T2_Box& output_box,
			      T2_UsrArgReader& arg, T2_AccMetaObj*& acmp)
{
  u_int w, h, psize;

  if (arg.getDataSize() < sizeof(w)+sizeof(h)+sizeof(psize)) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: not enough arguments for svmImageAccMetaConstructor.");
     return T2_UDFRet_ERROR;
  }

  arg >> w >> h >> psize;
  acmp = (T2_AccMetaObj *) new svm_ImageAccMetaAVG(w, h, psize);

  t2_Assert (output_box.getNumberDimensions() == 2);
  t2_Assert ((u_int) output_box.getLow()[0] == 0);
  t2_Assert ((u_int) output_box.getLow()[1] == 0);
  t2_Assert (w-1 == (u_int) output_box.getHigh()[0]);
  t2_Assert (h-1 == (u_int) output_box.getHigh()[1]);

  return T2_UDFRet_OK;
}

