
/* $Id: svm_index.C,v 1.4 2000/02/09 22:07:23 chialin Exp $ */

/* svm_index.C */


#include	<iostream.h>
#include	"svm_utility.h"
#include	"svm_index.h"


///////////////////////////////////////////////////////////////////////////
//
// app_ImageIndex methods
//
///////////////////////////////////////////////////////////////////////////
svm_ImageIndex::svm_ImageIndex(T2_Array<int> fds): fps(0), curfp(0)
{
  u_int nfiles = fds.getNumberElements();	// # of assigned index files
  fps.clear(nfiles);
  for (u_int i=0; i<nfiles; i++) {
     fps[i] = fdopen(fds[i], "r");
     if (fps[i] == NULL) {
	t2_err_ret(__FILE__, __LINE__, "ERROR: fdopen on index file fails.");
	for (u_int j=0; j<i; j++)
	   fclose(fps[j]);
        return;
     }
  }
}


T2_UDFRet
svm_ImageIndex::fetchInit(const T2_System& system, const T2_Box& qr)
{
  curfp = 0;
  fseek(fps[0], 0, SEEK_SET);		// rewind the file
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_ImageIndex::fetch(const T2_System& system, const T2_Box& qr,
		      bool& eod, T2_Box& mbr, T2_ClusterAuxInfoWriter& info,
		      T2_VArray<T2_BlockRequest>& chk)
{
  // line buffer
  const size_t max_line_size = 512;
  char line[max_line_size];

  // set number of dimensions
  mbr.setNumberDimensions(2);

  while (curfp < fps.getNumberElements()) {
    FILE* fp = fps[curfp];
    while (get_uncomment_line(line, max_line_size, fp, '#') != 0) {
      // assume that each entry in the index file has the following fields, all
      // in one line:
      // <x_pos> <y_pos> <width> <height> <# bytes per pixel> <offset> <size>
      // <logical data file id>
      // where <x_pos> <y_pos> are the coordinates for the top-left corner 
      // of a chunk,
      u_int x_pos, y_pos, width, height, pixelsize, offset, size, fid;
      sscanf(line, "%d %d %d %d %d %d %d %d",
	    &x_pos, &y_pos, &width, &height, &pixelsize, &offset, &size, &fid);

      // compute mbr
      mbr.getLow()[0] = x_pos;
      mbr.getLow()[1] = y_pos;
      mbr.getHigh()[0] = x_pos + width - 1;
      mbr.getHigh()[1] = y_pos + height - 1;

      if (qr ^ mbr) {			// data found
         // store x_pos and y_pos in info, just for demonstration purpose
         info.allocBuffer(sizeof(x_pos) + sizeof(y_pos));   // allocate buffer
         info << x_pos << y_pos;
 
         // set block-request information 
         chk << T2_BlockRequest(fid, offset, size);

         eod = false;
         return T2_UDFRet_OK;
      }
    }		// while (get_uncomment_line(..., fp) != 0)

    if (++curfp < fps.getNumberElements())
       fseek(fps[curfp], 0, SEEK_SET);		// rewind the next file

  }		// while (curfp < fps.getNumberElements())

  eod = true;
  return T2_UDFRet_OK;
}
  


///////////////////////////////////////////////////////////////////////////
//
// svm_ImageIndex constructor
//
// Ignores arg.
//
///////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageIndexConstructor(const T2_System& system, T2_UsrArgReader& arg,
			  const T2_Array<int>& fd, T2_Index*& idxp)
{
  if (fd.getNumberElements() == 0) { 	// no assigned index files
     // this should never happed 
     t2_err_quit(__FILE__, __LINE__,
		"ERROR: no assigned index files.");
     return T2_UDFRet_ERROR;
  }

  idxp = (T2_Index *) new svm_ImageIndex(fd);
  return T2_UDFRet_OK;
}
  
