
/* $Id: svm_output.h,v 1.1 2000/02/04 00:21:21 chialin Exp $ */

/* svm_output.h */

#ifndef _SVM_OUTPUT_H_
#define _SVM_OUTPUT_H_

#include	"t2_output.h"
#include	"svm_image.h"


/////////////////////////////////////////////////////////////////////////
//
// svm_OutputSubImage
//
// A class that stores an output sub-image of an output image.
// It uses a protocol to send the sub-image to a display client.
//
/////////////////////////////////////////////////////////////////////////
class svm_OutputSubImage: public T2_Output {
public:
  svm_OutputSubImage(u_int pno, u_int ptp,
		     u_int np, u_int nt, u_int xp, u_int yp, u_int w, u_int h,
		     u_char* b, u_int s, u_int v, u_int xr, u_int yr);
  ~svm_OutputSubImage() {};

  // virtual functions from T2_Output
  virtual size_t getOutputBufferSize(const T2_QueryInfo& qinfo);
  virtual T2_UDFRet flushOutput(const T2_QueryInfo& qinfo, T2_OutputBuffer& buf,
				T2_OutputDataPtrList& ptr);

protected:
  u_int packet_no,			// packet number from aggregation func
	packet_type;			// packet type from aggregation func

  u_int nprocs,				// # back-end nodes
	ntiles,				// # output tiles
	x_pos, y_pos,			// position of the sub-image in the
					// entire output image
	sub_x_res, sub_y_res;		// width and height of sub-image
  u_char* bytes;			// pointer to bytes of the sub-image
					// pixels (the bytes are owned by
					// another object, so no need to
					// delete them here)
  u_int pixelsize,			// number of bytes per pixel
	maxval;				// max value for each component byte
  u_int x_res, y_res;			// width and height of entire output
					// image

private:
};


inline 
svm_OutputSubImage::
svm_OutputSubImage(u_int pno, u_int ptp,
		   u_int np, u_int nt, u_int xp, u_int yp, u_int w, u_int h,
		   u_char* b, u_int s, u_int v, u_int xr, u_int yr)
		: packet_no(pno), packet_type(ptp),
		  nprocs(np), ntiles(nt), x_pos(xp), y_pos(yp),
		  sub_x_res(w), sub_y_res(h), bytes(b), pixelsize(s), maxval(v),
		  x_res(xr), y_res(yr)
{
}



/////////////////////////////////////////////////////////////////////////
//
// svm_OutputSubImagePPM
//
// A class that is exactly the same as svm_OutputSubImage, except that
// the output is in ppm format.
//
/////////////////////////////////////////////////////////////////////////
class svm_OutputSubImagePPM: public svm_OutputSubImage {
public:
  svm_OutputSubImagePPM(u_int pno, u_int ptp,
			u_int np, u_int nt, u_int xp, u_int yp, u_int w,u_int h,
			u_char* b, u_int s, u_int v, u_int xr, u_int yr);
  ~svm_OutputSubImagePPM() {};

  // virtual functions from T2_Output
  virtual size_t getOutputBufferSize(const T2_QueryInfo& qinfo);
  virtual T2_UDFRet flushOutput(const T2_QueryInfo& qinfo, T2_OutputBuffer& buf,
				T2_OutputDataPtrList& ptr);

protected:
private:
};


inline 
svm_OutputSubImagePPM::
svm_OutputSubImagePPM(u_int pno, u_int ptp,
		      u_int np, u_int nt, u_int xp, u_int yp, u_int w, u_int h,
		      u_char* b, u_int s, u_int v, u_int xr, u_int yr)
	: svm_OutputSubImage(pno, ptp, np, nt, xp, yp, w, h, b, s, v, xr, yr)
{
}



#endif	/* ! _SVM_OUTPUT_H_ */
