
/* $Id: svm_prj.h,v 1.2 2000/02/14 19:40:39 chialin Exp $ */

/* svm_prj.h */


#ifndef _SVM_PRJ_H_
#define _SVM_PRJ_H_

#include	"t2_prj.h"


///////////////////////////////////////////////////////////////////////
//
// svm_ImageProjection
//
///////////////////////////////////////////////////////////////////////
class svm_ImageProjection: public T2_ProjectFuncObj {
public:
  svm_ImageProjection(int xo, int yo, u_int xs, u_int ys);
  ~svm_ImageProjection() {};

  // virtual methods
  T2_UDFRet project(const T2_System& system,
		    const T2_Box& ibox, const T2_Region& rg,
                    const T2_Point& input_pt,
		    bool& succeed, T2_Region& output_rg);
  virtual u_int getNumberInputDimensions() { return 2; };
  virtual u_int getNumberOutputDimensions() { return 2; };

  // the T2 default implementation of the virtual function projectBox()
  // is overridden, 'cause this may be more efficient
  virtual T2_UDFRet projectBox(const T2_System& system, const T2_Box& inbox,
			       bool& succeed, T2_Box& outbox);

  // a help method to project a point in the input space into a point
  // in the output space without considering the bounds in the output space;
  // this method is used by both project() and projectBox()
  void projectPoint(const T2_Point& inpoint, T2_Point& outpoint) const;

  // access methods
  int getXOrigin() const { return x_orig; };
  int getYOrigin() const { return y_orig; };
  u_int getXSize() const { return x_size; };
  u_int getYSize() const { return y_size; };

protected:
  int x_orig, y_orig;	// the coordinats in the input attribute space for the
			// origin of the output attribute space
  u_int x_size, y_size;	// the number of input units that are contained by
			// an output unit along each dimension
			// (ie this is the scaling factor);
			// this representation, obviously, only supports
			// projection from a finer grid to a coarser grid

private:
};


inline
svm_ImageProjection::svm_ImageProjection(int xo, int yo, u_int xs, u_int ys)
			: x_orig(xo), y_orig(yo), x_size(xs), y_size(ys)
{
}

inline void
svm_ImageProjection::projectPoint(const T2_Point& inpoint, T2_Point& outpoint)
const
{
  assert (inpoint.getNumberDimensions() == 2);
  assert (outpoint.getNumberDimensions() == 2);

  int x = (int) inpoint[0],
      y = (int) inpoint[1];

  if (x < x_orig)
     outpoint[0] = (x - x_orig) / (int) x_size - 1;
  else
     outpoint[0] = (x - x_orig) / (int) x_size;
  if (y < y_orig)
     outpoint[1] = (y - y_orig) / (int) y_size - 1;
  else
     outpoint[1] = (y - y_orig) / (int) y_size;
}



#endif	/* _SVM_PRJ_H_ */
