
/* $Id: t2_dsinfo.h,v 1.3 2000/03/14 23:25:55 chialin Exp $ */

/* t2_dsinfo.h */

#ifndef _T2_DSINFO_H_
#define _T2_DSINFO_H_

#include	<iostream.h>
#include	<fstream.h>
#include	<string.h>
#include	<sys/types.h>
#include	<regex.h>
#include	<strings.h>
#include	"t2_header.h"
#include	"t2_txtfile.h"
#include	"t2_dsreg.h"


// forward declaration
class T2_DSInfoEntry;

/////////////////////////////////////////////////////////////////////////
//
// T2_IndexInfoEntry
//
/////////////////////////////////////////////////////////////////////////
class T2_IndexInfoEntry {
friend class T2_DSInfoEntry;
public:
  T2_IndexInfoEntry();
  ~T2_IndexInfoEntry() { delete [] idx_name; delete [] idx_description; };

  // copy constructor and operator =
  T2_IndexInfoEntry(const T2_IndexInfoEntry& e);
  T2_IndexInfoEntry& operator =(const T2_IndexInfoEntry& e);

  // const access methods
  T2_IndexID getIndexID() const { return idx_id; };
  const char* getIndexName() const { return idx_name; };
  const char* getIndexDescription() const { return idx_description; };

  // non-const access methods
  T2_IndexID& getIndexID() { return idx_id; };
  char*& getIndexName() { return idx_name; };
  char*& getIndexDescription() { return idx_description; };

  // write to ostream for debugging purpose
  ostream& write(ostream& os) const;

  // copy a string to another one
  static void str_alloc_cpy(char*& dest, const char* src);

protected:
  T2_IndexID idx_id;
  char *idx_name,
       *idx_description;

  // read from a file
  bool read(T2_TxtFile& file);

private:
};


inline void
T2_IndexInfoEntry::str_alloc_cpy(char*& dest, const char* src)
{
  delete [] dest;
  size_t len = strlen(src) + 1;
  t2_Alloc(dest, len, char);
  memcpy(dest, src, len);
}

inline
T2_IndexInfoEntry::T2_IndexInfoEntry(): idx_id(0), idx_name(NULL),
					idx_description(NULL)
{
}

inline
T2_IndexInfoEntry::T2_IndexInfoEntry(const T2_IndexInfoEntry& e)
			: idx_id(e.idx_id)
{
  if (e.idx_name == NULL)
     idx_name = NULL;
  else
     str_alloc_cpy(idx_name, e.idx_name);
  if (e.idx_description == NULL)
     idx_description = NULL;
  else
     str_alloc_cpy(idx_description, e.idx_description);
}

inline T2_IndexInfoEntry&
T2_IndexInfoEntry::operator =(const T2_IndexInfoEntry& e)
{
  idx_id = e.idx_id;
  if (e.idx_name == NULL)
     idx_name = NULL;
  else
     str_alloc_cpy(idx_name, e.idx_name);
  if (e.idx_description == NULL)
     idx_description = NULL;
  else
     str_alloc_cpy(idx_description, e.idx_description);
  return *this;
}

inline ostream&
operator << (ostream& os, const T2_IndexInfoEntry& e)
{
  return e.write(os);
}



/////////////////////////////////////////////////////////////////////////
//
// T2_DSInfoEntry
//
/////////////////////////////////////////////////////////////////////////
class T2_DSInfoEntry {
public:
  T2_DSInfoEntry();
  ~T2_DSInfoEntry();

  // query methods
  u_int getNumberIterators() const { return iternames.getNumberElements(); };
  u_int getNumberIndexes() const { return indexes.getNumberElements(); };

  // const access methods
  T2_DSID getDatasetID() const { return ds_id; };
  const char* getDatasetName() const { return dataset_name; };
  const char* getDatasetDescription() const { return dataset_description; };
  size_t getBlobFileSize() const { return blob_fsize; };
  const char* getBlobFileName() const { return blob_fname; };
  const char* getIteratorName(u_int i) const { return (char *) iternames[i]; };
  const T2_IndexInfoEntry& getIndex(u_int i) const { return indexes[i]; };

  // non-const access methods
  T2_DSID& getDatasetID() { return ds_id; };
  char*& getDatasetName() { return dataset_name; };
  char*& getDatasetDescription() { return dataset_description; };
  size_t& getBlobFileSize() { return blob_fsize; };
  char*& getBlobFileName() { return blob_fname; };
  char*& getIteratorName(u_int i) { return (char *&) iternames[i]; };
  T2_IndexInfoEntry& getIndex(u_int i) { return indexes[i]; };

  // accessing arrays
  T2_Array<T2_CharPtr>& getAllIteratorNames() { return iternames; };
  T2_Array<T2_IndexInfoEntry>& getAllIndexes() { return indexes; };

  // clear
  void clear();

  // read an entry from file
  bool read(T2_TxtFile& file);

  // write to ostream for debugging purpose
  ostream& write(ostream& os) const;

  // copy a char[] to another char[]
  static void str_alloc_cpy(char*& dest, const char* src);

protected:
  T2_DSID ds_id;			// dataset id
  char *dataset_name;			// name for dataset
  char *dataset_description;
  size_t blob_fsize;			// size of blob file
  char *blob_fname;			// filename of blob object

  T2_Array<T2_CharPtr> iternames;	// iterator names
					// (iterator id is implicitly
					// its array index in iternames[])
  T2_Array<T2_IndexInfoEntry> indexes;	// index entries

private:
};

inline
T2_DSInfoEntry::T2_DSInfoEntry()
			: ds_id(0), dataset_name(NULL),
			  dataset_description(0),
			  blob_fsize(0), blob_fname(NULL),
			  iternames(3), indexes(3)
{
}

inline
T2_DSInfoEntry::~T2_DSInfoEntry()
{
  delete [] dataset_name;
  delete [] dataset_description;
  delete [] blob_fname;
}

inline void
T2_DSInfoEntry::clear()
{
  ds_id = 0;
  delete [] dataset_name; dataset_name = NULL;
  delete [] dataset_description; dataset_description = NULL;
  delete [] blob_fname; blob_fname = NULL;
  blob_fsize = 0;
  iternames.clear(0);
  indexes.clear(0);
}

inline void
T2_DSInfoEntry::str_alloc_cpy(char*& dest, const char* src)
{
  delete [] dest;
  size_t len = strlen(src) + 1;
  t2_Alloc(dest, len, char);
  memcpy(dest, src, len);
}

inline ostream&
operator << (ostream& os, const T2_DSInfoEntry& e)
{
  return e.write(os);
}



/////////////////////////////////////////////////////////////////////////
//
// T2_DSInfoFileReader
//
/////////////////////////////////////////////////////////////////////////
class T2_DSInfoFileReader {
public:
  T2_DSInfoFileReader(const char* fname= NULL);
  ~T2_DSInfoFileReader() { file.close(); if (re_init == true) regfree(&re); };

  // open and close file
  void open(const char*);
  void close() { file.close(); ndatasets = 0; };

  // query methods
  operator const void*() const { return (const void *) file; };

  // init regular-expression search function;
  // return false if initialization fails
  bool fetchRegExpInit(const char* pattern,
		       int flag = REG_EXTENDED|REG_NOSUB|REG_ICASE);

  // get the next entry that satisfies the given constraints;
  // return false if no more match is found
  bool fetchRegExpName(T2_DSInfoEntry& e);

  // fetch entry for exact match;
  // returns true if a match is found, false otherwise;
  // since names are assumed to be unique, only the 1st one found is
  // returned;
  bool fetchExactMatch(T2_DSInfoEntry& e, const char* name);

  // return true if s is a end-of-record marker
  static bool check_eor_mark(const char* s);

protected:
  T2_TxtFile file;
  u_int ndatasets;		// number of datasets

  // for search
  regex_t re;
  bool re_init;			// has re ever been initialized?
  u_int cur;			// # entries that fetchRegExpName has read

private:
};


inline void
T2_DSInfoFileReader::open(const char* fname)
{
  close();
  if (fname == NULL)
      return;
  file.open(fname, true);
  if (!file) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open dataset info file '%s'.", fname);
     return;
  }
}

inline
T2_DSInfoFileReader::T2_DSInfoFileReader(const char* fname)
			: file('#', 2048), ndatasets(0), re_init(false)
{
  if (fname != NULL)
     open(fname);
}



/////////////////////////////////////////////////////////////////////////
//
// T2_DSInfoFileWriter
//
/////////////////////////////////////////////////////////////////////////
class T2_DSInfoFileWriter {
public:
  T2_DSInfoFileWriter(const char* fname= NULL);
  ~T2_DSInfoFileWriter() { close(); };

  // open and close file
  void open(const char*);
  void close();

  // query methods
  operator const void*() const { return (valid == true)? this:NULL; };
  u_int getNumberDatasets() const { return ndatasets; };
  u_int getNumberIndexes() const { return nindexes; };

  // add a dataset entry to the file, and assign id's to the dataset and
  // its associated indexes;
  // returns true if succeed, and set dsid to the assigned dataset id,
  // store in idx_ids the assigned index id's
  // if error, returns NULL and set eflag to 1
  bool addEntry(T2_DSRegistrationEntry& e, T2_DSID& dsid,
		T2_Array<T2_IndexID>& idx_ids, T2_Array<T2_CharPtr>& iterators);

protected:
  const u_int max_ndigits_ndatasets;	// max # digits for total # datasets
  fstream file;
  u_int ndatasets,			// # datasets added so far
	nindexes;			// # indexes added so far
  bool valid;				// is a file open? true = yes

private:
};


inline
T2_DSInfoFileWriter::T2_DSInfoFileWriter(const char* fname)
			: max_ndigits_ndatasets(20), file(),
			  ndatasets(0), nindexes(0), valid(false)
{
  if (fname != NULL)
     open(fname);
}



#endif	/* ! _T2_DSINFO_H_ */
