
/* $Id: feinquiry.C,v 1.4 2000/02/24 17:06:37 chialin Exp $ */

/* feinquiry.C */


#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	"t2_feinquiry.h"
#include	"t2_conscat.h"
#include	"t2_dsinfo.h"


// prototypes
bool
process_dataset_inquiry(T2_Socket& sock, T2_FEInquiry& inq,
			T2_DSInfoFileReader& dsinfo),
process_function_inquiry(T2_Socket& sock, T2_FEInquiry& inq,
			 T2_ConsCatalogReader& conscat);

///////////////////////////////////////////////////////////////////////////
//
// some help functions to deal with char[]
//
///////////////////////////////////////////////////////////////////////////
inline bool
bwrite_chararray(T2_OBStream& obs, const char* s)
{
  t2_say5(cout, "inside bwrite_chararray for '" << t2_stringNullWrap(s)
		<< "'" <<  endl);

  if (!obs)
     return false;

  if (s == NULL)
     obs << (size_t) 0;
  else {
     size_t len = strlen(s)+1;

     t2_say5(cout, "writing out len=" << len << endl);

     obs << len;
     if (obs.bwrite(s, len, sizeof(char)) != (ssize_t) len)
	return false;
  }

  return true;
}


inline bool
bread_chararray(T2_IBStream& ibs, char*& s)
{
  t2_say5(cout, "inside bread_chararray" << endl);

  if (!ibs)
     return false;

  t2_say5(cout, "gonna call read len" << endl);

  size_t len;
  ibs >> len;

  t2_say5(cout, "got len = " << len << endl);

  if (!ibs)
     return false;

  if (len == 0)
     s = NULL;
  else {
     t2_Alloc (s, len, char);

     t2_say5(cout, "gonna read " << len << " bytes" << endl);

     if (ibs.bread(s, len, sizeof(char)) != (ssize_t) len)
	return false;
  }

  t2_say5(cout, "bread_charray recvs '" << t2_stringNullWrap(s) << "'" << endl);

  return true;
}


///////////////////////////////////////////////////////////////////////////
//
// T2_FEInquiry methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_FEInquiry::bwrite(T2_OBStream& obs) const
{
  if (!obs)
     return false;

  obs << (u_char) qtype << (u_char) ftype;

  if (!obs || bwrite_chararray(obs, name) == false)
     return false;

  switch (exact_match) {
     case false:	obs << (u_char) 'f';
			break;
     case true:		obs << (u_char) 't';
			break;
  }
  obs << field_flag << blobsize_limit;
  if (!obs)
     return false;
  return true;
}


bool
T2_FEInquiry::bread(T2_IBStream& ibs)
{
  qtype = inquiry_error;
  delete_name();

  if (!ibs)
     return false;

  u_char qt, ft, b;
  ibs >>  qt >> ft;

  if (!ibs || bread_chararray(ibs, name) == false) {
     return false;
  }

  ibs >> b;
  switch (b) {
     case 'f':		exact_match = false;
			break;
     case 't':		exact_match = true;
			break;
     default:		t2_err_ret(__FILE__, __LINE__,
				"ERROR: invalid bool value %d", (int) b);
			return false;
  }
  ibs >> field_flag >> blobsize_limit;
  if (!ibs)
     return false;

  qtype = (inquiry_type) qt;
  ftype = (T2_UDFType) ft;
  return true;
}


ostream&
T2_FEInquiry::write(ostream& os) const
{
  switch (qtype) {
     case inquiry_error:	os << "inq type = error"; break;
     case inquiry_dataset:	os << "inq type = dataset"; break;
     case inquiry_function:	os << "inq type = function"; break;
     default:			os << "invalid inq " << (int) qtype;
  }
  os << "	ftype= " << t2_UDFType2shortstring(ftype) << endl;
  os << "name= '" << t2_stringNullWrap(name) << "'" << endl
     << "exact_match= " << (int) exact_match
     << "   field_flag= " << field_flag
     << "   blobsize_limit= " << blobsize_limit
     << "   del_name= " << (int) del_name
     << endl;
  return os;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_FEInquiryResultsWriter methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_FEInquiryResultsWriter::initiate()
{
  if (!sock)
     return false;

  // keep track of how many bytes of blob object have been sent
  blobsize_cur = 0;

  t2_say5(cout, "gonna write field_flag = " << field_flag << endl);

  // send field_flag just to make sure that the reader knows what it is 
  // expecting
  sock << field_flag;
  if (!sock)
     return false;

  return true;
}


bool
T2_FEInquiryResultsWriter::writeResult(T2_ConsCatalogEntry& e)
{
  if (!sock)
     return false;

  t2_say5(cout,
	  "gonna write out begin label " << (int) one_more_result << endl);

  sock << (u_char) one_more_result;	// beginning of another entry

  t2_say5(cout, "writing out type " << t2_UDFType2shortstring(e.type)
		<< " ascii=" << (int) e.type << endl);

  t2_Assert (e.type != T2_UDF_Unknown);
  sock << (u_char) e.type;		// always send type

  if (T2_FEInquiry::checkFunctionIDField(field_flag) == true) {
     t2_say5(cout, "writing out id " << e.id << endl);
     sock << e.id;
  }

  t2_say5(cout, "gonna check function name" << endl);

  if (T2_FEInquiry::checkFunctionNameField(field_flag) == true) {
     t2_Assert (e.name != NULL);
     t2_say5(cout, "gonna write out name= " << e.name << endl);

     if (bwrite_chararray(sock, e.name) == false)
        return false;
  }
  
  t2_say5(cout, "gonna check function description" << endl);

  if (T2_FEInquiry::checkFunctionDescriptionField(field_flag) == true) {
     t2_say5(cout, "gonna write funtion description='"
		  << t2_stringNullWrap(e.description) << "'" << endl);

     if (bwrite_chararray(sock, e.description) == false)
        return false;
  }

  if (!sock)
     return false;
  return true;
}


bool
T2_FEInquiryResultsWriter::writeResult(T2_DSInfoEntry& e)
{
  if (!sock)
     return false;

  sock << (u_char) one_more_result;	// beginning of another entry

  t2_say5(cout, "gonna check dataset id" << endl);

  if (T2_FEInquiry::checkDatasetIDField(field_flag) == true) {
     t2_say5(cout, "gonna write dataset id= " << e.getDatasetID() << endl);
     sock << e.getDatasetID();
  }

  t2_say5(cout, "gonna check dataset name" << endl);

  t2_Assert (e.getDatasetName() != NULL);
  if (T2_FEInquiry::checkDatasetNameField(field_flag) == true) {
     t2_say5(cout, "gonna write out datasetname ='" << e.getDatasetName()
		   << "'" << endl);
     if (bwrite_chararray(sock, e.getDatasetName()) == false)
	return false;
  }

  t2_say5(cout, "gonna check dataset description" << endl);

  if (T2_FEInquiry::checkDatasetDescriptionField(field_flag) == true) {
     t2_say5(cout, "gonna write dataset description='"
		<< t2_stringNullWrap(e.getDatasetDescription()) << "'" << endl);

     if (bwrite_chararray(sock, e.getDatasetDescription()) == false)
	return false;
  }

  t2_say5(cout, "gonna check blob data size" << endl);

  if (T2_FEInquiry::checkDatasetBlobDataSizeField(field_flag) == true) {
     t2_say5(cout, "gonna write blob filesize= " << e.getBlobFileSize() 
		   << endl);
     sock << e.getBlobFileSize();
  }

  t2_say5(cout, "gonna check blobdata field" << endl);

  if (T2_FEInquiry::checkDatasetBlobDataField(field_flag) == true) {
     t2_say5(cout, "gonna write blob data" << endl);

     // here is what gets sent across the socket:
     // (1) if file cannot be open, then 0 is sent across the socket to indicate
     //     that 0 bytes will be written into the socket
     // (2) if file can be open, then the following is written to the socket:
     //     (a) number of bytes expected to be written to the socket
     //     (b) the content of the file (exactly the number of bytes claimed
     //         in (a) is written to the socket)
     //     (c) the number of valid bytes out of the bytes actually written
     //         over the socket (just in case error occurs during reading 
     //         the file)
     if (e.getBlobFileName() == NULL || e.getBlobFileSize() == 0 ||
	 (blobsize_limit > 0 && blobsize_cur > blobsize_limit))
	sock << (size_t) 0;
     else if (blobsize_limit > 0 &&
	      blobsize_cur + e.getBlobFileSize() > blobsize_limit)
     {
	t2_say5(cout, "*******  CUTTING OFF ************************" << endl);

	sock << (size_t) 0;
	blobsize_cur = blobsize_limit+1;	// no more blob data
     }
     else {
	t2_say5(cout, "opening blob file '" << e.getBlobFileName() << "'"
			<< endl);

        int fd = open(e.getBlobFileName(), O_RDONLY);
	if (fd < 0) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open blob file '%s'.", e.getBlobFileName());
	   sock << (size_t) 0;
	}
	else {
	   sock << e.getBlobFileSize();
	   const size_t bufsize = 1024;
	   char buf[bufsize];
	   size_t ss;
	   for (ss=0; ss<e.getBlobFileSize();) {
	      t2_say5(cout, "ss = " << ss << endl);

	      ssize_t nbytes = read(fd, buf, bufsize);
	      if (nbytes <= 0) {		// read error
		 t2_say5(cout, "read error nbytes = " << nbytes << endl);

		 // not sure what to do now ....
		 // for now, I'll just send garbage bytes until in total
		 // e.getBlobFileSize() bytes have been sent over the
		 // the socket, and at the end send to the recv'er the actual
		 // valid bytes
		 // so far, ss bytes have been written ...
		 t2_Assert (e.getBlobFileSize() > ss);
		 size_t garbage = e.getBlobFileSize() - ss;
		 for (;garbage >= bufsize; garbage -= bufsize) {
		    ssize_t nb = sock.bwrite(buf, bufsize, sizeof(char));
		    if (nb != (ssize_t) bufsize) {	// serious write error
		       if (sock.getErrorValue() != EPIPE)
		          t2_err_ret(__FILE__, __LINE__,
			      "ERROR: socket write error after writing %d out of %d bytes from blob file '%s'",
			      e.getBlobFileSize()-garbage, e.getBlobFileSize(),
			      e.getBlobFileName());
		       close(fd);
		       return false;
		    }
		 }
		 ssize_t nb = sock.bwrite(buf, garbage, sizeof(char));
		 if (nb != (ssize_t) garbage) {		// serious write error
		    if (sock.getErrorValue() != EPIPE)
		       t2_err_ret(__FILE__, __LINE__,
			   "ERROR: socket write error after writing %d out of %d bytes from blob file '%s' to the socket.",
			   e.getBlobFileSize()-garbage, e.getBlobFileSize(),
			   e.getBlobFileName());
		    close(fd);
		    return false;
		 }
	      }
	      else {		// if (nbytes <= 0)
		 ssize_t nb = sock.bwrite(buf, nbytes, sizeof(char));
		 if (nb != nbytes) {	// serious write error
		    if (sock.getErrorValue() != EPIPE)
		       t2_err_ret(__FILE__, __LINE__,
			  "ERROR: only able to write %d out of %d bytes from blob file '%s' to the socket.",
			  ss+nb, e.getBlobFileSize(), e.getBlobFileName());
		    close(fd);
		    return false;
		 }
		 ss += nbytes;
	      }			// if (nbytes < 0) else

	   }			// for (ss=0; ss<e.getBlobFileSize();)

	   t2_say5(cout, "at the end ss = " << ss << " blobsize_cur = " 
			<< blobsize_cur << endl);
	   t2_Assert (ss <= e.getBlobFileSize());
	   sock << ss;		// this is actual number of valid bytes sent

	   blobsize_cur += ss;

	   close(fd);
        }			// if (fd < 0) else
     }				// if (e.getBlobFileName() == NULL) else
     
  }	// if (T2_FEInquiry::checkDatasetBlobDataField(field_flag) == true)

  t2_say5(cout, "gonna check number of iterators" << endl);

  if (T2_FEInquiry::checkDatasetIteratorNameField(field_flag) == true) {

     t2_say5(cout, "gonna write  number of iterators= " 
			<< e.getNumberIterators() << endl);

     sock << e.getNumberIterators();
     for (u_int k=0; k<e.getNumberIterators(); k++) {
	t2_say5(cout, "gonna write iterator= '" 
		<< t2_stringNullWrap(e.getIteratorName(k)) << "'" << endl);

	if (bwrite_chararray(sock, e.getIteratorName(k)) == false)
	   return false;
     }
  }

  t2_say5(cout, "gonna check number of index" << endl);

  if (T2_FEInquiry::checkDatasetIndexIDField(field_flag) == true ||
      T2_FEInquiry::checkDatasetIndexNameField(field_flag) == true ||
      T2_FEInquiry::checkDatasetIndexDescriptionField(field_flag) == true)
  {
      t2_say5(cout, "gonna write number indexes= " << e.getNumberIndexes() 
			<< endl);
     sock << e.getNumberIndexes();

     for (u_int k=0; k<e.getNumberIndexes(); k++) {
	t2_say5(cout, "gonna check indexid" << endl);

	T2_IndexInfoEntry& ie = e.getIndex(k);
	if (T2_FEInquiry::checkDatasetIndexIDField(field_flag) == true) {
	   t2_say5(cout, "gonna write indexid= " << ie.getIndexID() << endl);
	   sock << ie.getIndexID();
	}

	t2_say5(cout, "gonna check indexname" << endl);

	if (T2_FEInquiry::checkDatasetIndexNameField(field_flag) == true) {
	   t2_say5(cout, "gonna write indename='" 
			<< t2_stringNullWrap(ie.getIndexName()) << "'" << endl);

	   if (bwrite_chararray(sock, ie.getIndexName()) == false)
	      return false;

	}

	t2_say5(cout, "gona check index description" << endl);

	if (T2_FEInquiry::checkDatasetIndexDescriptionField(field_flag) == true)
	{
	   t2_say5(cout, "gonna writ3e index description='" 
			<< t2_stringNullWrap(ie.getIndexDescription()) 
			<< "'" << endl);

	   if (bwrite_chararray(sock, ie.getIndexDescription()) == false)
	      return false;


	}

	t2_say5(cout, "move to next index" << endl);
     }
  }

  return true;
}


bool
T2_FEInquiryResultsWriter::terminate(result_tag t)
{
  if (!sock)
     return false;

  sock << (u_char) t;			// write tag
  if (!sock)
     return false;
  return true;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_FEInquiryResultsReader methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_FEInquiryResultsReader::initiate()
{
  if (!sock)
     return false;

  // recv field_flag just to make sure that the reader knows what it is 
  // expecting
  sock >> field_flag;

  t2_say5(cout, "recv field_flag = " << field_flag << endl);

  if (!sock)
     return false;
  return true;
}


bool
T2_FEInquiryResultsReader::readResults(T2_FEFunctionInquiryResults& r)
{
  r.getStatus() = T2_FEInquiry::bad_socket_error;

  if (!sock)
     return false;

  while (1) {
     t2_say5(cout, "gonna read begin label" << endl);

     u_char t;
     sock >> t;

     t2_say5(cout, "just read begin label = " << t << " ascii=" 
			<< (int) t << endl);

     if (!sock) {
	r.getStatus() = T2_FEInquiry::bad_socket_error;
	t2_err_ret(__FILE__, __LINE__, "ERROR: cannot read from socket");
	return false;
     }

     switch ((T2_FEInquiryResultsWriter::result_tag) t) {
	case T2_FEInquiryResultsWriter::no_more_results:
		// end of transmission
		r.getStatus() = T2_FEInquiry::results_ok;
		return true;

	case T2_FEInquiryResultsWriter::one_more_result:
		// an entry coming, go read it
		break;

	case T2_FEInquiryResultsWriter::invalid_inquiry:
		// invalid inquiry
		r.getStatus() = T2_FEInquiry::invalid_inquiry;
		return false;

	case T2_FEInquiryResultsWriter::t2_frontend_not_ready:
		// some thing is wrong with the front-end
		r.getStatus() = T2_FEInquiry::t2_frontend_not_ready;
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: T2 front-end returns error");
		return false;

	default:	// invalid label
		r.getStatus() = T2_FEInquiry::t2_frontend_not_ready;
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: invalid label %d for inquiry results",
			(int) t);
		return false;
     }			// switch (t)

    t2_say5(cout, "gonna read type" << endl);

     T2_FEFunctionEntry& e = r.appendElement();
     sock >> t;				// always recv type

     t2_say5(cout, "got type = " << t << " asci=" << (int) t << endl);

     e.type = (T2_UDFType) t;
     t2_Assert (e.type != T2_UDF_Unknown);

     t2_say5(cout, "this type is actually " << t2_UDFType2shortstring(e.type) 
			<< endl);

     if (T2_FEInquiry::checkFunctionIDField(field_flag) == true) {
	t2_say5(cout, "gona read id" << endl);

        sock >> e.id;

	t2_say5(cout, "id = " << e.id << endl);
     }

     t2_say5(cout, "gonna chekc funtion name" << endl);

     if (T2_FEInquiry::checkFunctionNameField(field_flag) == true) {
	t2_say5(cout, "gonna get e.name" << endl);

	if (bread_chararray(sock, e.name) == false)
	   return false;

	t2_say5(cout, "got e.name = '" << t2_stringNullWrap(e.name) 
			<< "'" << endl);
     }

     t2_say5(cout, "gonna check function descipriton" << endl);

     if (T2_FEInquiry::checkFunctionDescriptionField(field_flag) == true) {
	t2_say5(cout, "gonna read function descipriton" << endl);

	if (bread_chararray(sock, e.description) == false)
	   return false;

	t2_say5(cout, "got function descipriton = '" 
			<< t2_stringNullWrap(e.description) << "'" << endl);
     }

     t2_say5(cout, "finish reading one function result" << endl << e << endl);

  }			// while (1)

  // should never reach this point
  t2_err_quit(__FILE__, __LINE__,
		"ERROR: program should never reach this point");
  return false;
}


bool
T2_FEInquiryResultsReader::readResults(T2_FEDatasetInquiryResults& r)
{
  r.getStatus() = T2_FEInquiry::bad_socket_error;

  if (!sock)
     return false;

  t2_say5(cout, "gonna read begin lable" << endl);

  while (1) {
     u_char t;
     sock >> t;

     t2_say(cout, "just read begin label = " << t << " ascii=" << (int) t 
			<< endl);

     if (!sock) {
	t2_err_ret(__FILE__, __LINE__, "ERROR: cannot read from socket");
	return false;
     }

     switch ((T2_FEInquiryResultsWriter::result_tag) t) {
	case T2_FEInquiryResultsWriter::no_more_results:
		// end of transmission
		r.getStatus() = T2_FEInquiry::results_ok;
		return true;

	case T2_FEInquiryResultsWriter::one_more_result:
		// an entry coming, go read it
		break;

	case T2_FEInquiryResultsWriter::invalid_inquiry:
		// invalid inquiry
		r.getStatus() = T2_FEInquiry::invalid_inquiry;
		return false;

	case T2_FEInquiryResultsWriter::t2_frontend_not_ready:
		// some thing is wrong with the front-end
		r.getStatus() = T2_FEInquiry::t2_frontend_not_ready;
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: T2 front-end returns error");
		return false;

	default:	// invalid label
		r.getStatus() = T2_FEInquiry::t2_frontend_not_ready;
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: invalid label %d for inquiry results",
			(int) t);
		return false;

     }			// switch (t)

     T2_FEDatasetEntry& e = r.appendElement();

     if (T2_FEInquiry::checkDatasetIDField(field_flag) == true)
	sock >> e.getDatasetID();

     t2_say5(cout, "gonna check data name" << endl);

     if (T2_FEInquiry::checkDatasetNameField(field_flag) == true) {
	t2_say5(cout, "gonna read data name" << endl);

	if (bread_chararray(sock, e.getDatasetName()) == false)
	   return false;

	t2_say5(cout, "just read name = " << e.getDatasetName() << endl);
     }

     t2_say5(cout, "gonna check dataset description" << endl);

     if (T2_FEInquiry::checkDatasetDescriptionField(field_flag) == true) {
	t2_say5(cout, "gonna read dataset description" << endl);

	if (bread_chararray(sock, e.getDatasetDescription()) == false)
	   return false;

	t2_say5(cout, "just read dataset description='"
			<< t2_stringNullWrap(e.getDatasetDescription()) 
			<<  "'" << endl);
     }

     t2_say5(cout, "gonna check blobk size" << endl);

     if (T2_FEInquiry::checkDatasetBlobDataSizeField(field_flag) == true) {
	t2_say5(cout, "gonna read blob size" << endl);

	sock >> e.getBlobFileSize();

	t2_say5(cout, "just read blob size= " << e.getBlobFileSize() << endl);
     }

     t2_say5(cout, "gonna check blob data" << endl);

     if (T2_FEInquiry::checkDatasetBlobDataField(field_flag) == true) {
	t2_say5(cout, "gonna read blob data" << endl);

	size_t nbytes_expected;
	sock >> nbytes_expected;

	t2_say5(cout, "nbytes_expected = " << nbytes_expected << endl);

	if (nbytes_expected > 0) {
	   char*& data = e.getBlobObject();
	   t2_Alloc (data, nbytes_expected+1, char);
	   const size_t bufsize = 1024;
	   char* p = data;
	   size_t ss = nbytes_expected;
	   for (; ss>=bufsize;) {
	      ssize_t nread = sock.bread(p, bufsize, sizeof(char));
	      if (nread <= 0) {		// read error
		 // not sure what to do now, return false for now
		 delete [] data;
		 data = NULL;
		 return false;
	      }
	      ss -= nread;
	      p += nread;
	   }
	   ssize_t nread = sock.bread(p, ss, sizeof(char));
	   if (nread <= 0) {		// read error
	      // not sure what to do now, return false for now
	      t2_err_ret(__FILE__, __LINE__, "ERROR: socket read error.");
	      delete [] data;
	      data = NULL;
	      return false;
	   }
	   size_t nbytes_valid;
	   sock >> nbytes_valid;
	   t2_Assert (nbytes_valid <= nbytes_expected);
	   data[nbytes_valid] = '\0';
	}				// if (nbytes_expected > 0)
	else
	   t2_Assert (e.getBlobObject() == NULL);

     }

     t2_say5(cout, "gonna check any iterator" << endl);

     if (T2_FEInquiry::checkDatasetIteratorNameField(field_flag) == true) {
	t2_say5(cout, "gonna read any iterator" << endl);

	u_int niters = 0;
	sock >> niters;

	t2_say5(cout, "just got niterator = " << niters << endl);

	T2_Array<T2_CharPtr>& iternames = e.getAllIteratorNames();
	iternames.clear(niters);
	for (u_int k=0; k<niters; k++) {
	   if (bread_chararray(sock, iternames[k]) == false)
	      return false;

	   t2_say5(cout, "just got iterator= '" 
				<< t2_stringNullWrap(iternames[k]) << endl);
	}
     }

     t2_say5(cout, "gonna check any index field" << endl);

     if (T2_FEInquiry::checkDatasetIndexIDField(field_flag) == true ||
         T2_FEInquiry::checkDatasetIndexNameField(field_flag) == true ||
         T2_FEInquiry::checkDatasetIndexDescriptionField(field_flag) == true)
     {
	t2_say5(cout, "gonna read nindexes " << endl);

	u_int nindexes = 0;
	sock >> nindexes;

	t2_say5(cout, "just read nindexes= " << nindexes << endl);

	T2_Array<T2_IndexInfoEntry>& allindexes = e.getAllIndexes();
	allindexes.clear(nindexes);
	for (u_int k=0; k<nindexes; k++) {
	   t2_say5(cout, "gonna check index id" << endl);

	   T2_IndexInfoEntry& ie = allindexes[k];
	   if (T2_FEInquiry::checkDatasetIndexIDField(field_flag) == true) {
	      t2_say5(cout, "gonna read index id" << endl);

	      sock >> ie.getIndexID();

	      t2_say5(cout, "just read index id = " << ie.getIndexID() << endl);
	   }

	   t2_say5(cout, "gonna check index name" << endl);

	   if (T2_FEInquiry::checkDatasetIndexNameField(field_flag) == true) {
	      t2_say5(cout, "gonna read index name" << endl);

	      if (bread_chararray(sock, ie.getIndexName()) == false)
	         return false;

	      t2_say5(cout, "just read index name = '" 
			<< t2_stringNullWrap(ie.getIndexName()) << "'" << endl);
	   }

	   t2_say5(cout, "gonna check index description" << endl);

	   if (T2_FEInquiry::checkDatasetIndexDescriptionField(field_flag)
	       == true)
	   { 
	      t2_say5(cout, "gonna rea index description" << endl);

	      if (bread_chararray(sock, ie.getIndexDescription()) == false)
	         return false;

	      t2_say5(cout, "just read index description ='" 
			<< t2_stringNullWrap(ie.getIndexDescription()) 
			<< "'" << endl);
	    }

        }
     }

  }					// while (1)

  // should never reach this point
  t2_err_quit(__FILE__, __LINE__,
		"ERROR: program should never reach this point");
  return false;
}



///////////////////////////////////////////////////////////////////////////
//
// some functions to handle the case when error occurs
//
///////////////////////////////////////////////////////////////////////////
void
ack_frontend_notready(T2_Socket& sock)
{
  // there is something wrong with the T2 front-end, so it cannot answer
  // the inquiry right now
  if (!sock)				// not much to do when sock is err
     return;

  T2_FEInquiryResultsWriter res_writer(sock, 0, 0);
  if (res_writer.initiate() == false) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to initiate result writes.");
     return;
  }
  // terminate result exchange
  if (res_writer.terminate(T2_FEInquiryResultsWriter::t2_frontend_not_ready)
      == false)
  {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to terminate result writes.");
     return;
  }
}


void
ack_invalid_inquiry(T2_Socket& sock)
{
  // there is something wrong with the inquiry
  if (!sock)				// not much to do when sock is err
     return;

  T2_FEInquiryResultsWriter res_writer(sock, 0, 0);
  if (res_writer.initiate() == false) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to initiate result writes.");
     return;
  }
  // terminate result exchange
  if (res_writer.terminate(T2_FEInquiryResultsWriter::invalid_inquiry)
      == false)
  {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to terminate result writes.");
     return;
  }
}


///////////////////////////////////////////////////////////////////////////
//
// process_inquiry
//
// a function that processes an inquiry for the T2 front-end
//
///////////////////////////////////////////////////////////////////////////
bool 
process_inquiry(T2_Socket& sock, T2_ConsCatalogReader& conscat,
		T2_DSInfoFileReader& dsinfo)
{
  T2_FEInquiry inq;
  if (inq.bread(sock) == false) {			// read failed
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to read inquiry from socket");
     return false;
  }

  t2_say5(cout, "inquiry received:" << endl << inq << endl);

  switch (inq.getInquiryType()) {
     case T2_FEInquiry::inquiry_dataset:	// inquiring datasets
		if (!dsinfo) 			// cannot lookup datasets
		   ack_frontend_notready(sock);
		else 
		   if (process_dataset_inquiry(sock, inq, dsinfo) == false)
		      return false;
		break;

     case T2_FEInquiry::inquiry_function:	// inquiring cons functions
		if (!conscat) 			// cannot lookup datasets
		   ack_frontend_notready(sock);
		else 
		   if (process_function_inquiry(sock, inq, conscat) == false)
		      return false;
		break;

     case T2_FEInquiry::inquiry_error:
		ack_invalid_inquiry(sock);
		return false;
     default:
		t2_err_quit(__FILE__, __LINE__,
			"ERROR: invalid inquiry type %d",
			(int) inq.getInquiryType());

  }			// switch (inq.getInquiryType())

  t2_say5(cout, "gonna return from process_inquiry" << endl);

  return true;
}


// return false only when write fails
// in cases due to other type of errors, take care of it, abort the writes,
// and return true
bool
process_dataset_inquiry(T2_Socket& sock, T2_FEInquiry& inq,
			T2_DSInfoFileReader& dsinfo)
{
  // these should be checked before calling this function
  t2_Assert ((const void*) sock != NULL);
  t2_Assert ((const void*) dsinfo != NULL);
  t2_Assert (inq.getInquiryType() == T2_FEInquiry::inquiry_dataset);

  T2_FEInquiryResultsWriter res_writer(sock, inq.getFieldFlag(),
				       inq.getTotalBlobDataSizeLimit());

  // initiate result exchange
  if (res_writer.initiate() == false) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to initiate result writes.");
     return false;
  }

  // search and send results
  T2_DSInfoEntry e;
  if (inq.getExactMatchFlag() == true) {	// only exact match
     t2_say5(cout, "gonna call dsinfo.fetchExactMatch" << endl);

     if (dsinfo.fetchExactMatch(e, inq.getName()) == true) {
	t2_say5(cout, "gonna call writeResult" << endl);

	if (res_writer.writeResult(e) == false) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to write dataset info inquiry result");
	   return false;
	}
     }
     else {
	t2_say5(cout, "fetchExactMatch returns false" << endl);
     }
  }
  else {					// regular expression matching
     t2_say(cout, "gonna call dsinfo.fetchRegExpInit" << endl);

     if (dsinfo.fetchRegExpInit(inq.getName()) == false) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to initiate regular expression fetch");
	res_writer.terminate(T2_FEInquiryResultsWriter::invalid_inquiry);
	return true;
     }

     t2_say5(cout, "gonna call dsinfo.fetchRegExpName" << endl);

     T2_DSInfoEntry e;
     while (dsinfo.fetchRegExpName(e) == true) {
	t2_say5(cout, "find entry " << endl << e << endl);

	if (res_writer.writeResult(e) == false) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to write dataset info inquiry result");
	   return false;
	}
     }

     t2_say5(cout, "all entries have been sent to socket" << endl);
  }

  t2_say5(cout, "gonna call result_writer.terminate" << endl);

  // terminate result exchange
  if (res_writer.terminate(T2_FEInquiryResultsWriter::no_more_results)
      == false)
  {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to terminate result writes.");
     return false;
  }

  t2_say5(cout, "gonna return from process_dataset_inquiry" << endl);
  return true;
}


// return false only when write fails
// in cases due to other type of errors, take care of it, abort the writes,
// and return true
bool
process_function_inquiry(T2_Socket& sock, T2_FEInquiry& inq,
			 T2_ConsCatalogReader& conscat)
{
  t2_say5(cout, "calling process_function_inquiry" << endl);

  // these should be checked before calling this function
  t2_Assert ((const void*) sock != NULL);
  t2_Assert ((const void*) conscat != NULL);
  t2_Assert (inq.getInquiryType() == T2_FEInquiry::inquiry_function);

  T2_FEInquiryResultsWriter res_writer(sock, inq.getFieldFlag(), 0);

  // initiate result exchange
  if (res_writer.initiate() == false) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to initiate inquiry result writes.");
     return false;
  }

  // search and send results
  T2_ConsCatalogEntry e;
  if (inq.getExactMatchFlag() == true) {	// only exact match
     t2_say5(cout, "gonna call conscat.fetchExactMatch" << endl);

     if (conscat.fetchExactMatch(e, inq.getName(), inq.getFunctionType())
	 == true)
     {
	if (e.type != T2_UDF_Dataset && e.type != T2_UDF_Index) {
	   // only return results that are not dataset/index constructors

	   t2_say5(cout, "gonna call writeResult" << endl);

	   if (res_writer.writeResult(e) == false) {
	      t2_err_ret(__FILE__, __LINE__,
		  "ERROR: fail to write constructor function inquiry results");
	      return false;
	   }
        }
     }
     else {
	t2_say5(cout, "fetchExactMatch returns false" << endl);
     }
  }
  else {					// regular expression matching
     if (conscat.fetchRegExpInit(inq.getName(), inq.getFunctionType()) == false)
     {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to initiate regular expression fetch");
	res_writer.terminate(T2_FEInquiryResultsWriter::invalid_inquiry);
	return true;
     }

     T2_ConsCatalogEntry e;
     while (conscat.fetchRegExpName(e) == true) {
	if (e.type == T2_UDF_Dataset || e.type == T2_UDF_Index)
	   continue;		// skip dataset/index constructors

	if (res_writer.writeResult(e) == false) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to write constructor function inquiry result");
	   return false;
	}
     }
  }

  t2_say5(cout, "gonna call result_writer.terminate" << endl);

  // terminate result exchange
  if (res_writer.terminate(T2_FEInquiryResultsWriter::no_more_results) == false)
  {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to terminate result writes.");
     return false;
  }

  t2_say5(cout, "gonna return from process_function_inquiry" << endl);
  return true;
}

