/*
  File name: t2dld_hilbert.h

  Data placement (declustering/clustering) using hilbert curve.

  $Id: t2dld_hilbert.h,v 1.1 2000/02/23 16:56:54 chialin Exp $
*/
#ifndef T2DLD_HILBERT_H
#define T2DLD_HILBERT_H

#include <stdlib.h>
#include "hilbert.h"
#include "t2dld_placement.h"

#define  MAX_DIMENSIONS_ALLOWED     4

class T2DLD_Hilbert3D : public T2DLD_PlacementFuncObj {
public:
  T2DLD_Hilbert3D() {};
  virtual ~T2DLD_Hilbert3D() {};
  
  /*
    initialization;
    argument ndims is the number of dimensions for all bounding boxes,
    argument ndisks is the number of disks available for declustering;
    return true if succeed, false otherwise
  */
  bool init(const u_int ndims, const u_int ndisks) { 
    nDims  = ndims; 
    if (nDims>MAX_DIMENSIONS_ALLOWED) {
      fprintf(stderr, "Warning: Hilbert curve works upto %d dimensions.\n",
	      MAX_DIMENSIONS_ALLOWED);
      fprintf(stderr, "Will use the first three attributes of the points.\n");
      fflush(stderr);
      nDims = MAX_DIMENSIONS_ALLOWED;
    }
    nDisks = ndisks; 
    return true; 
  }

  /*
    perform data declustering/clustering;
    argument mover_instr is an array of ndisks T2DLD_MoverInstruction objects;
    return true if succeed, false otherwise
  */
  bool placeClusters(const T2DLD_DataClusterArray& tclusters,
		     T2_Array<T2DLD_MoverInstruction>& mover_instr);

protected:
private:
  u_int nDims;
  u_int nDisks; 
};



#endif /* T2DLD_HILBERT_H */
