
/* $Id: t2dld_stdfunc.h,v 1.5 1999/03/30 21:45:56 chialin Exp $ */

/* t2dld_stdfunc.h */

#ifndef _T2DLD_STDFUNC_H_
#define _T2DLD_STDFUNC_H_


#include	"t2dld_header.h"
#include	"t2dld_getclusters.h"
#include	"t2dld_placement.h"
#include	"t2dld_index.h"

////////////////////////////////////////////////////////////////////////////
//
// T2DLD_STDGetClustersFunc
//
// A standard T2 getClusters function object, which uses T2DLD_STDIndexFile
// to access index files.
//
////////////////////////////////////////////////////////////////////////////
class T2DLD_STDGetClustersFunc: public T2DLD_GetClustersFuncObj {
public:
  T2DLD_STDGetClustersFunc(): index(), read_index(false) {};
  ~T2DLD_STDGetClustersFunc() {};

  // initialization;
  // argument index_file is the file that stores information
  // for all data clusters in the data subset;
  // return true if succeeds, false otherwise
  virtual bool init(const char* index_file);

  // get number of dimensions for bounding boxes
  virtual u_int getNumberDimensions();

  // get a set of data clusters
  virtual void getClusters(T2DLD_TapeClusterArray& c);

protected:
  T2DLD_STDIndexFile index;
  bool read_index;		// have read index file already?

private:
};


/////////////////////////////////////////////////////////////////////////////
//
// constructor for standard T2 getClusters function object
//
/////////////////////////////////////////////////////////////////////////////
T2DLD_GetClustersFuncObj* t2dld_stdGetClustersFunc();



/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_RoundRobinPlacement
//
// A simple implementation of a round-robin algorithm that simply
// assigns the clusters in a round-robin fashion.
//
/////////////////////////////////////////////////////////////////////////////
class T2DLD_RoundRobinPlacement: public T2DLD_PlacementFuncObj {
public:
  T2DLD_RoundRobinPlacement(): T2DLD_PlacementFuncObj() {};
  ~T2DLD_RoundRobinPlacement() {};

  // initialization;
  // argument ndims is the number of dimensions for all bounding boxes,
  // argument ndisks is the number of disks available for declustering;
  // return true if succeed, false otherwise
  virtual bool init(const u_int ndims, const u_int ndisks) { return true; };

  // perform data declustering/clustering;
  // argument mover_instr is an array of ndisks T2DLD_MoverInstruction objects;
  // return true if succeed, false otherwise
  virtual bool placeClusters(const T2DLD_DataClusterArray& tclusters,
                             T2_Array<T2DLD_MoverInstruction>& mover_instr);

protected:
private:
};



/////////////////////////////////////////////////////////////////////////////
//
// constructor functions for T2 standard placement function objects.
//
/////////////////////////////////////////////////////////////////////////////
T2DLD_PlacementFuncObj* t2dld_stdPlacementFunc();



#endif	/* ! _T2DLD_STDFUNC_H_ */
