/*
  File name: mc_offbandcomm.H

  Contains code to perform off-band communication in Meta-Chaos.

  $Id: mc_offbandcomm.H,v 1.2 1998/09/01 11:56:08 kurc Exp $
*/
#ifndef MC_OFFBANDCOMM_H
#define MC_OFFBANDCOMM_H

#include "move.h"
#include "mc_system.H"
#include "data_types.H"

class MC_OffBandComm {
private:
public:
  MC_OffBandComm() { };
  ~MC_OffBandComm() { };

  int bcast(MC_Client& to, void *data, int size, DATA_Type type, int tag);
  int recv(MC_Client& from, void *data, int size, DATA_Type type, int tag);
  
  int shortBcast(MC_Client& to, short *data, int nelems, int tag) {
    return shortBcastOffBand(to.program(),(short*)data,nelems,tag);
  };
  int intBcast(MC_Client& to, int *data, int nelems, int tag) {
    return intBcastOffBand(to.program(),(int*)data,nelems,tag);
  }; 
  int floatBcast(MC_Client& to, float *data, int nelems, int tag) {
    return floatBcastOffBand(to.program(),(float*)data,nelems,tag);
  };
  int doubleBcast(MC_Client& to, double *data, int nelems, int tag) {
    return doubleBcastOffBand(to.program(),(double*)data,nelems,tag);
  };
  int stringBcast(MC_Client& to, char *data, int nelems, int tag) {
    return stringBcastOffBand(to.program(),(char*)data,nelems,tag);
  };
  int charBcast(MC_Client& to, char *data, int nelems, int tag) {
    return byteBcastOffBand(to.program(),(char*)data,nelems,tag);
  };

  int shortRecv(MC_Client& from, short *data, int nelems, int tag) {
    return shortRecvOffBand(from.program(),(short*)data,nelems,tag);
  };
  int intRecv(MC_Client& from, int *data, int nelems, int tag) {
    return intRecvOffBand(from.program(),(int*)data,nelems,tag);
  };
  int floatRecv(MC_Client& from, float *data, int nelems, int tag) {
    return floatRecvOffBand(from.program(),(float*)data,nelems,tag);
  };
  int doubleRecv(MC_Client& from, double *data, int nelems, int tag) {
    return doubleRecvOffBand(from.program(),(double*)data,nelems,tag);
  };
  int stringRecv(MC_Client& from, char *data, int nelems, int tag) {
    return stringRecvOffBand(from.program(),(char*)data,nelems,tag);
  };
  int charRecv(MC_Client& from, char *data, int nelems, int tag) {
    return byteRecvOffBand(from.program(),(char*)data,nelems,tag);
  };

  // If we don't know about the other program then we can 
  // use the following recv operations. 
  int recv(void *data, int size, DATA_Type type, int tag);
  int shortRecv(short *data, int nelems, int tag) {
    return shortRecvOffBand(NULL,(short*)data,nelems,tag);
  };
  int intRecv(int *data, int nelems, int tag) {
    return intRecvOffBand(NULL,(int*)data,nelems,tag);
  };
  int floatRecv(float *data, int nelems, int tag) {
    return floatRecvOffBand(NULL,(float*)data,nelems,tag);
  };
  int doubleRecv(double *data, int nelems, int tag) {
    return doubleRecvOffBand(NULL,(double*)data,nelems,tag);
  };
  int stringRecv(char *data, int nelems, int tag) {
    return stringRecvOffBand(NULL,(char*)data,nelems,tag);
  };
  int charRecv(char *data, int nelems, int tag) {
    return byteRecvOffBand(NULL,(char*)data,nelems,tag);
  };
};

#endif /* MC_OFFBANDCOMM_H */
