/*
  File name: mc_options.H

  Options class for schedule computation.
*/
#ifndef MC_OPTIONS_H
#define MC_OPTIONS_H

#include "compute_schedule.h"
#include "responsable.h"
#include "mc_system.H"
#include "mc_otherdata.H"

/*
 * Class for setting and getting the options for computing the 
 * Meta-Chaos schedule. 
 */
class MC_Options {
private:
  WhereToCompute whr;
  T_OptimiseComm commOpt;
public:
  MC_Options() {
    whr     = ComputeOnReceiver;
    commOpt = DoNotOptComm;
  };

  MC_Options(WhereToCompute w, T_OptimiseComm c) { whr = w; commOpt = c; };
  
  MC_Options(MC_Options& opt) { whr = opt.whr; commOpt = opt.commOpt; };

  virtual ~MC_Options() { };

  WhereToCompute  getWhereTo() { return whr; };
  void            setWhereTo(WhereToCompute w) { whr = w; };
  T_OptimiseComm  getCommOpt() { return commOpt; };
  void            setCommOpt(T_OptimiseComm c) { commOpt = c; };

  // A default implementation is provided 
  virtual int  distributeWork(MC_Client&           otherProg,
			      MC_OtherProgData&    dataOther,
			      int                  zmGlobalOffset,
			      int                  zmSize,
			      int                  *outZMSize) {
    int ttsize = dataOther.getNumDataElems();
    return DistributeRegularly(otherProg.program(),(void*)&ttsize,
			       zmGlobalOffset,zmSize,outZMSize);
  };
};

#endif /*  MC_OPTIONS_H */
