/*
  File name: mc_schedule.H

  Class definitions for computation of schedule and exchange of data 
  between two programs using this schedule.
*/
#ifndef MC_SCHEDULE_H
#define MC_SCHEDULE_H

#include "zm.h"
#include "comm.h"
#include "responsable.h"
#include "communication.h"
#include "optimization.h"
#include "requete.h"
#include "mc_assert.H"
#include "mc_wrapper.h"
#include "mc_options.H"
#include "mc_data.H"
#include "mc_system.H"
#include "data_types.H"

// Class for schedule computation and communication 
class MC_Schedule {
private:
  int     which_schedule;
  sched_s *sched_send;
  sched_s *sched_recv;
  pgme_s  *otherProg;

  void    emissionRequete(liste_s*             liste_of_zm,
			  MC_SetOfRegions&     setOfRegion,
			  MC_Client&           clnt,
			  MC_DataSender&       mcu,
			  MC_Options&          opt,
			  MC_OtherProgData&    knownAboutOtherPgmeData);

  void   traitementRequete(MC_Liste&          mc_liste,
			   MC_SetOfRegions&   setOfRegion,
			   MC_Server&         srvr,
			   MC_Client&         clnt,
			   MC_DataReceiver&   mcu,
			   T_OptimiseComm     optComm);

  void   traitementLocale(MC_Liste&           mc_liste,
			  MC_SetOfRegions&    setOfRegion,
			  MC_Server&          srvr,
			  MC_DataSame&        mcu,
			  sched_s             **sched_send,
			  sched_s             **sched_recv);
public:
  MC_Schedule() {
    which_schedule  = 0;
    sched_send = sched_recv = NULL;
    otherProg  = NULL;
  };

  MC_Schedule(MC_DataSender& mcu, MC_Server& srvr, MC_Client& clnt,
	      MC_Options& opt, MC_OtherProgData& kopd) {
    which_schedule  = 0;
    sched_send = sched_recv = NULL;
    otherProg  = NULL;
    computeSchedule(mcu,srvr,clnt,opt,kopd);
  };

  MC_Schedule(MC_DataReceiver& mcu, MC_Server& srvr, MC_Client& clnt,
	      MC_Options& opt, MC_OtherProgData& kopd) {
    which_schedule  = 0;
    sched_send = sched_recv = NULL;
    otherProg  = NULL;
    computeSchedule(mcu,srvr,clnt,opt,kopd);
  };

  MC_Schedule(MC_DataSame& mcu, MC_Server& srvr,
	      MC_Options& opt, MC_OtherProgData& kopd) {
    which_schedule  = 0;
    sched_send = sched_recv = NULL;
    otherProg  = NULL;
    computeSchedule(mcu,srvr,opt,kopd);
  };

  ~MC_Schedule() {
    if (which_schedule!=0)
      FreeSched(sched_send);
    if (which_schedule==3)
      FreeSched(sched_recv);
  };

  void computeSchedule(MC_DataSender& mcu, MC_Server& srvr, MC_Client& clnt,
		       MC_Options& opt, MC_OtherProgData& kopd);
  void computeSchedule(MC_DataReceiver& mcu, MC_Server& srvr, MC_Client& clnt,
		       MC_Options& opt, MC_OtherProgData& kopd);
  void computeSchedule(MC_DataSame& mcu, MC_Server& srvr,
		       MC_Options& opt, MC_OtherProgData& kopd);

  int sendData(void *data, DATA_Type type, int msgtag);
  int recvData(void *data, DATA_Type type, int msgtag);

  int sendDataInt(void *data, int msgtag);
  int sendDataFlt(void *data, int msgtag);
  int sendDataDbl(void *data, int msgtag);
  int sendDataSht(void *data, int msgtag);

  int recvDataInt(void *data, int msgtag);
  int recvDataFlt(void *data, int msgtag);
  int recvDataDbl(void *data, int msgtag);
  int recvDataSht(void *data, int msgtag);
};

inline 
int MC_Schedule::sendDataInt(void *data, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_send!=NULL));
  iDataMoveSend(otherProg,sched_send,(int*)data,msgtag);
  return 1;
}

inline 
int MC_Schedule::sendDataFlt(void *data, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_send!=NULL));
  fDataMoveSend(otherProg,sched_send,(float*)data,msgtag);
  return 1;
}

inline 
int MC_Schedule::sendDataDbl(void *data, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_send!=NULL));
  dDataMoveSend(otherProg,sched_send,(double*)data,msgtag);
  return 1;
}

inline 
int MC_Schedule::sendDataSht(void *data, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_send!=NULL));
  sDataMoveSend(otherProg,sched_send,(short*)data,msgtag);
  return 1;
}

inline 
int MC_Schedule::recvDataInt(void *data, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_recv!=NULL));
  iDataMoveRecv(otherProg,sched_recv,(int*)data,msgtag);
  return 1;
}

inline 
int MC_Schedule::recvDataFlt(void *data, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_recv!=NULL));
  fDataMoveRecv(otherProg,sched_recv,(float*)data,msgtag);
  return 1;
}

inline 
int MC_Schedule::recvDataDbl(void *data, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_recv!=NULL));
  dDataMoveRecv(otherProg,sched_recv,(double*)data,msgtag);
  return 1;
}

inline 
int MC_Schedule::recvDataSht(void *data, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_recv!=NULL));
  sDataMoveRecv(otherProg,sched_recv,(short*)data,msgtag);
  return 1;
}

#endif /* MC_SCHEDULE_H */

