/*
  File name: mc_schedule.C

  Contains code for schedule computation.
*/
#ifndef MC_SCHEDULE_C
#define MC_SCHEDULE_C

#include "mc_schedule.H"

int
MC_Schedule::sendData(void *data, DATA_Type type, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_send!=NULL));

  switch (type) {
  case DATA_FLOAT:
    fDataMoveSend(otherProg,sched_send,(float*)data,msgtag);
    return 1;
    break;
  case DATA_INT:
    iDataMoveSend(otherProg,sched_send,(int*)data,msgtag);
    return 1;
    break;
  case DATA_DOUBLE:
    dDataMoveSend(otherProg,sched_send,(double*)data,msgtag);
    return 1;
    break;
  case DATA_SHORT:
    sDataMoveSend(otherProg,sched_send,(short*)data,msgtag);
    return 1;
    break;
  default:
    return 0;
    break;
  }
  return 0;
}

int
MC_Schedule::recvData(void *data, DATA_Type type, int msgtag)
{
  MC_ASSERT((otherProg!=NULL));
  MC_ASSERT((sched_recv!=NULL));

  switch (type) {
  case DATA_FLOAT:
    fDataMoveRecv(otherProg,sched_recv,(float*)data,msgtag);
    return 1;
    break;
  case DATA_INT:
    iDataMoveRecv(otherProg,sched_recv,(int*)data,msgtag);
    return 1;
    break;
  case DATA_DOUBLE:
    dDataMoveRecv(otherProg,sched_recv,(double*)data,msgtag);
    return 1;
    break;
  case DATA_SHORT:
    sDataMoveRecv(otherProg,sched_recv,(short*)data,msgtag);
    return 1;
    break;
  default:
    return 0;
    break;
  }
  return 0;
}

void
MC_Schedule::emissionRequete(liste_s           *liste_of_zm,
			     MC_SetOfRegions&  setOfRegion,
			     MC_Client&        clnt,
			     MC_DataSender&    mcu,
			     MC_Options&       opt,
			     MC_OtherProgData& knownAboutOtherPgmeData)
{
  int             i;
  membre_liste_s  *l=PremierListe(liste_of_zm);
  liste_s         **tabListeZM;
  ZM_s            *zm;
  ZM_s            *ptZm;
  membre_liste_s  *ptMembreAUtiliser;
  int             zmSize;
  int             zmGlobalOffset;
  int             zmProc;
  int             zmOffset;
  int             zmUsed;
  int             ouTzmSize;
  int             RespProc;
  int             fin=0;
  pgme_s          *other_pgme = clnt.program();

  if (l==NULL) {
    for (i=0;i<NumNodePgme(other_pgme);i++)
      MC_SendListe(NULL,TidPgme(other_pgme,i),FLAG_REQUETE,pvm_pack_ZM);
    return;
  }

  tabListeZM=(liste_s **)malloc(sizeof(liste_s *)*NumNodePgme(other_pgme));
  MC_ASSERT(tabListeZM!=NULL);
  for (i=0;i<NumNodePgme(other_pgme);i++) {
    tabListeZM[i]=AllocListe();
    MC_ASSERT(tabListeZM[i]!=NULL);
  }

  zm             = (ZM_s *)l->elem ;
  zmSize         = zm->size;
  zmGlobalOffset = zm->globalOffset;
  zmProc         = zm->proc;
  zmOffset       = zm->offset;
  zmUsed         = 0;
  fin            = 0; 
  while (fin!=1) { 
    if (zmSize==0) {
      zm            = (ZM_s *)l->elem;
      zmSize        = zm->size;
      zmGlobalOffset= zm->globalOffset;
      zmProc        = zm->proc;
      zmOffset      = zm->offset;
      zmUsed        = 0;
    }
    RespProc= opt.distributeWork(clnt,knownAboutOtherPgmeData,
				 zmGlobalOffset,zmSize,&ouTzmSize);
    if (zmUsed==0) {
      ptZm=zm;
      ptMembreAUtiliser=l;
      l=(membre_liste_s*)MSuivListe(l);
      zmUsed++;
    } else {
      ptZm=alloc_ZM();
      ptMembreAUtiliser=AllocMListe();
    }

    set_ZM(ptZm,zmGlobalOffset,zmProc,zmOffset,ouTzmSize);
    AddElemListeUsingMListe(tabListeZM[RespProc],ptMembreAUtiliser,ptZm);

    zmSize         -= ouTzmSize;
    zmGlobalOffset += ouTzmSize;
    zmOffset       += ouTzmSize;
    if (zmSize==0 && l==NULL)
      fin=1;
  }

  for (i=0;i<NumNodePgme(other_pgme);i++)
    MC_SendListe(tabListeZM[i],TidPgme(other_pgme,i),FLAG_REQUETE,pvm_pack_ZM);
  
  for (i=0;i<NumNodePgme(other_pgme);i++)
    MC_FreeListeZM(tabListeZM[i],free_ZM);

  FreeTListe(liste_of_zm);
  Free(tabListeZM);
}

void
MC_Schedule::computeSchedule(MC_DataSender&    mcu,
			     MC_Server&        srvr,
			     MC_Client&        clnt,
			     MC_Options&       opt,
			     MC_OtherProgData& kopd)
{
  int                i;
  liste_s            *liste_of_zm_for_the_set=NULL;
  liste_s            *liste_of_zm_for_region=NULL;
  MC_OtherProgData&  knownAboutOtherPgmeData = kopd;
  MC_SetOfRegions&   setOfRegion  = mcu.getSetOfRegions();
  pgme_s             *other_pgme  = clnt.program();

  if (this->sched_send!=NULL) 
    return;

  liste_of_zm_for_the_set=AllocListe();
  MC_ASSERT(liste_of_zm_for_the_set!=NULL);
  for (setOfRegion.reset(); !(setOfRegion.end()); setOfRegion.advance()) {

    liste_of_zm_for_region=
      mcu.derefLocal(setOfRegion.getRegion(),
		     setOfRegion.getRegionGlobalOffset()).getListePtr(); 

    liste_of_zm_for_the_set=
      MoveListe(liste_of_zm_for_the_set,liste_of_zm_for_region);
   }

  compress_liste_zm(liste_of_zm_for_the_set);

  /* emission_requete(....) */
  emissionRequete(liste_of_zm_for_the_set,setOfRegion,clnt,mcu,opt,
		  knownAboutOtherPgmeData);

  this->sched_send  = ReceiveSchedule(other_pgme);
  this->sched_recv  = this->sched_send;
  this->otherProg   = other_pgme;

  which_schedule = 1;
}

void
MC_Schedule::traitementRequete(MC_Liste&          mc_liste,
			       MC_SetOfRegions&   setOfRegion,
			       MC_Server&         srvr,
			       MC_Client&         clnt,
			       MC_DataReceiver&   mcu,
			       T_OptimiseComm     optComm)
{
  int     i;
  pgme_s  *receiver_pgme = srvr.program();
  pgme_s  *sender_pgme   = clnt.program();
  sched_s **tab_sched_send;
  sched_s **tab_sched_recv;
  liste_s *liste_zm2; 
  liste_s *liste_sd;
  liste_s *liste_of_zm = mc_liste.getListePtr();

  liste_zm2 = mcu.derefGlobal(mc_liste,setOfRegion,srvr).getListePtr();

  compress_liste_zm(liste_zm2);

  liste_sd=combineListeZM2SD(liste_zm2,liste_of_zm);

  MC_FreeListeZM(liste_of_zm,free_ZM);
  MC_FreeListeZM(liste_zm2,free_ZM);

  tab_sched_recv=(sched_s **)
    malloc(sizeof(sched_s *)*NumNodePgme(receiver_pgme));
  MC_ASSERT(tab_sched_recv!=NULL);

  for (i=0;i<NumNodePgme(receiver_pgme);i++)
    tab_sched_recv[i]=CreateSched(NumNodePgme(sender_pgme)); 

  tab_sched_send=(sched_s **)
    malloc(sizeof(sched_s *)*NumNodePgme(sender_pgme));
  MC_ASSERT(tab_sched_send!=NULL);

  for (i=0;i<NumNodePgme(sender_pgme);i++)
      tab_sched_send[i]=CreateSched(NumNodePgme(receiver_pgme)); 

  computeScheduleForSendAndRecv(tab_sched_recv,
				tab_sched_send,
				liste_sd,
				receiver_pgme,
				sender_pgme);

  MC_FreeListeSD(liste_sd,free_SD);

  sendSchedule(tab_sched_recv,tab_sched_send,receiver_pgme,sender_pgme);

  for (i=0;i<NumNodePgme(receiver_pgme);i++)
    FreeSched(tab_sched_recv[i]);
  for (i=0;i<NumNodePgme(sender_pgme);i++)
    FreeSched(tab_sched_send[i]);
  Free(tab_sched_recv);
  Free(tab_sched_send);
}

void 
MC_Schedule::computeSchedule(MC_DataReceiver&  mcu, 
			     MC_Server&        srvr,
			     MC_Client&        clnt,
			     MC_Options&       opt,
			     MC_OtherProgData& kopd) 
{
  liste_s            *liste_requete_zm=NULL;
  pgme_s             *other_pgme    = clnt.program();
  pgme_s             *receiver_pgme = srvr.program();
  T_OptimiseComm     optComm        = opt.getCommOpt();

  if (this->sched_recv!=NULL)
    return;

  liste_requete_zm=reception_requete(other_pgme);

  MC_Liste mc_liste(liste_requete_zm);

  traitementRequete(mc_liste,mcu.getSetOfRegions(),srvr,clnt,mcu,optComm);

  this->sched_recv  = ReceiveSchedule(receiver_pgme);
  this->sched_send  = this->sched_recv;
  this->otherProg   = clnt.program();

  which_schedule = 2;
}

void
MC_Schedule::traitementLocale(MC_Liste&           mc_liste,
			      MC_SetOfRegions&    setOfRegion,
			      MC_Server&          srvr,
			      MC_DataSame&        mcu,
			      sched_s             **sched_send,
			      sched_s             **sched_recv)
{
  int     i;
  sched_s **tab_sched_send;
  sched_s **tab_sched_recv;
  liste_s *liste_zm2=mcu.derefGlobal(mc_liste,setOfRegion,srvr).getListePtr();
  liste_s *liste_sd;
  liste_s *liste_of_zm=mc_liste.getListePtr();
  pgme_s  *pgme = srvr.program();

  compress_liste_zm(liste_zm2);
  
  liste_sd=combineListeZM2SD(liste_zm2,liste_of_zm);
  MC_FreeListeZM(liste_of_zm,free_ZM);
  MC_FreeListeZM(liste_zm2,free_ZM);

  tab_sched_recv=(sched_s **)malloc(sizeof(sched_s *)*NumNodePgme(pgme));
  MC_ASSERT(tab_sched_recv!=NULL);
  for (i=0;i<NumNodePgme(pgme);i++)
    tab_sched_recv[i]=CreateSched(NumNodePgme(pgme)); 

  tab_sched_send=(sched_s **)malloc(sizeof(sched_s *)*NumNodePgme(pgme));
  MC_ASSERT(tab_sched_send!=NULL);
  for (i=0;i<NumNodePgme(pgme);i++)
    tab_sched_send[i]=CreateSched(NumNodePgme(pgme)); 

  computeScheduleForSendAndRecv(tab_sched_recv,tab_sched_send,
				liste_sd,pgme,pgme);
  MC_FreeListeSD(liste_sd,free_SD);

  sendScheduleLocal(tab_sched_recv,tab_sched_send,
		    pgme,sched_send,sched_recv);

  for (i=0;i<NumNodePgme(pgme);i++)
    FreeSched(tab_sched_recv[i]);
  for (i=0;i<NumNodePgme(pgme);i++)
    FreeSched(tab_sched_send[i]);
  Free(tab_sched_recv);
  Free(tab_sched_send);
}

void
MC_Schedule::computeSchedule(MC_DataSame&      mcu, 
			     MC_Server&        srvr,
			     MC_Options&       opt,
			     MC_OtherProgData& kopd)
{
  int i;
  liste_s *liste_of_zm_for_the_set=NULL;
  liste_s *liste_of_zm_for_region=NULL;
  MC_OtherProgData&  knownAboutOtherPgmeData = kopd;
  MC_SetOfRegions&   setOfRegionSender    = mcu.getSetOfRegionsReceiver();
  MC_SetOfRegions&   setOfRegionReceiver  = mcu.getSetOfRegionsReceiver();
  sched_s            *pt_sched_send;
  sched_s            *pt_sched_recv;

  if (this->sched_send!=NULL || this->sched_recv!=NULL) 
    return;

  liste_of_zm_for_the_set=AllocListe();
  MC_ASSERT(liste_of_zm_for_the_set!=NULL);
  for (setOfRegionSender.reset(); !(setOfRegionSender.end()); 
       setOfRegionSender.advance()) {
    
    liste_of_zm_for_region=
      mcu.derefLocal(setOfRegionSender.getRegion(),
		     setOfRegionSender.getRegionGlobalOffset()).getListePtr(); 

    liste_of_zm_for_the_set=
      MoveListe(liste_of_zm_for_the_set,liste_of_zm_for_region);
  }

  compress_liste_zm(liste_of_zm_for_the_set);

  MC_Liste liste(liste_of_zm_for_the_set);
  traitementLocale(liste,setOfRegionReceiver,srvr,mcu,
		   &(this->sched_send),
		   &(this->sched_recv));
  this->otherProg = srvr.program();

  pt_sched_send=this->sched_send;
  pt_sched_recv=this->sched_recv;
  
  which_schedule = 3;

  compress_schedule(pt_sched_send);
  compress_schedule(pt_sched_recv);
}

#endif /* MC_SCHEDULE_C */
