
/* $Id: w-rtree.C,v 1.8 1999/06/30 19:55:57 chialin Exp $ */

/* w-rtree.C */

#include	<iostream.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<unistd.h>
#include	"w_rtree.h"


////////////////////////////////////////////////////////////////////////////
//
// W_GistM methods
//
////////////////////////////////////////////////////////////////////////////
rc_t
W_GistM::attach(int fd, gist_ext_t* extension)
{
  assert (extension != NULL);
  if (fd < 0 || isOpen)		// invalid fd or gist_m is already open
     return eFILEERROR;

  ext = extension;
  W_DO(file.attach(fd));
  isOpen = true;
  return RCOK;
}



////////////////////////////////////////////////////////////////////////////
//
// W_RTree methods
//
////////////////////////////////////////////////////////////////////////////
W_RTree::W_RTree(): rt_index(), open_mode(0), cursor(), datasz(0)
{ 
    opened = 0; dimension = 0; qry = NULL;
    rectQry = NULL; eof=false; 
}


W_RTree::W_RTree(const unsigned int dim)
		: rt_index(), open_mode(0), cursor(), datasz(0)
{ 
    opened = 0; dimension = dim; qry = NULL; 
    rectQry = NULL; eof = false;
}


W_RTree::~W_RTree()
{
    close();
    delete qry;
    delete rectQry;
}


int
W_RTree::open(const char *fname, int mode)
{
  if (opened)			// already opened an R-tree file
     return 0;

  int r;
  opened = 0;
  open_mode = mode;
  switch (mode) {
      case W_CREATE_FILE: {
	r = rt_index.create(fname,&rt_rect_ext);
	if (r!=RCOK) {		// index already exists, overwrite it
	   unlink(fname);
	   r = rt_index.create(fname,&rt_rect_ext);
	   if (r==RCOK) break;
	   return 0;
	}
	break;
      }
      case W_APPEND_FILE: {
	r = rt_index.open(fname,&rt_rect_ext);
	if (r!=RCOK) {		// index does not exist, create it
	   r = rt_index.create(fname,&rt_rect_ext);
	   if (r==RCOK) break;
	   return 0;
         }
	break;
      }
      case W_READ_FILE: {
	r = rt_index.open(fname,&rt_rect_ext);
	if (r!=RCOK) return 0;
	break;
      }
      default:
	return 0;
  }

  opened = 1;
  return 1;
}


int
W_RTree::close()
{ 
    if (opened) {
	opened = 0;
	if (rt_index.close() == RCOK)
	   return 1;
    }
    return 0;
} 


int
W_RTree::flush()
{
    if (!opened || open_mode == W_READ_FILE) {
	// no file is open or file was open for read-only
	return 0;
    }
    if (rt_index.flush() == RCOK)
	return 1;
    return 0;
}


int
W_RTree::insert(const double inp[], void *data, size_t size)
{
    int r;
    if (dimension<=0) return 0;
    r = rt_index.insert((void *)inp,sizeof(double)*dimension*2,
			(void*)data,(smsize_t)size);
    datasz = size;
    if (r==0) return 1;
    return 0;
}


int
W_RTree::initSearch(const double qw[])
{
    rc_t r;

    delete qry;
    delete rectQry;

    if (!opened) return 0;
    if (dimension<=0) return 0;

    rectQry = new rt_rect(dimension,qw);
    qry     = new rt_query_t(rt_query_t::rt_overlap, rt_query_t::rt_rectarg, 
			     (void *)rectQry);

    r = rt_index.fetch_init(cursor,qry);
    if (r!=RCOK) return 0;
    return 1;
}


int
W_RTree::getNext(double out[], u_int& dim, void *data, size_t& size)
{
    if (!opened) return 0;
    if (dimension<=0) return 0;

    rt_rect key(dimension*2);
    smsize_t keysz = 2*dimension*sizeof(double);
    datasz = size;
    int r = rt_index.fetch(cursor,(void*)key.coord,keysz,(void*)data,
			   datasz,eof);

//    cout << "rt_index.fetch returns " << r << endl;
//    switch (r) {
//      case eBADSLOTNUMBER: cout << "rc = eBADSLOTNUMBER" << endl; break;
//      case eRECWONTFIT: cout << "rc = eRECWONTFIT" << endl; break;
//      case eNOTOPEN: cout << "rc = eNOTOPEN" << endl; break;
//      case eFILEERROR: cout << "rc = eFILEERROR" << endl; break;
//      case eNOBUFFERS: cout << "rc = eNOBUFFERS" << endl; break;
//      case eOUTOFBOUNDS: cout << "rc = eOUTOFBOUNDS" << endl; break;
//      case eOUTOFSPACE: cout << "rc = eOUTOFSPACE" << endl; break;
//    }

    if (r!=RCOK) { key.dealloc(); return 0; }
    if (eof) { key.dealloc(); return -1; }

    size = (unsigned int) datasz;
    dim  = (unsigned int) dimension;

    for (unsigned int i=0;i<2*dimension;i++) 
      out[i] = key.coord[i];
    key.dealloc();
    return 1;
}



////////////////////////////////////////////////////////////////////////////
//
// W_StaticRTree methods
//
////////////////////////////////////////////////////////////////////////////
int
W_StaticRTree::attach(int fd)
{
  if (fd < 0)
     return 0;
  rc_t r = rtree.rt_index.attach(fd, &rt_rect_ext);
  if (r==RCOK) { rtree.opened = 1; return 1; }
  return 0;
}

