
/* $Id: t2_accmsgrecv.h,v 1.3 2000/01/11 20:58:58 chialin Exp $ */

/* t2_accmsgrecv.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 8/11/98
 */


#ifndef _T2_ACCMSGRECV_H_
#define _T2_ACCMSGRECV_H_

#include	"t2_header.h"
#include	"t2_bits.h"


////////////////////////////////////////////////////////////////////////////
//
// T2_AccMsgSrc
//
// A class that contains a src node and a pointer to the acc msg received.
//
////////////////////////////////////////////////////////////////////////////
struct T2_AccMsgSrc {
  T2_ProcID src;
  T2_AccMsgRecvBuffer* msg;

  T2_AccMsgSrc() { src = 0; msg = NULL; };
  T2_AccMsgSrc(T2_ProcID p, T2_AccMsgRecvBuffer* m) { src = p; msg = m; };
  ~T2_AccMsgSrc() {};
};



////////////////////////////////////////////////////////////////////////////
//
// T2_AccMsgRecvBufList
//
// A list of received acc msgs and a bit vector to keep track of which
// processor has already sent me an acc msg.
//
////////////////////////////////////////////////////////////////////////////
class T2_AccMsgRecvBufList {
public:
  T2_AccMsgRecvBufList(u_int nprocs = 0);
  ~T2_AccMsgRecvBufList();

  // query methods
  u_int getNumberMsgs() const { return accmsgrecvbuf.getNumberElements(); };
  bool isSet(u_int i) const;
  const T2_BitVector& getBitVector() const  { return procs; };

  // set a bit in the bit vector
  void setBitVector(u_int i);

  // get and set the number of processors
  u_int getNumberProcessors() const { return accmsgrecvbuf.getMaxSetSize(); };
  void setNumberProcessors(u_int nprocs);

  // access 
  bool getAccMsgRecvBuffer(u_int i, T2_ProcID& src, T2_AccMsgRecvBuffer*& msg)
									const;
  bool insertAccMsgRecvBuffer(T2_ProcID src, T2_AccMsgRecvBuffer* msg);

  // reset the data sturcture
  void clearAll() { accmsgrecvbuf.emptySet(); procs.clrAll(); };

protected:
  T2_TSet<T2_AccMsgSrc> accmsgrecvbuf;	// buffer/src for recv'd acc msg
  T2_BitVector procs;

private:
};


inline void
T2_AccMsgRecvBufList::setNumberProcessors(u_int nprocs)
{
  t2_Assert (getNumberMsgs() == 0);
  accmsgrecvbuf.clear(nprocs);
  procs.setNumberBits(nprocs);
}

inline
T2_AccMsgRecvBufList::T2_AccMsgRecvBufList(u_int nprocs):
				accmsgrecvbuf(nprocs), procs(nprocs)
{
  setNumberProcessors(nprocs);
}

inline
T2_AccMsgRecvBufList::~T2_AccMsgRecvBufList()
{
  t2_Assert (getNumberMsgs() == 0);
}

inline bool
T2_AccMsgRecvBufList::isSet(u_int i) const
{
  t2_Assert (i<getNumberProcessors());
  return procs.isSet(i);
}

inline void
T2_AccMsgRecvBufList::setBitVector(u_int i)
{
  t2_Assert (i<getNumberProcessors());
  t2_Assert (isSet(i) == false);
  procs.set(i);
}

inline bool
T2_AccMsgRecvBufList::getAccMsgRecvBuffer(u_int i, T2_ProcID& src,
					  T2_AccMsgRecvBuffer*& msg) const
{
  if (i >= getNumberMsgs())
     return false;

  const T2_AccMsgSrc& e = accmsgrecvbuf[i];
  src = e.src;
  msg = e.msg;
  return true;
}

inline bool
T2_AccMsgRecvBufList::insertAccMsgRecvBuffer(T2_ProcID src,
					     T2_AccMsgRecvBuffer* msg)
{
  t2_Assert ((u_int) src < getNumberProcessors());
  t2_Assert (!accmsgrecvbuf.isFull());
  t2_Assert (!procs.isSet(src));

  procs.set(src);
  return accmsgrecvbuf.in(T2_AccMsgSrc(src, msg));
}



#endif	/* ! _T2_ACCMSGRECV_H_ */
