
/* $Id: t2_index.h,v 1.7 2000/02/09 20:45:23 chialin Exp $ */

/* t2_index.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/8/98
 * Dataset index class
 */

#ifndef _T2_INDEX_H_
#define _T2_INDEX_H_


#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_cluster.h"
#include	"t2_shape.h"
#include	"t2_blockreq.h"
#include	"t2_usrarg.h"
#include	"t2_system.h"


/////////////////////////////////////////////////////////
//
// T2_ClusterAuxInfoWriter
//
// A class to hold additional information of a cluster
// generated by the T2_Index::fetch(). The information
// could be any thing that the index sees appropriate,
// and will be passed to the user-defined chunk
// processing functions when the chunk is retrieved.
//
/////////////////////////////////////////////////////////
class T2_ClusterAuxInfoWriter: public T2_UsrArgWriter {
public:
  T2_ClusterAuxInfoWriter(T2_UsrArg& a): T2_UsrArgWriter(a) {};
  ~T2_ClusterAuxInfoWriter() {};

  // throw away the old data of arg and allocate new space of s bytes in arg
  size_t allocBuffer(size_t s) { return T2_UsrArgWriter::allocBuffer(s); };

  // grow the buffer to size s (NOT increment by s), and copy stuff from
  // old buffer to new buffer (not implemented yet)
  // size_t growBuffer(size_t s);

  // do not use copy constructor and operator =
  T2_ClusterAuxInfoWriter(const T2_ClusterAuxInfoWriter& w);
  T2_ClusterAuxInfoWriter& operator =(const T2_ClusterAuxInfoWriter& w);

protected:
private:
};


inline
T2_ClusterAuxInfoWriter::T2_ClusterAuxInfoWriter(const T2_ClusterAuxInfoWriter& w)
				: T2_UsrArgWriter(w)
{
  t2_err_quit(__FILE__, __LINE__,
             "do not use T2_ClusterAuxInfoWriter copy constructor");
}

inline T2_ClusterAuxInfoWriter&
T2_ClusterAuxInfoWriter::operator =(const T2_ClusterAuxInfoWriter&)
{
  t2_err_quit(__FILE__, __LINE__,
             "do not use T2_ClusterAuxInfoWriter operator =");
  return *this;
}



/////////////////////////////////////////////////////////
//
// T2_ClusterAuxInfoReader
//
// A class to read data written by
//  T2_ClusterAuxInfoWriter. 
//
/////////////////////////////////////////////////////////
class T2_ClusterAuxInfoReader: public T2_UsrArgReader {
public:
  T2_ClusterAuxInfoReader(T2_UsrArg& a): T2_UsrArgReader(a) {};
  ~T2_ClusterAuxInfoReader() {};

protected:
private:
};



/////////////////////////////////////////////////////////
//
// T2_Index
//
// T2 base class for an index of a dataset.
//
/////////////////////////////////////////////////////////
class T2_Index {
public:
  T2_Index() {};
  virtual ~T2_Index() {};

  // some user-defined functions
  virtual T2_UDFRet fetchInit(const T2_System& system, const T2_Box& qr) = 0;
  virtual T2_UDFRet fetch(const T2_System& system, const T2_Box& qr,
			  bool& eod, T2_Box& mbr, T2_ClusterAuxInfoWriter& info,
			  T2_VArray<T2_BlockRequest>& chk) = 0;

protected:
private:
};



/////////////////////////////////////////////////////////
//
// T2_Index constructors
//
/////////////////////////////////////////////////////////
typedef T2_UDFRet (T2_IndexConstructor)
	          (const T2_System& system, T2_UsrArgReader& arg,
		   const T2_Array<int>& fd, T2_Index*& idxp);
typedef T2_UDFRet (*T2_IndexConstructorPtr)
	          (const T2_System& system, T2_UsrArgReader& arg,
		   const T2_Array<int>& fd, T2_Index*& idxp);



#endif	/* _T2_INDEX_H_ */
