
/* $Id: t2_qenv.h,v 1.14 2000/03/16 00:46:24 chialin Exp $ */

/* t2_qenv.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/2/98
 * a query environment
 */


#ifndef _T2_QENV_H_
#define _T2_QENV_H_


#include	<iostream.h>
#include	"t2_header.h"
#include	"t2_system.h"
#include	"t2_qspec.h"
#include	"t2_acc.h"
#include	"t2_aggr.h"
#include	"t2_prj.h"
#include	"t2_dataset.h"
#include	"t2_shape.h"
#include	"t2_rts.h"
#include	"t2_clusterreq.h"
#include	"t2_output.h"
#include	"t2_qplan.h"
#include	"t2_qstatus.h"
#include	"t2_accmsgrecv.h"
#include	"t2_queryinfo.h"


#ifdef T2_PBACKEND

   #include	"t2_outhandle.h"

#else

   typedef char		T2_OutputHandle;

#endif	/* T2_PACKEND */


// forward declaration
class T2_QEnv;
class T2_WorkList;

//////////////////////////////////////////////////////////////////////
//
// T2_DatasetEnv
//
//////////////////////////////////////////////////////////////////////
class T2_DatasetEnv {
public:
  T2_DatasetEnv();
  T2_DatasetEnv(const T2_DatasetEnv&);
  ~T2_DatasetEnv();

  // operator =
  T2_DatasetEnv& operator =(const T2_DatasetEnv&);

  // initialization from a query dataset specification
  bool init(T2_QSpecDataset& qspec, T2_RTS& rts);

  // set qenv pointer
  void setQEnv(T2_QEnv* e) { qenv = e; };
  
  // valid?
  operator const void* () const;

  // const access methods
  T2_DSID getDatasetID() const { return ds_id; };
  u_int getIteratorID() const { return iter_id; };
  T2_IndexID getIndexID() const { return index_id; };
  const T2_Box& getQueryRegion() const { return qregion; };
  const T2_ClusterRequestList& getClusterRequestList() const { return reqs; };

  // non-const access methods
  T2_Dataset* getDatasetPtr() { return dataset; };
  T2_ProjectFuncObj* getProjectFuncObjPtr() { return prj; };
  T2_ClusterRequestList& getClusterRequestList() { return reqs; };
  T2_QEnv* getQEnv() { return qenv; };
  T2_DatasetEnv* getAddress() { return this; };
  T2_ProjectFuncObj* getProjectFuncObj() { return prj; };

protected:
  T2_Dataset* dataset;
  T2_DSID ds_id;		// dataset id
  u_int iter_id;		// dataset iterator id

  T2_IndexID index_id;		// index id

  T2_Box qregion;		// query region

  T2_ProjectFuncObj *prj;	// projection function object

  T2_ClusterRequestList reqs;	// a list of cluster requests for qenv

  T2_QEnv* qenv;		// a back pointer to the owning qenv

private:
};


inline
T2_DatasetEnv::operator const void* () const
{
  if (dataset==NULL || prj==NULL)
     return 0;
  return this;
}



//////////////////////////////////////////////////////////////////////
//
// T2_OutputEnv
//
//////////////////////////////////////////////////////////////////////
class T2_OutputEnv {
public:
  T2_OutputEnv(): type(t2_oUnknown), outhandle(NULL) {};
  ~T2_OutputEnv() { delete outhandle; };

  // query methods
  T2_OutputHandleType getType() const { return type; };

  // initialization from query spec
  bool init(T2_QEnv* p, T2_RTS& rts, T2_QSpecOutput& qspec);

  // initialize before query execution
  bool initBeforeQueryExecution(T2_RTS& rts);

  // if *outhandle is blocking
  bool isBlocking() const;

  // if *outhandle is blocking and not all isends for local-reduction phase
  // complete
  bool blockingAndISendsNotComplete() const;

  // handle the output
  bool handleOutput(T2_Output& output, T2_Accumulator& acc, T2_Iteration i);

  // finalize the output handle -- called when a query is completed done
  void finalize();

  // get output handler
  T2_OutputHandle* getOutputHandle() { return outhandle; };

protected:
  T2_OutputHandleType type;
  T2_OutputHandle* outhandle;

private:
};


inline bool
T2_OutputEnv::initBeforeQueryExecution(T2_RTS& rts)
{
#ifdef T2_PBACKEND

  t2_Assert (outhandle != NULL);
  return outhandle->initBeforeQueryExecution(rts);

#else

  return true;

#endif
}

inline bool
T2_OutputEnv::isBlocking() const
{
#ifdef T2_PBACKEND

  t2_Assert (outhandle != NULL);
  return outhandle->isBlocking();

#else

  return false;

#endif
}

inline bool
T2_OutputEnv::blockingAndISendsNotComplete() const
{
#ifdef T2_PBACKEND

  t2_Assert (outhandle != NULL);
  return outhandle->blockingAndISendsNotComplete();

#else

  return false;

#endif
}

inline void
T2_OutputEnv::finalize()
{
#ifdef T2_PBACKEND

  t2_Assert (outhandle != NULL);
  outhandle->finalize();

#endif
}



//////////////////////////////////////////////////////////////////////
//
// T2_QEnv
//
//////////////////////////////////////////////////////////////////////
class T2_QEnv {
public:
  T2_QEnv();
  T2_QEnv(const T2_QEnv& e);
  ~T2_QEnv();

  // operator =
  T2_QEnv& operator =(const T2_QEnv& e);

  // initialize every thing but datasets from a query specification
  bool init(u_int qid, T2_QSpec& qspec, T2_RTS& rts);

  // valid?
  operator const void*() const { return (valid == true)? this : NULL; };
  bool isValid() const { return valid; };
  void invalidate() { valid = false; };

  // kill the query during query execution
  void killQueryNow();

  // access methods to qenv_id
  u_int getQEnvID() const { return qenv_id; };

  // access methods
  T2_AccMetaObj* getAccMetaPtr() { return accmeta; };
  T2_Accumulator* getAccumulator() { return acc; };
  u_int getNumberDatasets() const { return datasets.getNumberElements(); };
  const T2_DatasetEnv& getDatasetEnv(u_int i) const { return datasets[i]; };
  T2_DatasetEnv& getDatasetEnv(u_int i) { return datasets[i]; };
  T2_Array<T2_DatasetEnv>& getDatasetEnvList() { return datasets; };
  u_int getAccNavigatorID() const { return acc_navid; };
  T2_AggregateFuncObj* getAggregateFuncObject() { return aggrf; };
  T2_Output* getOutput() { return output; };
  const T2_QueryInfo& getQueryInfo() const { return qinfo; };

  // access methods for query windows
  u_int getNumberQueryWindows() const { return qwindows.getNumberElements(); };
  const T2_VArray<T2_Region>& getQueryWindows() const { return qwindows; };
  T2_VArray<T2_Region>& getQueryWindows() { return qwindows; };
  const T2_Region& getQueryWindow(u_int i) const { return qwindows[i]; };
  T2_Region& getQueryWindow(u_int i) { return qwindows[i]; };
  void insertQueryWindow(T2_Region& b) { qwindows << b; };

  // create the accumulator
  bool allocateAccumulator(u_int iter);

  // delete the accumulator
  void deleteAccumulator() { delete acc; acc = NULL; };

  // finalize the output from accumulator
  void finalizeOutput();

  // delete the output
  void deleteOutput() { delete output; output = NULL; };

  // set and access T2_QQPlan
  T2_QQPlan* getQQPlan() const { return qqplan; };
  void setQQPlan(T2_QQPlan* q) { qqplan = q; };

  // method to initialize stuff at the beginning of query execution
  // (ie only invoked once for a query)
  bool initBeforeQueryExecution(T2_RTS& rts);

#ifdef T2_PBACKEND
  // method to initialize stuff at the beginning of a strip-mined iteration;
  // returns false if T2_QQPlan says all iterations are already done,
  // true otherwise
  bool initIteration(T2_WorkList& worklist);
#endif

  // method to clear stuff at the end of a strip-mined iteration
  void finalizeIteration();

  // access the # of cluster requests yet to be processed
//  u_int getNumberClustersYetToProcess() const {return ncluster_yetprocessed;};
//  u_int decNumberClustersYetToProcess() { return --ncluster_yetprocessed; };

  // access to T2_OutputEnv
  const T2_OutputEnv& getOutputEnv() const { return outenv; };
  T2_OutputEnv& getOutputEnv() { return outenv; };

  // access to stuff related to the global-combine phase
  T2_Array<size_t>& getAccMsgSizes() { return accmsgsizes; };
  const T2_Array<size_t>& getAccMsgSizes() const { return accmsgsizes; };
  T2_Array<T2_AccMsgSendBuffer>& getAccMsgSendBuffer() { return accmsgsendbuf;};
  const T2_Array<T2_AccMsgSendBuffer>& getAccMsgSendBuffer() const;
  u_int getNumberAccMsgRecvBuffers() const;
  void getNumberAccMsgRecvBuffers(u_int &i, u_int& j) const;
  bool getAccMsgRecvBuffer(u_int i, T2_ProcID& src, T2_AccMsgRecvBuffer*& msg)
									const;
//  bool insertAccMsgRecvBuffer(T2_ProcID src, T2_AccMsgRecvBuffer* msg);
//  void emptyAccMsgRecvBuffer() { accmsgrecvbuf.emptySet(); };
  bool deferAccMsgProcess(T2_ProcID src, T2_AccMsgRecvBuffer* msg);
  T2_AccMsgRecvBufList** getAccMsgRecvBufPointer() { return accmsgrecvbuf; };

  // dealing with query status
  void initGlobalCombine(u_int naccmsgs_toprocess, u_int nsends);
  void finalizeGlobalCombine();
  void initOutputHandle(u_int nwrites) { status.initOutputHandle(nwrites); };
  void initQueryDone() { status.initQueryDone(); outenv.finalize(); };
  bool decNumberClustersYetToProcessAndCheckEndLocalReduction();
  bool decNumberAccMsgsYetProcessAndCheckEndGlobalCombine();
  T2_QState decNumberAccSendsYetCompleteAndCheckEndOfStatus();
  bool decNumberOutputWritesYetCompleteAndCheckEndOutputHandle();
  bool endOutputHandle() { return status.endOutputHandle(); };
  u_int decNumberOutputWritesYetComplete();

//  u_int getNumberPendingAccISends() const { return naccsends_notcomplete; };
//  u_int incNumberPendingAccISends(u_int i=1);
//  u_int decNumberPendingAccISends() { return --naccsends_notcomplete; };
//  u_int getNumberPendingAccIRecvs() const { return naccrecvs_notcomplete; };
//  u_int incNumberPendingAccIRecvs(u_int i=1);
//  u_int decNumberPendingAccIRecvs() { return --naccrecvs_notcomplete; };
//  u_int getNumberAccMsgsYetToProcess() const { return naccmsgs_yetprocessed;};
//  u_int incNumberAccMsgsYetToProcess(u_int i=1);
//  u_int decNumberAccMsgsYetToProcess() { return --naccmsgs_yetprocessed; };

  // is the global combine phase completed?
//  bool doneGlobalCombine() const;

  // access to query status
  const T2_QStatus& getQueryStatus() const { return status; };
  T2_QStatus& getQueryStatus() { return status; };

protected:
  enum invalidKey { NO_ACCMETA, NO_AGGR, NO_DATASET, NO_OUTHANDLE};
  u_int qenv_id;		// the index of this qenv in the qenv array

  bool valid;			// is the query a valid query?

  T2_AccMetaObj* accmeta;	// accumulator meta object
  u_int acc_navid;		// accumulator navigation function id
  T2_Accumulator *acc;		// accumulator
  T2_Output *output;		// final output

  T2_AggregateFuncObj* aggrf;	// aggregation function object

  T2_Array<T2_DatasetEnv> datasets;	// a list of dataset objects

  T2_VArray<T2_Region> qwindows;   // a list of query windows, one per iteration

  T2_QQPlan* qqplan;		// a pointer to its T2_QQPlan in T2_QPlan

  T2_OutputEnv outenv;		// stuff for handling final output

  // stuff used in the global combine phase
  T2_Array<size_t> accmsgsizes;		// acc msg size, one per processor
  T2_Array<T2_AccMsgSendBuffer> accmsgsendbuf;	// buffer for sending acc msg,
						// one per processor
  T2_AccMsgRecvBufList *accmsgrecvbuf[2];	// buffer/src of recv'd acc 
						// msgs for current and next
						// iterations (tiles)


  T2_QStatus status;		// current query status

  T2_QueryInfo qinfo;		// information about the query

  // some error reporting methods
  ostream& reportInitError(ostream& os, invalidKey k, const T2_QSpec& qspec,
			   u_int i=0) const;

private:
};


inline
T2_QEnv::T2_QEnv():
	qenv_id(0),
	valid(false), accmeta(NULL), acc_navid(0), acc(NULL), output(NULL),
	aggrf(NULL), datasets(0) , qwindows(10), qqplan(NULL),
	outenv(),
	accmsgsizes(0), accmsgsendbuf(0), 
	status(), qinfo(qqplan)
{
  accmsgrecvbuf[0] = new T2_AccMsgRecvBufList;
  accmsgrecvbuf[1] = new T2_AccMsgRecvBufList;
}

inline
T2_QEnv::T2_QEnv(const T2_QEnv& e):
	qenv_id(0),
	valid(false), accmeta(NULL), acc_navid(0), acc(NULL), output(NULL),
	aggrf(NULL), datasets(0) , qwindows(10), qqplan(NULL),
	outenv(),
	accmsgsizes(0), accmsgsendbuf(0),
	status(), qinfo(qqplan)
{
  t2_Assert(!e);
  accmsgrecvbuf[0] = accmsgrecvbuf[1] = NULL;
}

inline T2_QEnv&
T2_QEnv::operator =(const T2_QEnv& e)
{
  t2_Assert(!e);
  accmsgrecvbuf[0] = accmsgrecvbuf[1] = NULL;
  return *this;
}

inline
T2_QEnv::~T2_QEnv()
{
  delete accmeta;
  delete aggrf;
  delete acc;
  delete output;
  delete accmsgrecvbuf[0];
  delete accmsgrecvbuf[1];
}

inline bool
T2_QEnv::initBeforeQueryExecution(T2_RTS& rts)
{
  t2_Assert(*this);

  if (outenv.initBeforeQueryExecution(rts) == false) {
     return false;
  }

  return true;
}

inline const T2_Array<T2_AccMsgSendBuffer>&
T2_QEnv::getAccMsgSendBuffer() const
{
  return accmsgsendbuf;
}

inline u_int
T2_QEnv::getNumberAccMsgRecvBuffers() const
{
  t2_Assert (accmsgrecvbuf[0] != NULL);
  return accmsgrecvbuf[0]->getNumberMsgs();
}

inline void
T2_QEnv::getNumberAccMsgRecvBuffers(u_int& i, u_int& j) const
{
  t2_Assert (accmsgrecvbuf[0] != NULL);
  t2_Assert (accmsgrecvbuf[1] != NULL);
  i = accmsgrecvbuf[0]->getNumberMsgs();
  j = accmsgrecvbuf[1]->getNumberMsgs();
}

inline bool
T2_QEnv::getAccMsgRecvBuffer(u_int i, T2_ProcID& src, T2_AccMsgRecvBuffer*& msg)
const
{
  t2_Assert (accmsgrecvbuf[0] != NULL);
  return accmsgrecvbuf[0]->getAccMsgRecvBuffer(i, src, msg);
}

//inline u_int
//T2_QEnv::incNumberAccMsgsYetToProcess(u_int i)
//{
//  return (naccmsgs_yetprocessed += i);
//}

//inline bool
//T2_QEnv::doneGlobalCombine() const
//{
//  if (getNumberPendingAccISends() == 0 && 
//      getNumberAccMsgsYetToProcess() == 0 &&
//      getQueryStatus() == t2_qGlobal)
//     return true;
//  return false;
//}

inline void
T2_QEnv::initGlobalCombine(u_int naccmsgs_toprocess, u_int nsends)
{
  // note that number of acc msgs to recv (argument 2 for method
  // status.initGlobalCombine()) equals to the number of acc msgs
  // to process minus the number of acc msgs for this iteration
  // already received and stored in acc msg recv buffers
  status.initGlobalCombine(naccmsgs_toprocess,
			   naccmsgs_toprocess - getNumberAccMsgRecvBuffers(),
			   nsends);
}

inline bool
T2_QEnv::decNumberClustersYetToProcessAndCheckEndLocalReduction()
{
  t2_Assert (qqplan != NULL);
  return status.decNumberClustersYetToProcessAndCheckEndLocalReduction(*qqplan);
}

inline bool
T2_QEnv::decNumberAccMsgsYetProcessAndCheckEndGlobalCombine()
{
  status.decNumberAccMsgsYetProcess();
  return status.endGlobalCombine(outenv.isBlocking());
}

inline T2_QState
T2_QEnv::decNumberAccSendsYetCompleteAndCheckEndOfStatus()
{
  return
    status.decNumberAccSendsYetCompleteAndCheckEndOfStatus(outenv.isBlocking());
}

inline bool
T2_QEnv::decNumberOutputWritesYetCompleteAndCheckEndOutputHandle()
{
  status.decNumberOutputWritesYetComplete();
  return status.endOutputHandle();
}

inline u_int
T2_QEnv::decNumberOutputWritesYetComplete()
{
  return status.decNumberOutputWritesYetComplete();
}


#endif	/* _T2_QENV_H_ */
