
/* aggr.C */


#include	<iostream.h>
#include	"t2_aggr.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_AggregateFuncObj methods
//
/////////////////////////////////////////////////////////////////////////
T2_UDFRet
T2_AggregateFuncObj::aifAcc(const T2_QueryInfo& qinfo, T2_Accumulator& acc)
{
  t2_say3(cout, "calling T2_AggregateFuncObj::aifAcc" << endl);

  T2_AccIterator* iter = NULL;
  if (acc.navigateAll(qinfo, iter) != T2_UDFRet_OK) {
      t2_err_ret(__FILE__, __LINE__, "accumulator.navigateAll returns ERROR");
      return T2_UDFRet_ERROR;
  }
  if (iter == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"accumulator.navigateAll returns a NULL pointer"); 
     return T2_UDFRet_ERROR;
  }

  T2_UDFRet rv;
  bool eod = false;
  void* acc_elem = NULL;
  while ((rv = iter->getNextElement(qinfo,acc,eod,acc_elem)) == T2_UDFRet_OK) {
    if (eod == true)
       break;

    if ((rv = aifElem(qinfo, acc_elem)) != T2_UDFRet_OK) {
       t2_err_ret(__FILE__, __LINE__, "aggr.aifElem returns ERROR");
       break;
    }
    eod = false;
    acc_elem = NULL;
  }

  delete iter;
  return rv;
}


T2_UDFRet
T2_AggregateFuncObj::dafAcc(const T2_QueryInfo& qinfo,
			    const T2_Cluster& input, const T2_Box& mbr,
			    T2_ClusterAuxInfoReader& meta,
			    T2_Dataset& dataset, T2_DSID dsid,
			    T2_IteratorID dsiterid,
			    const T2_Box& iqr,
			    const T2_Region& tilereg, T2_ProjectFuncObj& proj,
			    u_int acciterid, T2_Accumulator& acc)
{
  // mbr is the bounding box of the retrieved input data cluster
  // iqr is the query window given by the query
  // tilereg is the currnet output query region of the accumulator for 
  //	 the current strip-mined tile

  t2_say4(cout, "enter T2_AggregateFuncObj::dafAcc" << endl);

  T2_Iterator *iter = NULL;
  if (dataset.genIterator(qinfo.getSystem(), dsiterid, input, mbr, meta,
			  iqr, tilereg, proj, iter) != T2_UDFRet_OK)
  {
     t2_err_ret(__FILE__, __LINE__, "dataset.genIterator returns ERROR");
     return T2_UDFRet_ERROR;
  }

  if (iter == NULL) {
     t2_err_ret(__FILE__, __LINE__,"dataset.genIterator returns NULL iterator");
     return T2_UDFRet_ERROR;
  }


# if (T2_VERBOSE > 4)
     u_int input_cnt = 0;
# endif

  T2_UDFRet rv;
  bool input_eod = false;
  const void* input_elem = NULL;
  T2_Point coord;
  while ((rv = iter->getNextElement(qinfo.getSystem(), input, input_eod,
				    input_elem, coord)) == T2_UDFRet_OK)
  {

#    if (T2_VERBOSE > 4)
	input_cnt++;
	if (input_eod == true)
	   t2_say5(cout, "input_iterator.getNextElement is called the "
			  << input_cnt
			  << "-th time and returns end-of-data=true" << endl);
	else
	   t2_say5(cout, "input_iterator.getNextElement is called the "
			  << input_cnt
			  << "-th time and returns end-of-data=false" << endl);
#    endif

     if (input_eod == true)
	break;					// all done

     bool intersects = false;
     T2_Region rg;
     if (proj.project(qinfo.getSystem(), iqr, tilereg, coord, intersects, rg)
	!= T2_UDFRet_OK)
     {
	t2_err_ret(__FILE__, __LINE__, "proj.project returns ERROR");
	delete iter;
	return T2_UDFRet_ERROR;
     }
     if (intersects == true) {
        T2_AccIterator *acciter = NULL;
	if (acc.navigate(qinfo, acciterid, rg, acciter) != T2_UDFRet_OK) {
	   t2_err_ret(__FILE__, __LINE__, "acc.navigate returns ERROR");
	   delete iter;
	   return T2_UDFRet_ERROR;
	}
	if (acciter == NULL) {
	   t2_err_ret(__FILE__, __LINE__,
		      "acc.navigate returns NULL acc iterator");
	   delete iter;
	   return T2_UDFRet_ERROR;
	}
	else {

#	   if (T2_VERBOSE > 4)
	      u_int acc_cnt = 0;
#	   endif

	   T2_UDFRet rv2;
	   bool acc_eod = false;
           void *acc_elem = NULL;
           while ((rv2 = acciter->getNextElement(qinfo,acc,acc_eod,acc_elem))
		    == T2_UDFRet_OK)
	   {

#	      if (T2_VERBOSE > 4)
		 acc_cnt++;
		 if (acc_eod == true)
		    t2_say5(cout, "acc_iterator.>getNextElement is called the "
			           << acc_cnt
				   << "-th time and returns end-of-data=true"
				   << endl);
		 else
		    t2_say5(cout, "acc_iterator.>getNextElement is called the "
			           << acc_cnt
				   << "-th time and returns end-of-data=false"
				   << endl);
#	      endif

	      if (acc_eod == true)
		 break;				// all done

              if (dafElem(qinfo,dsid,input_elem,acc_elem) != T2_UDFRet_OK) {
		  t2_err_ret(__FILE__, __LINE__, "aggr.dafElem returns ERROR");
		  delete iter;
		  delete acciter;
		  return T2_UDFRet_ERROR;
	      }

              acc_eod = false;
	      acc_elem = NULL;
           }
           delete acciter;
	}			// if (acciter == NULL) else
     }				// if (intersects == true)

# if (T2_VERBOSE > 4)
     else {			// if (intersects == true) else
	t2_say5(cout, "proj.project sets intersection flag to false" << endl);
     }
# endif

     input_eod = false;
     input_elem = NULL;
     coord.setNumberDimensions(0);

   }				// while ((rv = iter->getNextElement() != OK)

# ifndef T2_PBACKEND
     if (rv != T2_UDFRet_OK) {
	// print some message that iter->getNextElement returns error
	t2_err_ret(__FILE__, __LINE__,
		"sbe:%d: ERROR: dataset iterator.getNextElement returns error",
		qinfo.getMyNode());
     }
# endif

# if (T2_VERBOSE > 4)
     t2_say5(cout, "leaving T2_AggregateFuncObj::dafAcc after input iterator steps through "
		   << input_cnt << " input elements" << endl);
# endif

  delete iter;
  return rv;
}

