
/* $Id: qenv.C,v 1.15 2000/03/16 17:56:34 chialin Exp $ */

/* qenv.C */


#include	<iostream.h>
#include	"t2_qenv.h"
#include	"t2_rts.h"

#ifdef T2_PBACKEND
  #include	"t2_work.h"
#endif

extern int _g_mynode;

////////////////////////////////////////////////////////////////////
//
// T2_DatasetEnv methods
//
////////////////////////////////////////////////////////////////////
T2_DatasetEnv::T2_DatasetEnv():
	dataset(NULL), iter_id(0), index_id(0), qregion(), prj(NULL), reqs()
{
}


T2_DatasetEnv::T2_DatasetEnv(const T2_DatasetEnv& e):
	dataset(e.dataset), iter_id(e.iter_id), index_id(e.index_id),
        qregion(e.qregion), prj(e.prj), reqs()
{
  t2_Assert (e.dataset == NULL);
  t2_Assert (e.prj == NULL);
  t2_Assert (e.reqs.getNumberClusterRequests());
}


T2_DatasetEnv::~T2_DatasetEnv()
{
  delete prj;
  // no need to delete *dataset, which is managed by T2_RTS
}


T2_DatasetEnv&
T2_DatasetEnv::operator =(const T2_DatasetEnv& e)
{
  t2_Assert (e.dataset == NULL);
  t2_Assert (e.prj == NULL);
  t2_Assert (e.reqs.getNumberClusterRequests() == 0);

  dataset = e.dataset;
  iter_id = e.iter_id;
  index_id = e.index_id;
  qregion.setNumberDimensions(e.qregion.getNumberDimensions());
  qregion = e.qregion;
  prj = e.prj;

  return *this;
}


bool
T2_DatasetEnv::init(T2_QSpecDataset& qspec, T2_RTS& rts)
{
  // first, create the dataset object
  ds_id = qspec.getDatasetID();			// set dataset id
  dataset = rts.activateDataset(ds_id);
  if (dataset == NULL)
     return false;
  iter_id = qspec.getIteratorID();		// set iterator id

  // now, remeber the index id
  index_id = qspec.getIndexID();


  // now store the query region
  qregion.setNumberDimensions(qspec.getQueryBox().getNumberDimensions());
  qregion = qspec.getQueryBox();


  // now deal with the projection function
  T2_ProjectFuncConstructorPtr prj_cons =
	T2_RTS::getPrjFuncConstructorPtr(qspec.getProjID());
  if (prj_cons) {
     T2_UsrArgReader reader(qspec.getProjConstructorArg());
     if (prj_cons(rts.getSystem(), reader, prj) != T2_UDFRet_OK) {
	t2_err_ret(__FILE__, __LINE__,
		  "projection function constructor failed");
	return false;				// constructor failed
     }
  }
  if (prj == NULL)
     return false;


  // nothing can be done to the cluster request list, reqs[], yet


  return (*this)? true : false;
}




////////////////////////////////////////////////////////////////////
//
// T2_OutputEnv methods
//
////////////////////////////////////////////////////////////////////
bool
T2_OutputEnv::init(T2_QEnv* p, T2_RTS& rts, T2_QSpecOutput& qspec)
{
  type = qspec.getOutputHandleType();

#ifndef T2_PBACKEND

  return true;			// do nothing else for sequential back-end

#else

  t2_Assert (outhandle == NULL);

  switch (type) {
    case t2_oUnknown:			// not knowing what to do
		return false;
		break;

    case t2_oBFile:			// output to a binary file
		outhandle = (T2_OutputHandle *) new T2_OHBinaryFile(p, qspec);
		t2_Assert (outhandle != NULL);
		if (outhandle->isValid() == false)
		   return false;
		break;

    case t2_oSocket:
		outhandle = (T2_OutputHandle *) new T2_OHSocket(p, rts, qspec);
		t2_Assert (outhandle != NULL);
		if (outhandle->isValid() == false)
		   return false;
		break;

#ifdef T2_METACHAOS
    case t2_oMC:
		outhandle = (T2_OutputHandle *)
				new T2_OHMetaChaos(p, rts, qspec);
		t2_Assert (outhandle != NULL);
		if (outhandle->isValid() == false)
		   return false;
		break;
#endif

    default:
		t2_err_ret(__FILE__, __LINE__,
			  "unrecognized output handle type %d",
			  (int) type);
		return false;
  }

  return true;

#endif	/* ! T2_PBACKEND */

}


bool
T2_OutputEnv::handleOutput(T2_Output& output,
			   T2_Accumulator& acc, T2_Iteration i)
{
#ifndef T2_PBACKEND

  return true;				// does nothing for sequential back-end

#else

  t2_Assert (outhandle != NULL);
  return outhandle->handleOutput(output, acc, i);

#endif	/* ! T2_PBACKEND */

}



////////////////////////////////////////////////////////////////////
//
// T2_QEnv methods
//
////////////////////////////////////////////////////////////////////
bool
T2_QEnv::init(u_int qid, T2_QSpec& qspec, T2_RTS& rts)
{
  valid = false;
  qenv_id = qid;

  // make qinfo points to system
  qinfo.setSystem(rts.getSystem());
  qinfo.setQuerySpec(qspec);

  // deal with the accumulator;
  // we need to construct the accumulator meta object and keep the
  // navigation function id
  T2_AccMetaConstructorPtr acc_cons =
			T2_RTS::getAccMetaConstructorPtr(qspec.getAccID());
  if (acc_cons) {
     T2_UsrArgReader reader(qspec.getAccConstructorArg());
     if (acc_cons(qinfo, qspec.getOutputBox(), reader, accmeta)!=T2_UDFRet_OK) {
	t2_err_ret(__FILE__, __LINE__,
		  "accumulator-meta constructor returns ERROR");
	return false;				// acc-meta constructor failed
     }
  }
  else
     accmeta = NULL;
  acc_navid = qspec.getAccNavigatorID();
  if (accmeta == NULL) {
     reportInitError(cerr, NO_ACCMETA, qspec);
     return false;
  }

  // now, let's deal with the aggregation function
  // we need to construct the aggregation function object
  T2_AggregateFuncConstructorPtr aggr_cons =
			T2_RTS::getAggrFuncConstructorPtr(qspec.getAggrID());
  if (aggr_cons) {
     T2_UsrArgReader reader(qspec.getAggrConstructorArg());
     if (aggr_cons(qinfo, reader, aggrf) != T2_UDFRet_OK) {
	t2_err_ret(__FILE__, __LINE__, "aggr constructor returns ERROR");
	return false;
     }
  }
  else
     aggrf = NULL;
  if (aggrf == NULL) {
     reportInitError(cerr, NO_AGGR, qspec);
     return false;
  }


  // now, let's deal with output handling
  if (outenv.init(this, rts, qspec.getOutputSpec()) == false) {
     reportInitError(cerr, NO_OUTHANDLE, qspec);
     return false;
  }

  // now, let's deal with datasets
  u_int n = qspec.getNumberDatasets();
  datasets.clear(n);
  {
     for (u_int i=0; i<n; i++) {
        if (datasets[i].init(qspec.getDatasetSpec(i), rts) == false) {
           reportInitError(cerr, NO_DATASET, qspec, i);
           return false;
	}
	datasets[i].setQEnv(this);
     }
  }


  // nothing to do with query windows


  // now, let's initialize the stuff used in the global-combine phase
  T2_System& system = rts.getSystem();
  const u_int numnodes = system.getNumberNodes();
  t2_Assert (accmsgsizes.getNumberElements() == 0);
  t2_Assert (accmsgsendbuf.getNumberElements() == 0);
  accmsgsizes.clear(numnodes);
  accmsgsendbuf.clear(numnodes);
  t2_Assert (accmsgrecvbuf[0] != NULL);
  t2_Assert (accmsgrecvbuf[1] != NULL);
  t2_Assert (accmsgrecvbuf[0]->getNumberProcessors() == 0);
  t2_Assert (accmsgrecvbuf[1]->getNumberProcessors() == 0);
  accmsgrecvbuf[0]->setNumberProcessors(numnodes);
  accmsgrecvbuf[1]->setNumberProcessors(numnodes);


  // initialization succeeded !!
  valid = true;
  return true;
}


ostream&
T2_QEnv::reportInitError(ostream& os, T2_QEnv::invalidKey k,
			 const T2_QSpec& qspec, u_int i) const
{
  switch (k) {
    case NO_ACCMETA :
	os << _g_mynode
	   << ": T2_QEnv::init cannot generate accumulator meta object w/"
	   << " constructor id = " << qspec.getAccID() << endl;
	return os;
	break;

    case NO_AGGR :
	os << _g_mynode
	   << ": T2_QEnv::init cannot generate aggregation function object w/"
	   << " constructor id = " << qspec.getAggrID() << endl;
	return os;
	break;

    case NO_DATASET :
	os << _g_mynode
	   << ": T2_QEnv::init cannot generate dataset object w/"
	   << " id = " << qspec.getDatasetSpec(i).getDatasetID() << endl;
	return os;
	break;

    case NO_OUTHANDLE :
	os << _g_mynode
	   << ": T2_QEnv::init cannot initialize the output handling "
	   << "from the query specification" << endl;
	return os;
	break;

    default :
	t2_err_quit(__FILE__, __LINE__, 
		   "T2_QEnv::init gets unexpected invalid key %d",
		   (int) k);
  }
  return os;
}


bool
T2_QEnv::allocateAccumulator(u_int iter)
{
  t2_Assert (accmeta != NULL);
  t2_Assert (acc == NULL);
  t2_Assert (iter < getNumberQueryWindows());

  // allocate new accumulator
  if (accmeta->allocAcc(qinfo,iter,getQueryWindow(iter),acc) != T2_UDFRet_OK)
  {
      t2_err_ret(__FILE__, __LINE__, "ERROR: acc-meta.allocAcc returns ERROR");
      t2_Assert ((acc = NULL) == NULL);
      return false;
  }

  t2_Assert (acc != NULL);


# if (T2_TIME_QUERY > 1)
  {
     T2_QQPlan* qqp = getQQPlan();
     t2_Assert (qqp != NULL);
     t2_Assert (qqp->getNumberSubplans() > 0);
     T2_QueryIterTimer& qtimer = qqp->getCurrentSubplan();
     double tsnap = t2_dclock();
# endif

  // initialize accumulator elements
  t2_say3(cout, "gonna call aggr.aifAcc" << endl);
  t2_Assert (aggrf != NULL);
  if (aggrf->aifAcc(qinfo, *acc) != T2_UDFRet_OK) {
     t2_err_ret(__FILE__, __LINE__, "aggr.aifAcc returns ERROR");
     return false;
  }

# if (T2_TIME_QUERY > 1)
     qtimer.incAifAccTime(t2_dclock() - tsnap);
  }
# endif

  return true;
}


void
T2_QEnv::finalizeOutput()
{
  t2_Assert (acc != NULL);
  t2_Assert (aggrf != NULL);

  delete output;


#if (T2_TIME_QUERY > 1)
{
  T2_QQPlan* qqp = getQQPlan();
  t2_Assert (qqp != NULL);
  t2_Assert (qqp->getNumberSubplans() > 0);
  T2_QueryIterTimer& qtimer = qqp->getCurrentSubplan();
  double tsnap = t2_dclock();
#endif

  if (aggrf->finalize(getQueryInfo(), *acc, output) != T2_UDFRet_OK) {
     t2_err_ret(__FILE__, __LINE__, "aggr.finalize returns ERROR");

     // not sure what to do now; for now, just delete output and set it to NULL
     delete output;
     output= NULL;
  }
  if (output == NULL) {
     t2_err_ret(__FILE__, __LINE__, "aggr.finalize returns NULL output");

     // not sure what to do now ...
  }

#if (T2_TIME_QUERY > 1)
  qtimer.incFinalizeTime(t2_dclock() - tsnap);
}
#endif

}


#ifdef T2_PBACKEND
bool
T2_QEnv::initIteration(T2_WorkList& worklist)
{
  // due to dependencies among the methods, the correct invocation order
  // is to call T2_QPlan::initBeforeExecution first, then for each query
  // (ie for each T2_QEnv) call T2_QEnv::initIteration at the beginning of the
  // iteration, and call finalizeIteration at the end of the iteration


  if (valid == false) {
     t2_say3(cout, "query " << (int) this << " is invalid" << endl);
     return false;
  }

  const T2_QQPlan *qqplan = getQQPlan();
  t2_Assert (qqplan != NULL);
  t2_Assert (acc == NULL);
  t2_Assert (output == NULL);

  if (qqplan->allIterationsDone() == true) {
     t2_say3(cout, "query " << (int) this
           	<< " qqplan says allIterationsDone is true" << endl);
     return false;
  }

  // get current iteration
  T2_Iteration iter = qqplan->getCurrentIteration();
  t2_Assert (iter < qqplan->getNumberSubplans());
  t2_Assert (qqplan->getCurrentCluster() == 0);


  t2_say3(cout, "******** initIteration() is initializing for iteration "
	       << iter << " for qenv " << (int) this << endl);


  // allocate accumulator first
  if (allocateAccumulator(iter) == false) {
     return false;			// allocate accumulator failed
  }


#ifdef T2_TIME_QUERY
  T2_QueryIterTimer& qtimer = getQQPlan()->getCurrentSubplan();
  qtimer.setStartReduction();
#endif


  // now set the current query status to local-reduction phase
  status.initLocalReduction(*qqplan);

  if (status.endLocalReduction(*qqplan) == true) {	// skip local-reduction
     t2_Assert (!worklist.isFull());
     worklist.enqueueGlobalCombine(this);
  }

  return true;				// a new iteration begins
}
#endif	/* T2_PBACKEND */


void
T2_QEnv::finalizeIteration()
{

  t2_say2(cout, "finalizeIteration is called for iteration "
		<< getQQPlan()->getCurrentIteration() << " for qenv "
		<< (int) this << endl);

  // delete accumulator and output
  deleteAccumulator();
  deleteOutput();
}


bool
T2_QEnv::deferAccMsgProcess(T2_ProcID src, T2_AccMsgRecvBuffer* msg)
{
  t2_Assert (msg != NULL);
  t2_Assert (accmsgrecvbuf[0] != NULL);
  t2_Assert (accmsgrecvbuf[1] != NULL);
  t2_Assert ((u_int) src < accmsgrecvbuf[0]->getNumberProcessors());

  // first check if any acc msg from node src has been received
  if (accmsgrecvbuf[0]->isSet(src) == true) {
     // already a msg from src exists, so this new msg must be for next iter.
     t2_Assert (accmsgrecvbuf[1]->isSet(src) == false);
     bool r = accmsgrecvbuf[1]->insertAccMsgRecvBuffer(src, msg);
     t2_Assert (r == true);
     t2_Assert (accmsgrecvbuf[1]->isSet(src) == true);

     t2_say4(cout, 
		" ______deferAccMsg ret TRUE for msg from " << src
		<< " cause bit is already set "
		<< accmsgrecvbuf[0]->getBitVector() << " , "
		<< accmsgrecvbuf[1]->getBitVector()
		<< endl);

     return true;					// delay msg processing
  }

  // no msg from src has been received yet;
  // if the query is in global-combine (not all acc msgs have been processed)
  // then this acc msg must be for this strip-mined iteration (ie tile),
  // so do not put it in accmsgrecvbuf and return false to ask the
  // communication manager put the acc msg in the work list;
  // a query is in state t2_qPostGlobal if it has already processed all its 
  // acc msgs but its T2_OutputHandleWork work entry (the entry whose handler
  // would officially advance this query into the output-handling phse)
  // is still waiting in the queue and has not been processed; for
  // a query that is in state t2_qPostGlobal, since all acc msgs for
  // the current iteration has been processed, this new acc msg must be
  // for the next iteration, so should buffer the acc msg in accmsgrecvbuf
  // so that it can be processed in the next iteration

  T2_QStatus& qstatus = getQueryStatus();
  if (qstatus.getState() == t2_qGlobal) {
     // already in global-combine and not all acc msgs have been processed
     accmsgrecvbuf[0]->setBitVector(src);
     t2_Assert (accmsgrecvbuf[0]->isSet(src) == true);
     t2_Assert (accmsgrecvbuf[1]->isSet(src) == false);


//     bool r = accmsgrecvbuf[1]->insertAccMsgRecvBuffer(src, msg);
//     t2_Assert (r == true);
//     t2_Assert (accmsgrecvbuf[1]->isSet(src) == true);


     t2_say4(cout, 
		"______deferAccMsg ret FALSE for msg from " << src
		<< " cause in global-combine "
		<< accmsgrecvbuf[0]->getBitVector() << " , "
		<< accmsgrecvbuf[1]->getBitVector()
		<< endl);

     return false;				// do not delay msg processing
  }

  bool r = accmsgrecvbuf[0]->insertAccMsgRecvBuffer(src, msg);
  t2_Assert (r == true);
  t2_Assert (accmsgrecvbuf[0]->isSet(src) == true);

  t2_say4(cout, 
		"______deferAccMsg ret TRUE for msg from " << src
		<< " cause not in global-combine, but in state "
		<< (int) qstatus.getState() << " "
		<< accmsgrecvbuf[0]->getBitVector() << " , "
		<< accmsgrecvbuf[1]->getBitVector()
		<< endl);

  return true;
}


void
T2_QEnv::finalizeGlobalCombine()
{
  accmsgrecvbuf[0]->clearAll();

  // swap accmsgrecvbuf[0] with accmsgrecvbuf[1]
  t2_Assert (accmsgrecvbuf[0] != NULL);
  t2_Assert (accmsgrecvbuf[1] != NULL);
  T2_AccMsgRecvBufList *p = accmsgrecvbuf[0];
  accmsgrecvbuf[0] = accmsgrecvbuf[1];
  accmsgrecvbuf[1] = p;

  // advance to post global-combine phase
  t2_Assert (getQueryStatus().getState() == t2_qGlobal);
  getQueryStatus().setState(t2_qPostGlobal);
}


#ifdef T2_PBACKEND
void
T2_QEnv::killQueryNow()
{
  t2_Assert (isValid() == true);
  t2_say2(cout, "killing query " << getQEnvID() << endl);

  // release acc msgs that arrive earlier and are queued in accmsgrecvbuf[2]
  for (u_int i=0; i<2; i++) {
     T2_AccMsgRecvBufList* abuf = accmsgrecvbuf[i];
     t2_Assert (abuf != NULL);

     const u_int nrecvbufs = abuf->getNumberMsgs();
     for (u_int j=0; j<nrecvbufs; j++) {
	T2_ProcID src;
	T2_AccMsgRecvBuffer* msg;

#	ifdef T2_ASSERTION
	   bool q =
#	endif
	abuf->getAccMsgRecvBuffer(j, src, msg);
	t2_Assert (q == true);

	delete msg;
     }
     abuf->clearAll();
  }

  // if query never reaches output-handling phase for the 1st strip-mined
  // iteration (ie tile), do some error handling for the output, if needed
  t2_Assert (getOutputEnv().getOutputHandle() != NULL);
  getOutputEnv().getOutputHandle()->executionErrorHandling(getQueryInfo());

  // proceed to kill query
  invalidate();			// invalidate query
  finalizeIteration();		// release memory, if any
  outenv.finalize();		// finalize output handling
  if (qqplan != NULL)
     qqplan->forceIterationsDone();	// no more iterations
  status.setNumberClustersYetProcess(0);
  status.setNumberAccMsgsYetProcess(0);
  status.setNumberAccSendsYetComplete(0);

  // update global status
  T2_QGlobalStatus& global = status.getGlobalStatus();
  t2_Assert (global.getNumberQueriesNotDone() > 0);
  global.decNumberQueriesNotDone();
}
#endif /* T2_PBACKEND */
