
/* $Id: rts.C,v 1.17 2000/03/16 17:56:34 chialin Exp $ */

/* rts.C */

#include	<iostream.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	"t2_rts.h"
#include	"t2_queryinfo.h"
#include	"t2_sockdefs.h"
#include	"t2_diskassign.h"

#ifdef T2_METACHAOS
  #include	"t2_outhandle.h"
#endif

extern int _g_mynode;


////////////////////////////////////////////////////////////////////
//
// T2_RTS static variables
//
////////////////////////////////////////////////////////////////////
T2_Array<T2_AccMetaConstructorPtr> T2_RTS::acc_cons(0);
T2_Array<T2_AggregateFuncConstructorPtr> T2_RTS::aggr_cons(0);
T2_Array<T2_ProjectFuncConstructorPtr> T2_RTS::prj_cons(0);
T2_Array<T2_DatasetConstructorPtr> T2_RTS::ds_cons(0);
T2_Array<T2_IndexConstructorPtr> T2_RTS::index_cons(0);

#ifdef T2_METACHAOS
T2_Array<T2_MetaChaosUserConstructorPtr> T2_RTS::mcu_cons(0);
#endif

T2_RTSDSEntry T2_RTSDSList::emptyEntry(0);


////////////////////////////////////////////////////////////////////
//
// some default T2 constructors
//
////////////////////////////////////////////////////////////////////
T2_UDFRet
t2_defAccMetaConstructor(const T2_QueryInfo& qinfo, const T2_Box& output_box,
			 T2_UsrArgReader& arg, T2_AccMetaObj*& amp)
{
  t2_err_ret(__FILE__, __LINE__, "t2_defAccMetaConstructor is invoked\n");
  amp = NULL;
  return T2_UDFRet_ERROR;
}

T2_UDFRet
t2_defAggrFuncConstructor(const T2_QueryInfo& qinfo, T2_UsrArgReader& arg,
			  T2_AggregateFuncObj*& afp)
{
  t2_err_ret(__FILE__, __LINE__, "t2_defAggrFuncConstructor is invoked\n");
  afp = NULL;
  return T2_UDFRet_ERROR;
}

T2_UDFRet
t2_defPrjFuncConstructor(const T2_System& system, T2_UsrArgReader& arg,
			 T2_ProjectFuncObj*& prj)
{
  t2_err_ret(__FILE__, __LINE__, "t2_defPrjFuncConstructor is invoked\n");
  prj = NULL;
  return T2_UDFRet_ERROR;
}

T2_UDFRet
t2_defDatasetConstructor(const T2_System& system,
		         const T2_Array<int>& dfd, const T2_Array<int>& mfd,
			 T2_Dataset*& dsp)
{
  t2_err_ret(__FILE__, __LINE__, "t2_defDatasetConstructor is invoked\n");
  dsp = NULL;
  return T2_UDFRet_ERROR;
}

T2_UDFRet
t2_defIndexConstructor(const T2_System& system, T2_UsrArgReader& arg,
		       const T2_Array<int>& fd, T2_Index*& idxp)
{
  t2_err_ret(__FILE__, __LINE__, "t2_defIndexConstructor is invoked\n");
  idxp = NULL;
  return T2_UDFRet_ERROR;
}


#ifdef T2_METACHAOS
T2_UDFRet t2_defMetaChaosUserConstructor(const T2_System& system,
			T2_Output& output, T2_Accumulator& acc,
			MC_Server& svr, MC_Client& clt,
			T2_MetaChaosUser*& mcu_cons)
{
  t2_err_ret(__FILE__, __LINE__, "t2_defMetaChaosUserConstructor is invoked\n");
  mcu_cons = NULL;
  return T2_UDFRet_ERROR;
}
#endif



////////////////////////////////////////////////////////////////////
//
// T2_RTSDSEntry methods
//
////////////////////////////////////////////////////////////////////
void
T2_RTSDSEntry::kill()
{
  id = 0;
  delete dataset;
  dataset = NULL;
  deactivate();
  closeDataFiles();
  map.emptyMap();
  my_datafile_id.setNumberBits(0);
}


T2_RTSDSEntry::T2_RTSDSEntry(const T2_RTSDSEntry& e)
			: my_datafile_id(e.my_datafile_id)
{
  t2_Assert (e.dataset == NULL);
  id = e.id;
  dataset = e.dataset;
  map = e.map;
  activated = e.activated;
}


T2_RTSDSEntry&
T2_RTSDSEntry::operator =(const T2_RTSDSEntry& e)
{
  t2_Assert (e.dataset == NULL);
  t2_Assert (e.map.getNumberElements() == 0);
  id = e.id;
  dataset = e.dataset;
  map = e.map;
  my_datafile_id.assign(e.my_datafile_id);
  activated = e.activated;
  return *this;
}




////////////////////////////////////////////////////////////////////
//
// T2_RTSDSList methods
//
////////////////////////////////////////////////////////////////////
T2_RTSDSList::T2_RTSDSList(u_int l): list(l)
{
}


int
T2_RTSDSList::search(T2_DSID i) const
{
  for (u_int j=0; j<getNumberEntries(); j++) {
     if (list[j].getDatasetID() == i)	// found it !!
	return j;
  }

  return -1;				// cannot find it
}


int
T2_RTSDSList::searchInactivated() const
{
  for (u_int j=0; j<getNumberEntries(); j++) {
     if (!list[j].isActivated())	// found it !!
	return j;
  }

  return -1;				// cannot find it
}


T2_Dataset*
T2_RTSDSList::activateDataset(T2_DSID i, T2_DSCatalog& dscatalog,
			      const T2_System& system)
{
  int k = search(i);
  if (k >= 0) {				// found dataset with id = i
     list[k].activate();
     t2_Assert (list[k].getDatasetPtr() != NULL);
     return list[k].getDatasetPtr();
  }

  // dataset i is not found in list, need to create an entry for dataset i

  // read dataset info from catalog first
  T2_BitVector bits;			// bits specify local data files to read
  T2_DSCatEntry e;
  if (!dscatalog || dscatalog.getDSCatalogEntry(i, e, bits) == false) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot read info from catalog for dataset %d.", i);
     return NULL;
  }
  t2_Assert (e.getDatasetID() == i);

# if (T2_ASSERTION > 1)
    cout<<system.getMyNode() << ":" << "bits = ";
    for (u_int xx=0; xx<bits.getNumberBits(); xx++)
	if (bits.isSet(xx) == true)
	   cout << " " << xx;
    cout << " (out of " << bits.getNumberBits() << " data files)" << endl;
# endif

  // get dataset constructor
  T2_DatasetConstructorPtr dscons
		= T2_RTS::getDatasetConstructorPtr(e.getConstructorID());
  if (dscons == NULL) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: dataset constructor function with id %d for dataset %d is NULL",
	 e.getConstructorID(), i);
     return NULL;
  }

  const u_int ndf = e.getNumberDataFiles(),
              nmf = e.getNumberMetaFiles();

  u_int j;				// loop index

  // open all meta files
  T2_Array<int> metafd(nmf);
  for (j=0; j<nmf; j++)	{
     const u_int num_tries = 10;
     u_int nnn;
     for (nnn=0; nnn<num_tries; nnn++) {
	if ((metafd[j] = open(e.getMetaFileName(j), O_RDONLY)) > 0)
	   break;

	t2_say3(cout,
	    "WARNING: T2_RTSDSList cannot open dataset meta file "
	     << e.getMetaFileName(j) << ", will retry." << endl);
	usleep(1000);
     }

     if (nnn == num_tries) {		// all read attempts failed
	t2_err_ret(__FILE__, __LINE__,
	  "ERROR: T2_RTSDSList cannot open dataset meta file %s after %d tries",
	   e.getMetaFileName(j), num_tries);
	for (u_int jj=0; jj<j; jj++)
           close(metafd[jj]);
	return NULL;
     }

#    ifndef T2_PBACKEND
	t2_say2(cout, system.getMyNode() << ":");
#    endif
     t2_say2(cout, "RTS open " << e.getMetaFileName(j) 
		  << " (meta datafile " << j << ",fd=" << metafd[j]
		  << ") for dataset " << i << endl);
  }

  // open all data files assigned to local processor
  T2_Array<int> datafd(ndf);
  T2_Array<int> diskno(ndf);
  for (j=0; j<ndf; j++) {		// open data files
     if (bits.isSet(j)) {		// this is my local data file
	const u_int num_tries = 10;
	u_int nnn;
	for (nnn=0; nnn<num_tries; nnn++) {
	   if ((datafd[j] = open(e.getDataFileName(j), O_RDONLY)) > 0)
	      break;

	   t2_say3(cout,
		"WARNING: T2_RTSDSList cannot open dataset data file "
		 << e.getDataFileName(j) << ", will retry" << endl);
	   usleep(1000);
	}
	
	if (nnn == num_tries) {		// all read attempts failed
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: T2_RTSDSList cannot open dataset data file %s after %d tries",
		e.getDataFileName(j), num_tries);
	   u_int jj;
	   for (jj=0; jj<j; jj++)
	      close(datafd[jj]);
	   for (jj=0; jj<nmf; jj++)
	      close(metafd[jj]);
	   return NULL;
	}

#	ifndef T2_PBACKEND
	   t2_say2(cout, system.getMyNode() << ":");
#	endif
	t2_say2(cout, "RTS open " << e.getDataFileName(j) 
		     << " (datafile " << j << ", fd=" << datafd[j]
		     << ") for dataset " << i << endl);

	diskno[j] = e.getDataFileDiskNo(j);
     }
  }

  // insert a new entry into list for dataset i

  if (list.isFull()) {			// list full, need to remove an entry
     k = searchInactivated();
     if (k < 0) {
	t2_err_quit(__FILE__,__LINE__,
		    "%d: T2_RTSDSList is full, and all entry are activated",
		    _g_mynode);
	return NULL;
     }
     list[k].kill();			// entry k is to be evicted to make room
  }
  else {				// list is not full yet
     k = list.getNumberElements();
     list << T2_RTSDSEntry();
  }

  T2_RTSDSEntry& newentry = list[k];	// this is our new entry
  newentry.getDatasetID() = i;
  newentry.activate();
  newentry.getDSFileIDBitVector().assign(bits);

  t2_Assert (newentry.getDatasetPtr() == NULL);
  t2_Assert (newentry.getDSFileMap().isEmpty());

  for (j=0; j<ndf; j++) {
     if (bits.isSet(j))
	newentry.getDSFileMap().insert(j, datafd[j], diskno[j]);
  }

  // we finally are ready to create the dataset object
  if (dscons(system, datafd, metafd, newentry.getDatasetPtr()) != T2_UDFRet_OK)
  {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: constructor for dataset %d returns ERROR", i);
     newentry.kill();
     list.out(k);
     newentry.getDatasetPtr() = NULL;
  }
  else if (newentry.getDatasetPtr() == NULL) {
     // dataset constructor returns NULL !!
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: in T2_RTSDSList, constructor for dataset %d returns NULL",
	    i);
     newentry.kill();
     list.out(k);
  }

  for (j=0; j<nmf; j++)			// close meta data files
     close(metafd[j]);

  return newentry.getDatasetPtr();
}




////////////////////////////////////////////////////////////////////
//
// T2_RTS methods
//
////////////////////////////////////////////////////////////////////
T2_RTS::T2_RTS():
	    dscatalog(), dslist(), indexcatalog(),
	    initflag(false), system()
{
}


T2_RTS::~T2_RTS()
{
#ifdef T2_METACHAOS
  // initialize MC_Server
  T2_OHMetaChaos::exitMCServer();
#endif
}


T2_RTS&
T2_RTS::operator =(const T2_RTS& r)
{
  t2_Assert (!r.dscatalog);
  t2_Assert (!r.indexcatalog);
  t2_Assert (r.dslist.getNumberEntries() == 0);

  initflag = r.initflag;
  system = r.system;

  return *this;
}


bool
T2_RTS::initHostName(const char* datacatalog, const char* idxcatalog,
		     const char* hostname, const char* mc_pname)
{
  initflag = false;
  t2_Assert (datacatalog != NULL);
  t2_Assert (idxcatalog != NULL);
  t2_Assert (hostname != NULL);

  dscatalog.openHostName(datacatalog, hostname);
  if (!dscatalog)
     return false;

  indexcatalog.open(idxcatalog);
  if (!indexcatalog)
     return false;

# ifdef T2_METACHAOS
     // initialize MC_Server
     t2_Assert (mc_pname != NULL);
     T2_OHMetaChaos::initMCServer(mc_pname, getSystem().getNumberNodes(),
				  getSystem().getMyNode());
     getSystem().setMCProgramName(mc_pname);
# endif

  initflag = true;		// initialization succeeded !!
  return true;
}


bool
T2_RTS::initDiskAssignFile(const char* datacatalog, const char* idxcatalog,
			   const char* diskassign_file, const char* hostname,
			   const char* mc_pname)
{
  initflag = false;
  t2_Assert (datacatalog != NULL);
  t2_Assert (idxcatalog != NULL);
  t2_Assert (diskassign_file != NULL);
  t2_Assert (hostname != NULL);

  T2_DiskAssignmentFile diskassign;
  diskassign.open(hostname, getMyNode(), diskassign_file);
  if (!diskassign)  		// open failed
     return false;

  switch (diskassign.getAssignmentType()) {
     case T2_DiskAssignmentFile::assign_error:    // error
	t2_err_quit(__FILE__, __LINE__, "ERROR: error assignment type");
	break;
     case T2_DiskAssignmentFile::assign_disk_ids: // overwrite disk assignmt.
	dscatalog.openDiskID(datacatalog, diskassign.getDiskIDAssignment());
	break;
     case T2_DiskAssignmentFile::assign_proc_ids: // overwrite proc assignmt.
	dscatalog.openProcID(datacatalog, diskassign.getProcIDAssignment());
	break;
  }
  diskassign.close();

  if (!dscatalog)
     return false;

  indexcatalog.open(idxcatalog);
  if (!indexcatalog)
     return false;

# ifdef T2_METACHAOS
     // initialize MC_Server
     t2_Assert (mc_pname != NULL);
     T2_OHMetaChaos::initMCServer(mc_pname, getSystem().getNumberNodes(),
				  getSystem().getMyNode());
     getSystem().setMCProgramName(mc_pname);
# endif

  initflag = true;		// initialization succeeded !!
  return true;
}


bool
T2_RTS::initRank(const char* datacatalog, const char* idxcatalog,
		 const char* mc_pname)
{
  initflag = false;
  t2_Assert (datacatalog != NULL);
  t2_Assert (idxcatalog != NULL);

  // need to peek into data catalog to find out the number of logical
  // processor ids
  T2_TxtFile hdr('#', 2048);
  hdr.open(datacatalog, true);
  if (!hdr) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open data catalog file '%s'.", datacatalog);
     return false;
  }
  u_int num_procids, num_diskids;		// numbers of logical ids
  if (T2_DSCatalogHeader::readHeaderNumberIDs(hdr, num_procids, num_diskids)
      == false)
  {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: fail to read number of logical processor/disk ids from data catalog file '%s'.",
		datacatalog);
     return false;
  }
  hdr.close();

  // set up a bit vector that marks the logical processor ids assigned to
  // the local processor
  const u_int numnodes = getNumberNodes();
  const T2_ProcID mynode = getMyNode();
  t2_Assert(mynode >= 0);
  T2_BitVector my_procids(num_procids);
  for (u_int i=0; i<num_procids; i++)
     if (i % numnodes == (u_int) mynode)
	my_procids.set(i);
     else
	my_procids.clr(i);

# if (T2_ASSERTION > 1)
     cout << mynode << ": my_procids = ";
     for (u_int xx=0; xx<my_procids.getNumberBits(); xx++)
	if (my_procids.isSet(xx) == true)
	    cout << " " << xx;
     cout << "	(out of " << my_procids.getNumberBits() << " proc ids)" << endl;
# endif

  dscatalog.openProcID(datacatalog, my_procids);
  if (!dscatalog)
     return false;

# if (T2_ASSERTION > 1)
     cout << mynode << ": my_diskbitmap = ";
     for (u_int xx=0; xx<dscatalog.getDiskBitMap().getNumberBits(); xx++)
	if (dscatalog.getDiskBitMap().isSet(xx) == true)
	   cout << " " << xx;
     cout << "	(out of " << dscatalog.getDiskBitMap().getNumberBits() 
	  << " disk ids)" << endl;
# endif

  indexcatalog.open(idxcatalog);
  if (!indexcatalog)
     return false;

# ifdef T2_METACHAOS
     // initialize MC_Server
     t2_Assert (mc_pname != NULL);
     T2_OHMetaChaos::initMCServer(mc_pname, getSystem().getNumberNodes(),
				  getSystem().getMyNode());
     getSystem().setMCProgramName(mc_pname);
# endif

  initflag = true;		// initialization succeeded !!
  return true;
}


void
T2_RTS::initStaticArrays()
{
  void initAccMetaConstructors(T2_Array<T2_AccMetaConstructorPtr>&),
       initAggregateFuncConstructors(T2_Array<T2_AggregateFuncConstructorPtr>&),
       initProjectFuncConstructors(T2_Array<T2_ProjectFuncConstructorPtr>&),
       initDatasetConstructors(T2_Array<T2_DatasetConstructorPtr>&),
       initIndexConstructors(T2_Array<T2_IndexConstructorPtr>&);

  initAccMetaConstructors(acc_cons);
  initAggregateFuncConstructors(aggr_cons);
  initProjectFuncConstructors(prj_cons);
  initDatasetConstructors(ds_cons);
  initIndexConstructors(index_cons);

#ifdef T2_METACHAOS
  void initMetaChaosUserConstructors(T2_Array<T2_MetaChaosUserConstructorPtr>&);
  initMetaChaosUserConstructors(mcu_cons);
#endif

}


ostream&
T2_RTS::write(ostream& os)
{
  if (initflag == false) {
     os << "Not initialized yet!" << endl;
     return os;
  }

  os << "numnodese = " << getNumberNodes() 
     << "  mynode = " << getMyNode()
     << endl
     << "dataset catalog:" << endl
     << dscatalog << endl
     << "index catalog:" << endl
     << indexcatalog << endl;
  return os;
}

