
/* $Id: sproc.C,v 1.12 2000/03/16 17:56:34 chialin Exp $ */

/* sproc.C */

#include	<iostream.h>
#include	<stdio.h>
#include	<unistd.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_socket.h"
#include	"t2_packets.h"
#include	"t2_bfile.h"
#include	"t2_qspec.h"
#include	"t2_rts.h"
#include	"t2_qenv.h"
#include	"t2_blockreq.h"
#include	"t2_clusterreq.h"
#include	"t2_cluster.h"
#include	"t2_qplan.h"
#include	"t2_clusterinfo.h"


// forward declaration
bool
gen_requests(T2_QSpec& qspec,
	     T2_IndexCatalog& indexcat,
	     T2_RTSDSList& dslist,
	     T2_Array<T2_DatasetEnv>& datasetenv,
	     T2_BlockRequestListMap& blockreqmap,
	     const T2_System& system, bool verbose = false);
u_int
add_disk_requests_to_list(T2_VArray<T2_BlockRequest>& reqs, T2_DSFileMap& dfmap,
			  T2_ClusterRequestList& clusterreq,
			  T2_PhyBlockRequestList& blockreq,
			  const T2_Box& mbr, T2_UsrArg& cinfo_buf);
bool
strip_mine(T2_QQPlan& qqplan, T2_QEnv& qenv, const size_t mem,
	   T2_System& system);

void
execute_all_qplans(T2_Array<T2_QPlan>& qplan, T2_Array<T2_QEnv>& qenv,
		const char* ofname, const T2_Array<T2_RTS>& rts, bool verbose,
		void* stuff);
void
execute_qplan_init_acc(T2_Iteration iter, T2_QPlan& qplan, T2_QEnv& qenv);

void
execute_qplan_local_reduction(T2_Iteration iter, T2_QPlan& qplan,
			     const T2_System& system, const T2_Region& tilereg);
void
execute_all_qplans_global_combine(T2_Array<T2_QPlan>& qplan,
			 T2_Array<T2_QEnv>& qenv, const T2_Array<T2_RTS>& rts,
			 bool verbose);
void
execute_qplan_global_combine_send(T2_QPlan& qplan, T2_QEnv& qenv,
			          const T2_System& system,
				  T2_Array<T2_AccMsgSendBuffer>& msgbuf,
				  bool verbose);
void
execute_qplan_output_handling(T2_Iteration iter, T2_QEnv& qenv,
			      const char* ofname, const T2_System& system,
			      bool verbose, void *stuff);
void
execute_qplan_cleanup(T2_QEnv& qenv);


size_t
issue_read(T2_PhyBlockRequest& block);
//void
//process_input_data_chunk(T2_DatasetEnv& dsenv, const T2_Cluster& input,
//			 const T2_Region& oqw, T2_ClusterAuxInfoReader& meta,
//			 const T2_Box& mbr, const T2_System& system);
void
process_input_data_chunk(
//T2_DatasetEnv& dsenv, 
T2_ClusterRequest* cluster, const T2_Region& tilereg);
//const T2_Cluster& input,
//const T2_Region& oqw, 
//T2_ClusterAuxInfoReader& meta);

void
handle_output(T2_QEnv& qenv, const char* ofname, void* stuff);

void
handle_output_by_socket(T2_QEnv& qenv, T2_Output& output, T2_Socket& sock);
void
handle_output_by_qenv(T2_QEnv& qenv, void* stuff);
void
handle_output_write_file(T2_QEnv& qenv, const char* ofname);


/////////////////////////////////////////////////////////////////////
//
// some help functions
//
/////////////////////////////////////////////////////////////////////
inline void
closeFiles(u_int s, u_int e, const T2_Array<int>& fd)
{
  for (u_int j=s; j<=e && j<fd.getNumberElements(); j++) {
     if (fd[j] >= 0)
        close(fd[j]);
  }
}


/////////////////////////////////////////////////////////////////////
//
// initializing query environment
//
/////////////////////////////////////////////////////////////////////
inline bool
init_qenvs(T2_Array<T2_QEnv>& qenv, T2_Array<T2_RTS>& rts, T2_QSpec& qspec)
{
  t2_Assert (qenv.getNumberElements() == rts.getNumberElements());

  const u_int nqry = qenv.getNumberElements();
  for (u_int i=0; i<nqry; i++) {
     if (qenv[i].init(i, qspec, rts[i]) == false)	// qenv[i] init failed
	return false;
  }

  return true;					// all qenv's init succeeded
}


/////////////////////////////////////////////////////////////////////
//
// generating disk requests for all qenvs
//
/////////////////////////////////////////////////////////////////////
bool
gen_all_requests(T2_QSpec& qspec, T2_Array<T2_RTS>& rts,
		 T2_Array<T2_QEnv>& qenv, T2_Array<T2_QPlan>& qplan,
		 bool verbose)
{
  t2_Assert (rts.getNumberElements() == qenv.getNumberElements());
  t2_Assert (rts.getNumberElements() == qplan.getNumberElements());

  const u_int nqry = qenv.getNumberElements();
  for (u_int i=0; i<nqry; i++) {
     bool r =
	gen_requests(qspec,
		rts[i].getIndexCatalog(),
		rts[i].getRTSDatasetInfo(),
		qenv[i].getDatasetEnvList(),	// actually just need clusterreq
		qplan[i].getBlockRequestListMap(),
		rts[i].getSystem(), verbose);

     if (r == false)				// gen_requests failed
	return false;

     qplan[i].initBlockRequests();
  }

  return true;
}


/////////////////////////////////////////////////////////////////////
//
// strip-mining all queries
//
/////////////////////////////////////////////////////////////////////
void
strip_mine_all(T2_Array<T2_RTS>& rts, T2_Array<T2_QPlan>& qplan,
	       T2_Array<T2_QEnv>& qenv, size_t mem)
{
  t2_Assert (qplan.getNumberElements() == qenv.getNumberElements());

  const u_int nqry = qenv.getNumberElements();
  for (u_int i=0; i<nqry; i++) {
     // assure only one query per node
     t2_Assert (qplan[i].getNumberQQPlans() == 1);

     if (qenv[i]) {
        if (strip_mine(qplan[i].getQQPlan(0), qenv[i], mem, rts[i].getSystem())
	    == false)
	{
	   // strip_mine failed
	   qenv[i].invalidate();
	}
     }
     else
	t2_Assert (qplan[i].getQQPlan(0).getNumberSubplans() == 0);
  }
}


/////////////////////////////////////////////////////////////////////
//
// query processing function
//
/////////////////////////////////////////////////////////////////////
void query_processing(T2_Array<T2_RTS>& rts, T2_QSpec& qspec, size_t mem,
		      const char* ofname, bool verbose)
{
  const u_int numnodes = rts.getNumberElements();

  // initialize query environments for the given query
  t2_say2(cout, "sbe: initializing query environments" << endl);
  T2_Array<T2_QEnv> qenv(numnodes);
  if (init_qenvs(qenv, rts, qspec) == false)	// qenv init failed
     return;					// do nothing

  // initialize query plans
  t2_say2(cout, "sbe: initializing query plans" << endl);
  T2_Array<T2_QPlan> qplan(numnodes);
  for (u_int i=0; i<numnodes; i++)
     qplan[i].setNumberQQPlans(1);		// only one query per node

  // generate disk requests from qenv
  if (verbose) cout << "sbe: generating disk requests for queries" << endl;
  if (gen_all_requests(qspec, rts, qenv, qplan, verbose) == false)
     return;

  // strip-mining
  if (verbose) cout << "sbe: performing strip-mining" << endl;
  strip_mine_all(rts, qplan, qenv, mem);


  // check if all back-end nodes come up with the same number of 
  // strip-mined iterations (ie output tiles);
  // for now, just invalidate the query if not
  {
     u_int ntiles = 0;
     for (u_int q=0; q<numnodes; q++) {
	if (qenv[q].isValid() == false)		// invalid query
	   return;
	if (q == 0) {
	   ntiles = qplan[q].getQQPlan(0).getNumberSubplans();
	   continue;
	}
	if (ntiles != qplan[q].getQQPlan(0).getNumberSubplans()) {
	   cout << "sbe: back-end node 0 generates " << ntiles 
		<< " strip-mined iterations, but node " << q
		<< " has " << qplan[q].getQQPlan(0).getNumberSubplans()
		<< " iterations; invalid query execution."
		<< endl;
	   return;
	}
     }
  }

  // initialize output handling
  void *stuff = NULL;				// output handling stuff
  const T2_QSpecOutput& outspec = qspec.getOutputSpec();

  if (ofname == NULL) {
     switch (outspec.getOutputHandleType()) {
	case t2_oUnknown:
		t2_err_ret(__FILE__, __LINE__,
			   "don't know what to do with the output");
		return;

	case t2_oBFile:				// write to files
		t2_Assert (outspec.getFilePrefix() != NULL);
		ofname = outspec.getFilePrefix();
		break;

	case t2_oSocket:			// write to sockets
		t2_Assert (outspec.getHostName() != NULL);
		t2_say(cout, "sbe: connecting to " << outspec.getHostName()
			     << " port " << outspec.getPortNumber());
		{
		  T2_Socket* sock = new T2_Socket();
		  bool r = sock->connectByName(outspec.getHostName(),
					       outspec.getPortNumber());
		  if (r == false) {
		     errno = sock->getErrorValue();
		     t2_err_ret(__FILE__, __LINE__,
		     "cannot connect to output client %s at port %d, error %d)",
				outspec.getHostName(), outspec.getPortNumber(),
				sock->getErrorValue());
		     delete sock;
		     return;
		  }

		  if (outspec.useT2Protocol() == true) {
		     (*sock) << T2_BEClientHelloPacket(1, 0);

		     if (sock->getErrorValue() > 0) {
		        t2_err_ret(__FILE__, __LINE__,
				   "write T2_BEClientHelloPacket error");
		        sock->closeSocket();
		        delete sock;
		        return;
		     }
		  }

		  stuff = (void *) sock;
		}
		break;

	case t2_oMC:				// write to meta-chaos
		t2_err_ret(__FILE__, __LINE__,
			  "this version does not support Meta-Chaos interface");
		return;

	default:
		t2_err_ret(__FILE__, __LINE__,
			  "unrecognized output handling type = %d", 
			  outspec.getOutputHandleType());
		return;
     }
  }


  // sort the subplans here, if decided to


  // now execute the query plan
  execute_all_qplans(qplan, qenv, ofname, rts, verbose, stuff);

  // clean up output-handling stuff
  if (stuff != NULL) {
     switch (outspec.getOutputHandleType()) {
	case t2_oSocket: {
		  T2_Socket* sock = (T2_Socket *) stuff;
		  if (outspec.useT2Protocol() == true) {
		     T2_BEClientByePacket bye;
		     bye.send(*sock);
		  }
		  sock->closeSocket();
		  delete sock;
		}
		break;

	default:
		t2_err_ret(__FILE__, __LINE__,
			  "stuff should be NULL when output handling type = %d",
			  outspec.getOutputHandleType());
     }
  }
}


/////////////////////////////////////////////////////////////////////
//
// generating disk requests
//
/////////////////////////////////////////////////////////////////////
bool
gen_requests(T2_QSpec& qspec,
	     T2_IndexCatalog& indexcat,
	     T2_RTSDSList& dslist,
	     T2_Array<T2_DatasetEnv>& datasetenv,
	     T2_BlockRequestListMap& blockreqmap,
	     const T2_System& system, bool verbose)

{
  const u_int nds = qspec.getNumberDatasets();
  const u_int numnodes = system.getNumberNodes();
  const T2_ProcID mynode = system.getMyNode();

  for (u_int i=0; i<nds; i++) {		// step through each dataset
     u_int j;				// loop index
     T2_UDFRet rv = T2_UDFRet_OK;

     // identify all relevant objects
     T2_QSpecDataset& dsspec = qspec.getDatasetSpec(i);
     T2_DSID ds_id = dsspec.getDatasetID();
     T2_IndexID index_id = dsspec.getIndexID();
     T2_DatasetEnv& dsenv = datasetenv[i];
     T2_ClusterRequestList& clusterreq = dsenv.getClusterRequestList();
     T2_PhyBlockRequestList& blockreq =
				blockreqmap.touchBlockRequestList(ds_id);

     // get the dataset entry from rts
     bool flag;
     T2_RTSDSEntry& dsentry = dslist.search(ds_id, flag);
     if (flag == false) {
        t2_err_ret(__FILE__, __LINE__,
	   "ERROR: cannot find in T2_RTSDSList dataset with id = %d", ds_id);
	return false;
     }

     T2_DSFileMap& dfmap = dsentry.getDSFileMap();
     const u_int ndsfiles = dfmap.getNumberElements();
     if (ndsfiles == 0) {		// no assigned data files
	// assume no local data 
	continue;
     }

     // get the index entry from the index catalog
     T2_IndexCatEntry index_catentry;
     if (indexcat.getIndexCatalogEntry(index_id, index_catentry) == false) {
        t2_err_ret(__FILE__, __LINE__,
		  "cannot read from catalog info for index %d", index_id);
        return false;
     }

     // set the bitmap that marks the index files required to support the
     // assigned data files for the given dataset
     T2_BitVector my_idxfiles(0);
     u_int n_my_idxfiles = 
	index_catentry.support(dsentry.getDSFileIDBitVector(), my_idxfiles);
     if (n_my_idxfiles == 0) {		// no index file !!
	t2_err_ret(__FILE__, __LINE__,
	   "ERROR: no assigned index file while there is assigned data files.");
	return false;
     }

     t2_Assert (clusterreq.getNumberClusterRequests() == 0);

     // open index files;
     // fd[i] would store the file descriptior of the i-th index file
     // needed to support all local data files, and 
     // idx_fileid[i] would store the position of the i-th index file
     // as in the order all index files are stored in the index catalog
     // (ie, idx_fileid[2] = 5 means the 5th index file listed in the index
     // catalog is the 2nd index file needed to support the local data files)
     const u_int nfiles = index_catentry.getNumberFiles();
     T2_Array<int> fd(n_my_idxfiles);
     T2_Array<int> idx_fileid(n_my_idxfiles);
     u_int rj;
     for (j=0, rj=0; j<nfiles; j++) {
	if (my_idxfiles.isSet(j) == true) {
	  if ((fd[rj] = open(index_catentry.getFileName(j), O_RDONLY)) < 0) {
	     t2_err_ret(__FILE__,__LINE__,
			"cannot open index file %s",
			index_catentry.getFileName(j));
	     closeFiles(0, rj-1, fd);
	     return false;
	  }
	  idx_fileid[rj] = j;
	  rj++;
	}
     }

     // pointer index object
     T2_Index *index = NULL;

     // find the file descriptors for the data files of the dataset
     {
        // construct the index from its constructor
        T2_IndexConstructorPtr index_cons =
	    T2_RTS::getIndexConstructorPtr(index_catentry.getConstructorID());
        if (index_cons) {
	     T2_UsrArgReader reader(dsspec.getIndexConstructorArg());
	     rv = index_cons(system, reader, fd, index);
        }
     }
     if (index == NULL) {
	  t2_err_ret(__FILE__, __LINE__,
	     "ERROR: index constructor for index %d returns NULL index pointer",
	     index_id);
	  closeFiles(0, nfiles, fd);
	  return false;
     }
     if (rv == T2_UDFRet_ERROR) {
	  t2_err_ret(__FILE__, __LINE__,
		    "index constructor for index %d returns ERROR", index_id);
	  closeFiles(0, nfiles, fd);
	  return false;
     }


     // now generate the list of disk requests

     t2_say2(cout, "sbe:" << mynode
		    << ": gonna call index.fetchInit()" << endl);

     // first let the index initialize itself before search
     const T2_Box& qbox = dsspec.getQueryBox();
     index->fetchInit(system, qbox);

     // about ready to do an index lookup to find all logical chunks that
     // intersect the query box
     u_int wc = 0;				// just a counter
     T2_UsrArg cinfo_buf;			// buffer for aux cluster info
     T2_ClusterAuxInfoWriter cinfo(cinfo_buf);
     bool eod = true;				// end-of-data flag
     T2_Box mbr;				// an empty bounding box
     T2_VArray<T2_BlockRequest> reqs(20);	// physical block requests

     t2_say2(cout, "sbe:" << mynode << ": gonna call index.fetch()" << endl);

     // now retrieve the logical chunks and their corresponding physical
     // chunk requests
     while ((rv = index->fetch(system, qbox, eod, mbr, cinfo, reqs))
	     == T2_UDFRet_OK)
     {
	if (eod == true)
	   break;				// all done

	wc++;
	cinfo.close();
	const u_int nn = reqs.getNumberElements();

	t2_say5(cout, "sbe:" << mynode
			<< ": index fetch() called " << wc 
			<< "-th time, returns "
			<< nn << " block requests" << endl);

	if (nn == 0) {			// no chunk request !!
	   t2_err_ret(__FILE__, __LINE__,
		     "No chunk request was generated when index fetch() called %d-th time",
		     wc);
	   eod = true;
	   mbr.setNumberDimensions(0);
	   cinfo_buf.releaseBuffer();	// remove any buffer allocated by fetch
	   cinfo.open(cinfo_buf);	// must rebind cinfo with cinfo_buf
	   reqs.clear();
	   continue;
	}

	// for now, I am assuming that for a given logical chunk id, either
	// all its actual chunk requests go to local disks, or none of them do;
	// ie, no joins between processors can happen;
	// this makes scheduling much easier  :)

	// check the first chunk request to see if local reads are needed
	int fd_t;
	u_int diskno_t;
	if (dfmap.getFileInfo(reqs[0].getFileID(), fd_t, diskno_t) == false) {
	   eod = true;
	   mbr.setNumberDimensions(0);
	   cinfo_buf.releaseBuffer();	// remove any buffer allocated by fetch
	   cinfo.open(cinfo_buf);	// must rebind cinfo with cinfo_buf
	   reqs.clear();
	   continue;			// not a local data chunk
	}

	// this part is only added to deal with multiple processors
	if (reqs[0].getFileID() % numnodes != (u_int) mynode) {
	   // not my cluster
	   eod = true;
	   mbr.setNumberDimensions(0);
	   cinfo_buf.releaseBuffer();	// remove any buffer allocated by fetch
	   cinfo.open(cinfo_buf);	// must rebind cinfo with cinfo_buf
	   reqs.clear();
	   t2_say5(cout, "sbe:" << mynode
			<< ": NOT my local cluster, so skip it." << endl);
	   continue;
	}

	// so this is a local data cluster request, needs to add to req list
	t2_say5(cout, "sbe:" << mynode
			<< ": this is my local cluster, add to local req list"
			<< endl);
	u_int cnt =
		add_disk_requests_to_list(reqs, dfmap, clusterreq, blockreq,
					  mbr, cinfo_buf);
	// make sure that the buffer in info has been "stolen" and put into
	// the cluster list by add_disk_requests_to_list()
	t2_Assert (cinfo_buf.getBufferSize() == 0);
	t2_Assert (cinfo_buf.getDataSize() == 0);

	if (cnt != reqs.getNumberElements()) {
	   t2_err_ret(__FILE__, __LINE__,
		      "Only %d out of %d block requests are on local disks\nCurrent implementation assumes that all block requests of a cluster returned by\nan index must reside on the same processor",
		      cnt, reqs.getNumberElements());

	   delete index;
	   closeFiles(0, nfiles, fd);
	   return false;
	}

	// get ready to call fetch again
	eod = true;
	mbr.setNumberDimensions(0);
	cinfo_buf.releaseBuffer();	// remove any buffer allocated by fetch
	cinfo.open(cinfo_buf);		// must rebind cinfo with cinfo_buf
	reqs.clear();

     }		// while ((rv = index->fetch()) == T2_UDFRet_OK)

     t2_say2(cout, "sbe:" << mynode
		   << ": index.fetch sets eod=true when called the " << wc+1
		   << " times" << endl);
     t2_say2(cout, "sbe:" << mynode
		   << ": total # local cluster reqs = "
		   << clusterreq.getNumberClusterRequests() << endl);

     // make each cluster request in clusterreq point to dsenv, and init
     // its reference count
     for (j=0; j<clusterreq.getNumberClusterRequests(); j++) {
	T2_ClusterRequest& cls = clusterreq[j];
	cls.setDatasetEnv(dsenv.getAddress());
	cls.initRefCount();
     }

     // release memory
     delete index;
     closeFiles(0, nfiles, fd);

     if (rv != T2_UDFRet_OK) {			// some thing bad happened
	return false;
     }

  }		// for (u_int i=0; i<nds; i++) // step thru each dataset

  return true;
}



/////////////////////////////////////////////////////////////////////
//
// adding local disk requests into the appropriate lists,
// and returning the number of block requests added to list blockreq
//
/////////////////////////////////////////////////////////////////////
u_int
add_disk_requests_to_list(T2_VArray<T2_BlockRequest>& reqs,
			  T2_DSFileMap& dfmap,
			  T2_ClusterRequestList& clusterreq,
			  T2_PhyBlockRequestList& blockreq,
			  const T2_Box& mbr, T2_UsrArg& cinfo_buf)
{
  const u_int nn = reqs.getNumberElements();

  // insert a new cluster request into clusterreq
  T2_ClusterRequest& cluster = clusterreq.insert();
  cluster.setNumberBlockRequests(nn);
  cluster.setMBR(mbr);
  cluster.getMetaData().steal(cinfo_buf);

  u_int cnt = 0;
  for (u_int jj=0; jj<nn; jj++) {
      // get file descriptor and disk number
      int fdesc;
      u_int diskno;
      T2_BlockRequest& chk = reqs[jj];
      if (!dfmap.getFileInfo(chk.getFileID(), fdesc, diskno))	// not local !!
	 continue;

      // insert a new block request into blockreq
      T2_PhyBlockRequest& blockr = blockreq.touch(fdesc, chk.getOffset(),
					          chk.getNumberBytes());

      // now set the cross pointers between cluster request and block request
      // first make cluster request point to the new block request
      cluster.setBlockRequest(jj, blockr.getAddress());
      // now have the block request point back to the cluster request
      blockr.addStaticOwner(cluster.getAddress());

      cnt++;
  }

  return cnt;
}



/////////////////////////////////////////////////////////////////////
//
// strip-mining
//
/////////////////////////////////////////////////////////////////////
bool
strip_mine(T2_QQPlan& qqplan, T2_QEnv& qenv, const size_t mem,
	   T2_System& system)
{
  // get the accumulator meta object
  t2_Assert (qenv.getAccMetaPtr() != NULL);
  T2_AccMetaObj& accmeta = *(qenv.getAccMetaPtr());

  const T2_QueryInfo& qinfo = qenv.getQueryInfo();
  T2_VArray<T2_Region>& qwindows = qenv.getQueryWindows();
  t2_Assert (qenv.getNumberQueryWindows() == 0);
  if (accmeta.stripMine(qinfo, mem, T2_ClusterInfoList(qenv), qwindows)
     != T2_UDFRet_OK)
  {
     return false;
  }

  // partition the disk requests by iterations
  const u_int nn = qenv.getNumberDatasets();
  for (u_int i=0; i<nn; i++) {
     T2_DatasetEnv& dsenv = qenv.getDatasetEnv(i);
     qqplan.insertClusterRequests(dsenv.getClusterRequestList(), qwindows,
				  *(dsenv.getProjectFuncObjPtr()), system);
  }

  return true;
}



/////////////////////////////////////////////////////////////////////
//
// execute all query plans
//
/////////////////////////////////////////////////////////////////////
void
execute_all_qplans(T2_Array<T2_QPlan>& qplan, T2_Array<T2_QEnv>& qenv,
		  const char* ofname, const T2_Array<T2_RTS>& rts, bool verbose,
		  void* stuff)
{
  t2_Assert (qplan.getNumberElements() == qenv.getNumberElements());
  t2_Assert (qplan.getNumberElements() == rts.getNumberElements());

  u_int i;					// loop index
  const u_int nqry = qenv.getNumberElements();

  if (nqry == 0)
     return;

  t2_Assert (qplan[0].getNumberQQPlans() == 1);
  u_int nsubplans = qplan[0].getQQPlan(0).getNumberSubplans();
  for (i=0; i<nqry; i++) {
     t2_Assert(qplan[i].getNumberQQPlans() == 1);
     T2_QQPlan& qqplan = qplan[i].getQQPlan(0);
     t2_Assert(qqplan.getNumberSubplans() == qenv[i].getNumberQueryWindows());
     t2_Assert(qqplan.getNumberSubplans() == nsubplans);
  }

  if (verbose)
     cout << "==================================================" << endl
	  << "sbe: executing query plan for " << nsubplans << " iterations"
	  << endl;

  for (T2_Iteration splan = 0; splan < nsubplans; splan++) {
     if (verbose)
	cout << "==================================================" << endl
	     << "sbe: initializing accumulator for iteration " << splan
	     << endl;
     for (i=0; i<nqry; i++) {
	if (qenv[i].isValid() == true)
	   execute_qplan_init_acc(splan, qplan[i], qenv[i]);
     }

     if (verbose)
	cout << "sbe: entering local-reduction phase" << endl;
     for (i=0; i<nqry; i++) {
	if (qenv[i].isValid() == true)
	   execute_qplan_local_reduction(splan, qplan[i], rts[i].getSystem(),
				         qenv[i].getQueryWindow(splan));
     }

     if (verbose)
	cout << "sbe: entering global-combine phase" << endl;
     execute_all_qplans_global_combine(qplan, qenv, rts, verbose);

     if (verbose)
	cout << "sbe: generating the final output" << endl;
     for (i=0; i<nqry; i++) {
	if (qenv[i].isValid() == true)
	   execute_qplan_output_handling(splan, qenv[i], ofname,
				         rts[i].getSystem(), verbose, stuff);
     }

     for (i=0; i<nqry; i++) {
	if (qenv[i].isValid() == true)
	   execute_qplan_cleanup(qenv[i]);
     }
  }
}


/////////////////////////////////////////////////////////////////////
//
// initializing an accumulator for a query plan
//
/////////////////////////////////////////////////////////////////////
void
execute_qplan_init_acc(T2_Iteration iter, T2_QPlan& qplan, T2_QEnv& qenv)
{
  // allocate the accumulator first
  if (qenv.allocateAccumulator(iter) == false)
     qenv.invalidate();
}


/////////////////////////////////////////////////////////////////////
//
// carrying out local reduction phase for a query plan
//
/////////////////////////////////////////////////////////////////////
void
execute_qplan_local_reduction(T2_Iteration iter, T2_QPlan& qplan,
			      const T2_System& system, const T2_Region& tilereg)
{
  // can only handle a query plan for a single query
  t2_Assert (qplan.getNumberQQPlans() == 1);
  T2_QQPlan& qqplan = qplan.getQQPlan(0);

  // now step through the cluster requests
  T2_QSubplan& splan = qqplan.getSubplan(iter);
  const u_int nclusters = splan.getNumberClusterRequests();

  for (u_int jj=0; jj<nclusters; jj++) {
     T2_ClusterRequest* cluster = splan.getClusterRequestPtr(jj);
     t2_Assert (cluster != NULL);

     const u_int nblocks = cluster->getNumberBlockRequests();
     for (u_int b=0; b<nblocks; b++) {       // issue the block requests
	T2_PhyBlockRequest* blk = cluster->getBlockRequest(b);
	t2_Assert (blk != NULL);

	// allocate buffer space first
	t2_Assert (blk->hasNoBuffer());
	T2_PhyBlockRequestStatus s = blk->touch(cluster);

        t2_AssertOrDo(s == t2_bActivated || s == t2_bAvailable,
	      cout << "iteration " << iter << ", cluster " << jj
		   << " (" << (int) cluster->getAddress()
		   << "), block " << b << " touch returns unexpected status ="
		   << (int) s << endl);

        t2_AssertOrDo(!blk->hasNoBuffer(),
	      cout << "iteration " << iter << ", cluster " << jj
		   << " (" << (int) cluster->getAddress()
		   << "), block " << b << " allocBuffer failed, nbytes="
		   << blk->getNumberBytes() << ", s=" << (int) s
		   << endl);

	issue_read(*blk);	// issue a blocking disk read

	blk->becomePending();	// become pending
	blk->becomeAvailable();	// become available
	t2_Assert (blk->getRefCount() == 1);

     }				// for (u_int b=0; b<nblocks; b++)

     // process input data
//     process_input_data_chunk(*dsenv, data, tilereg, meta, cluster->getMBR(),
//			      system);
     process_input_data_chunk(cluster, tilereg);

     // release buffer space
     for (u_int bb=0; bb<nblocks; bb++) {
	cluster->getBlockRequest(bb)->release();
	t2_Assert (cluster->getBlockRequest(bb)->getRefCount() == 0);
	t2_Assert (cluster->getBlockRequest(bb)->getStatus() == t2_bCached ||
		   cluster->getBlockRequest(bb)->getStatus() == t2_bIdle);
     }

  }				// for (u_int jj=0; jj<nclusters; jj++)
}


/////////////////////////////////////////////////////////////////////
//
// carrying out global combine phase for all queries
//
/////////////////////////////////////////////////////////////////////
void
execute_all_qplans_global_combine(T2_Array<T2_QPlan>& qplan,
			 T2_Array<T2_QEnv>& qenv, const T2_Array<T2_RTS>& rts,
			 bool verbose)
{
  t2_Assert (qplan.getNumberElements() == qenv.getNumberElements());
  t2_Assert (qplan.getNumberElements() == rts.getNumberElements());

  u_int i;					// loop index
  const u_int nqry = qenv.getNumberElements();

  if (nqry == 0)
     return;

  const u_int numnodes = rts[0].getNumberNodes();
  for (i=0; i<nqry; i++) {
     t2_Assert (rts[i].getNumberNodes() == numnodes);
  }

//  if (numnodes == 1)				// no need for global combine
//     return;


  // check to see if global combine phase is needed; will invoke the method
  // needGlobalCombine() for every processor and check if the return values
  // are consistent
  if (verbose)
     cout << "sbe: about to invoke 'needGlobalCombine' to decide if global-combine phase"
	  << endl << "     is needed" << endl;
  bool need_global_combine;
  for (i=0; i<nqry; i++) {
     if (!qenv[i])
	continue;

     T2_AggregateFuncObj* aggr = qenv[i].getAggregateFuncObject();
     t2_Assert (aggr != NULL);
     T2_Accumulator *acc = qenv[i].getAccumulator();
     t2_Assert (acc != NULL);
     if (i == 0) {			// first query, set need_global_combine
        need_global_combine =
			aggr->needGlobalCombine(qenv[i].getQueryInfo(), *acc);
     }
     else {
        bool newval = aggr->needGlobalCombine(qenv[i].getQueryInfo(), *acc);
        if (need_global_combine != newval) {
	   // not consistent with earlier queries
	   cout << "T2 is currently invoking the method 'needGlobalCombine' from the"
		<< endl
		<< "aggregation function object of each simulated processor."
		<< endl
		<< "All the aggregation functions on the simulated processors should return the"
		<< endl
		<< "same boolean value."
		<< endl;
	   if (i == 1) {
	      cout << "However, processor 0 returns "
		   << T2_BoolObj(need_global_combine)
		   << "and processor 1 returns " << T2_BoolObj(newval)
		   << endl;
	   }
	   else {
	      cout << "However, processor 0-" << i-1 << " returns "
		   << T2_BoolObj(need_global_combine)
		   << "and processor " << i << " returns "
		   << T2_BoolObj(newval)
		   << endl;
	   }
	   t2_err_quit(__FILE__, __LINE__, "Program aborts");
	   return;
	}		/* if (need_global_combine != newval) */
     }			/* if (i == 0) else */
  }			/* for (i=0; i<nqry; i++) */

  if (!need_global_combine) {	// no need of global combine
     if (verbose)
        cout << "sbe: 'needGlobalCombine' says no need of a global-combine phase"
	     << endl;
     return;
  }


  // create the acc messages from each query
  if (verbose)
     cout << "sbe: yes, a global-combine phase is needed; about to call 'fillAccMsgBuffer'"
	  << endl
	  << "     to create the messages."
	  << endl;
  T2_Array<T2_AccMsgSendBuffer> *accmsg = 
				new T2_Array<T2_AccMsgSendBuffer>[numnodes];
  for (i=0; i<nqry; i++) {
     if (!qenv[i])
	continue;

     T2_AggregateFuncObj* aggr = qenv[i].getAggregateFuncObject();
     t2_Assert (aggr != NULL);
     T2_Accumulator *acc = qenv[i].getAccumulator();
     t2_Assert (acc != NULL);
     execute_qplan_global_combine_send(qplan[i], qenv[i], rts[i].getSystem(),
				       accmsg[i], verbose);
  }

  // process the acc messages created above
  if (verbose)
     cout << "sbe: about to call 'processAccMsg' for each received accumulator message"
	  << endl;

  for (i=0; i<nqry; i++) {
     if (!qenv[i])
	continue;

     T2_AggregateFuncObj* aggr = qenv[i].getAggregateFuncObject();
//     const T2_System& system = rts[i].getSystem();
     T2_Accumulator *acc = qenv[i].getAccumulator();
     for (T2_ProcID p=0; p<(int)numnodes; p++) {
	// accmsg[p][i] stores the msg that proc p sent to proc i
	T2_AccMsgRecvBuffer& msg = accmsg[p][i];

	// recall that a T2_AccMsgSendBuffer has a T2_UsrArg and a user data
	// pointer; need to find out if the msg data is in the T2_UsrArg or
	// in a buffer pointed to by the user data pointer
	const char* b;
	size_t msgsize;
	accmsg[p][i].getSendBufferInfo(b, msgsize);
	if (msgsize == 0)		// proc i recv no msg from proc p
	   continue;

	if (b == accmsg[p][i].getUserDataPtr()) {
	   // the msg is in a buffer pointed to by the user data pointer,
	   // so to make the interface works (processAccMsg must read from
	   // a T2_AccMsgRecvBuffer), simply need to copy it into a
	   // T2_AccMsgRecvBuffer object
	   T2_AccMsgRecvBuffer mbuf;
	   mbuf.allocBuffer(msgsize);
	   {
	      T2_UsrArgWriter w(mbuf);
	      w.write(b, msgsize);
	   }
	   T2_AccMsgBufferReader reader(mbuf);
	   aggr->processAccMsg(qenv[i].getQueryInfo(), p, reader, *acc);
	}
	else {
	   // the msg is in the T2_UsrArg buffer, so just reuse the buffer
	   T2_AccMsgBufferReader reader(msg);
	   aggr->processAccMsg(qenv[i].getQueryInfo(), p, reader, *acc);
	}		// if (b == accmsg[p][i].getUserDataPtr())
     }			// for (T2_ProcID p=0; p<(int)numnodes; p++)
  }			// for (i=0; i<nqry; i++)

  delete [] accmsg;
}


/////////////////////////////////////////////////////////////////////
//
// sending out acc messages for global combine phase for a query
//
/////////////////////////////////////////////////////////////////////
void
execute_qplan_global_combine_send(T2_QPlan& qplan, T2_QEnv& qenv,
			          const T2_System& system,
				  T2_Array<T2_AccMsgSendBuffer>& msgbuf,
				  bool verbose)
{
  T2_AggregateFuncObj* aggr = qenv.getAggregateFuncObject();
  t2_Assert (aggr != NULL);

  const u_int numnodes = system.getNumberNodes();
  T2_Accumulator *acc = qenv.getAccumulator();

  msgbuf.clear(numnodes);
  T2_Array<T2_AccMsgBufferWriter> writer(numnodes);

  for (u_int i=0; i<numnodes; i++) {
     writer[i].open(msgbuf[i]);

//     if (verbose)
//	cout << "sbe: proc " << system.getMyNode()
//	<< " estimates to send " << msgsize[i] << " bytes to proc " << i
//	<< endl;
  }

  if (aggr->fillAccMsgBuffer(qenv.getQueryInfo(),*acc,writer) != T2_UDFRet_OK) {
     t2_err_ret(__FILE__, __LINE__,
		"aggregation.fillAccMsgBuffer returns ERROR");
  }
}


/////////////////////////////////////////////////////////////////////
//
// handling the output for a query
//
/////////////////////////////////////////////////////////////////////
void
execute_qplan_output_handling(T2_Iteration iter, T2_QEnv& qenv,
			     const char* ofname, const T2_System& system,
			     bool verbose, void *stuff)
{
  char fname[256];
  const char *outfile = ofname;

  if (ofname != NULL) {
     outfile = fname;
     sprintf(fname, "%s-%d.%d", ofname, iter, system.getMyNode());
     if (verbose)
        cout << "sbe:" << system.getMyNode()
	     << ": generating output file " << outfile << endl;
  }

  handle_output(qenv, outfile, stuff);
}


/////////////////////////////////////////////////////////////////////
//
// cleaning up at the end of an iteration for a query plan
//
/////////////////////////////////////////////////////////////////////
void
execute_qplan_cleanup(T2_QEnv& qenv)
{
  qenv.deleteAccumulator();
}


/////////////////////////////////////////////////////////////////////
//
// execute a query plan
//
/////////////////////////////////////////////////////////////////////
//void
//execute_qplan(T2_QPlan& qplan, T2_QEnv& qenv, const char* ofname,
//	      const T2_System& system)
//{
//  t2_Assert (qplan.getNumberQQPlans() == 1);
//  T2_QQPlan& qqplan = qplan.getQQPlan(0);
//  t2_Assert (qqplan.getNumberSubplans() == qenv.getNumberQueryWindows());
//
//  const u_int niters = qqplan.getNumberSubplans();
//  for (u_int ii=0; ii<niters; ii++) {
//     // allocate the accumulator first
//     qenv.allocateAccumulator(ii);
//
//     // get the query window
//     const T2_Region& oqw = qenv.getQueryWindow(ii);
//
//     // now step through the cluster requests
//     T2_QSubplan& splan = qqplan.getSubplan(ii);
//     const u_int nclusters = splan.getNumberClusterRequests();
//
//
//cout << "there are " << nclusters << " clusters in qplan " << ii << endl;
//
//
//     for (u_int jj=0; jj<nclusters; jj++) {
//	T2_ClusterRequest* cluster = splan.getClusterRequestPtr(jj);
//	t2_Assert (cluster != NULL);
//	const u_int nblocks = cluster->getNumberBlockRequests();
//	for (u_int b=0; b<nblocks; b++) {	// issue the block requests
//	   T2_PhyBlockRequest* blk = cluster->getBlockRequest(b);
//	   t2_Assert (blk != NULL);
//
//	   // allocate buffer space first
//	   t2_Assert (blk->hasNoBuffer());
//	   blk->allocBuffer();
//
//	   // issue a disk read
//	   issue_read(*blk);
//	}			// for (u_int b=0; b<nblocks; b++)
//
//	// process input data
//	T2_DatasetEnv *dsenv = cluster->getDatasetEnv();
//	t2_Assert (dsenv != NULL);
//	T2_Cluster data(*cluster);
//	T2_UsrArgReader meta(cluster->getMetaData());
//	process_input_data_chunk(*dsenv, data, oqw, meta, cluster->getMBR());
//
//	// release buffer space
//	for (u_int bb=0; bb<nblocks; bb++)
//	   cluster->getBlockRequest(bb)->releaseBuffer();
//
//     }				// for (u_int jj=0; jj<nclusters; jj++) 
//
//
//     // do global combine phase here
//
//
//     // deal with the output here
//     {
//        char outfile[128];
//        sprintf(outfile, "%s.%d", ofname, ii);
//        handle_output(qenv, outfile, system);
//     }
//
//     // delete accumulator
//     qenv.deleteAccumulator();
//  }				// for (u_int ii=0; ii<niters; ii++)
//}
//

/////////////////////////////////////////////////////////////////////
//
// issue a disk read for a block request
//
/////////////////////////////////////////////////////////////////////
size_t
issue_read(T2_PhyBlockRequest& block)
{
  int fd = block.getFileDescriptor();
  lseek(fd, block.getOffset(), SEEK_SET);
  size_t n = read(block.getFileDescriptor(),
		  block.getBuffer().getBuffer(),
		  block.getNumberBytes());
  if (n != block.getNumberBytes()) {
     t2_err_ret(__FILE__, __LINE__,
	        "issue_read expects %d bytes, but read returns %d bytes",
		block.getNumberBytes(), n);
     block.getNumberBytes() = n;
  }
  return n;
}

/////////////////////////////////////////////////////////////////////
//
// process a retrieved input data chunk
//
/////////////////////////////////////////////////////////////////////
void
process_input_data_chunk(
//T2_DatasetEnv& dsenv, 
T2_ClusterRequest* cluster, const T2_Region& tilereg)
//const T2_Cluster& input,
//const T2_Region& oqw, 
//T2_ClusterAuxInfoReader& meta)
//const T2_Box& mbr, 
//const T2_System& system)
{
  t2_say5(cout, "sbe:" << system.getMyNode()
		<< ": process chunk mbr: " << mbr << endl);

  T2_DatasetEnv *dsenv = cluster->getDatasetEnv();
  t2_Assert (dsenv != NULL);

  T2_QEnv* qenv = dsenv->getQEnv();
  t2_Assert (qenv != NULL);

  T2_Cluster data(*cluster);
  T2_ClusterAuxInfoReader meta(cluster->getMetaData());

  T2_Dataset* dataset = dsenv->getDatasetPtr();
  t2_Assert (dataset != NULL);
  T2_Accumulator *accptr = qenv->getAccumulator();
  t2_Assert (accptr != NULL);
  T2_ProjectFuncObj *proj = dsenv->getProjectFuncObj();
  t2_Assert (proj != NULL);
  T2_AggregateFuncObj* aggr = qenv->getAggregateFuncObject();
  t2_Assert (aggr != NULL);
//  const T2_Box& iqr = dsenv->getQueryRegion();
//  const T2_DSID dsid = dsenv->getDatasetID();

  T2_UDFRet rv;
  rv = aggr->dafAcc(qenv->getQueryInfo(), data, cluster->getMBR(),
		    meta, *dataset, dsenv->getDatasetID(),
		    dsenv->getIteratorID(), dsenv->getQueryRegion(),
		    tilereg, *proj, qenv->getAccNavigatorID(), *accptr);
  if (rv != T2_UDFRet_OK) {
     t2_err_quit(__FILE__, __LINE__, "aggregation.dafAcc() returns ERROR");
     return;
  }
}


//void
//process_input_data_chunk(T2_DatasetEnv& dsenv, const T2_Cluster& input,
//			 const T2_Region& oqw, T2_ClusterAuxInfoReader& meta,
//			 const T2_Box& mbr, const T2_System& system)
//{
//  t2_say5(cout, "sbe:" << system.getMyNode()
//		<< ": process chunk mbr: " << mbr << endl);
//
//  T2_QEnv* qenv = dsenv.getQEnv();
//  t2_Assert (qenv != NULL);
//  T2_Dataset* dataset = dsenv.getDatasetPtr();
//  t2_Assert (dataset != NULL);
//  T2_ProjectFuncObj *proj = dsenv.getProjectFuncObj();
//  t2_Assert (proj != NULL);
//  T2_Accumulator *accptr = qenv->getAccumulator();
//  t2_Assert (accptr != NULL);
//  T2_Accumulator& acc = *accptr;
//  T2_AggregateFuncObj* aggr = qenv->getAggregateFuncObject();
//  t2_Assert (aggr != NULL);
//  const T2_Box& iqr = dsenv.getQueryRegion();
//  const T2_DSID dsid = dsenv.getDatasetID();
//
//  T2_Iterator* iter = NULL;
//  if (dataset->genIterator(system, dsenv.getIteratorID(), input,
//			   mbr, meta, iqr, oqw, *proj, iter) != T2_UDFRet_OK)
//  {
//     t2_err_quit(__FILE__, __LINE__, "dataset.genIterator() returns ERROR");
//     return;
//  }
//
//  if (iter == NULL) {
//     t2_err_ret(__FILE__, __LINE__, "dataset.genIterator() returns NULL");
//     return;
//  }
//  else {
//     T2_UDFRet iter_rv;
//     bool eod = true;
//     const void* input_elem = NULL;
//     T2_Point coord;
//     while ((iter_rv =iter->getNextElement(system,input,eod,input_elem,coord))
//	    == T2_UDFRet_OK)
//     {
//	if (eod == true)
//	   break;				// all done
//
//	if (input_elem == NULL) {
//	   t2_err_ret(__FILE__, __LINE__,
//		      "projection function returns NULL input element");
//	   delete iter;
//	   return;
//	}
//
//        T2_Region rg;
//	bool intersects = false;
//        if (proj->project(system,iqr,oqw,coord,intersects,rg) != T2_UDFRet_OK)
//	{
//	   t2_err_ret(__FILE__, __LINE__, "projection function returns ERROR");
//	   delete iter;
//	   return;
//	}
//	else if (intersects == true) {		// they actually intersect
//           T2_AccIterator *acciter = NULL;
//	   const T2_QueryInfo& qinfo = qenv->getQueryInfo();
//           if (acc.navigate(qinfo, qenv->getAccNavigatorID(), rg, acciter)
//		!= T2_UDFRet_OK)
//	   {
//		t2_err_ret(__FILE__, __LINE__,
//			  "accumulator navigate returns ERROR");
//		delete iter;
//		return;
//	   }
//
//	   if (acciter == NULL) {
//		t2_err_ret(__FILE__, __LINE__,
//		    "accumulator.navigate() returns NULL accumulator iterator");
//		delete iter;
//		return;
//	   }
//	   else {
//	      T2_UDFRet acciter_rv;
//	      bool acc_eod = true;
//	      void *acc_elem = NULL;
//              while ((acciter_rv = 
//		      acciter->getNextElement(qinfo, acc, acc_eod, acc_elem))
//		    == T2_UDFRet_OK)
//	      {
//		 if (acc_eod == true)
//		    break;			// all done
//
//		 if (acc_elem == NULL) {
//		    t2_err_ret(__FILE__, __LINE__,
//			     "accumulator-iterator.getNextElement returns NULL accumulator element");
//		    delete acciter;
//		    delete iter;
//		    return;
//		 }
//
//	         if (aggr->dafElem(qinfo, dsid, input_elem, acc_elem)
//		     != T2_UDFRet_OK)
//		 {
//		    t2_err_ret(__FILE__, __LINE__,
//				"aggr.dafElem returns ERROR");
//		    delete acciter;
//		    delete iter;
//		    return;
//		 }
//
//		 acc_eod = true;
//		 acc_elem = NULL;
//              }
//
//	      if (acciter_rv != T2_UDFRet_OK) {
//		 t2_err_ret(__FILE__, __LINE__,
//			  "accumulator-iterator.getNextElement returns ERROR");
//		 delete acciter;
//		 delete iter;
//		 return;
//	      }
//
//              delete acciter;
//	   }				// if (acciter == NULL) else
//        }				// else if (intersects == true)
//
//	eod = true;
//	input_elem = NULL;
//	coord.setNumberDimensions(0);
//     }					// while (iter->getNextElement() == OK)
//
//     if (iter_rv != T2_UDFRet_OK) {
//	t2_err_ret(__FILE__, __LINE__,
//		  "dataset-accumulator.getNextElement returns ERROR");
//     }
//
//     delete iter;
//  }					// if (iter == NULL) else
//}
//


/////////////////////////////////////////////////////////////////////
//
// handle the final output
//
/////////////////////////////////////////////////////////////////////
void
handle_output(T2_QEnv& qenv, const char* ofname, void* stuff)
{
  // create final output
  qenv.finalizeOutput();

  T2_Output *output = qenv.getOutput();
  t2_Assert (output != NULL);

  if (ofname == NULL) {			// no override for output handling
     handle_output_by_qenv(qenv, stuff);
  }
  else {				// forced to write output into files
     handle_output_write_file(qenv, ofname);
  }

  // remove final output
  qenv.deleteOutput();
}


void
handle_output_write_file(T2_QEnv& qenv, const char* ofname)
{
  t2_Assert (ofname != NULL);

  // note that finalizeOutput() has already been called
  T2_Output *output = qenv.getOutput();
  t2_Assert (output != NULL);

  // allocate output buffer
  size_t output_size = output->getOutputBufferSize(qenv.getQueryInfo());
  T2_OBuf obuf(output_size);
  T2_OutputBuffer outbuf(obuf);

  // get the set of data pointers to be written to a file
  T2_OutputDataPtrList ptrs;
  output->flushOutput(qenv.getQueryInfo(), outbuf, ptrs);

  // open a binary output file
  T2_OBFile obf(ofname);
  if (!obf) {
     t2_err_ret(__FILE__, __LINE__, "cannot open output file %s", ofname);
     return;
  }

  const u_int np = ptrs.getNumberPointers();
  for (u_int i=0; i<np; i++) {
     size_t size = ptrs.getDataSize(i);
     if (size == 0)
	continue;

     ssize_t s = obf.bwrite(ptrs.getDataPointer(i), 1, size);
     if (s < 0 || (u_int) s != size) {
	t2_err_ret(__FILE__, __LINE__,
		   "write %d bytes, but bwrite returns %d bytes", size, s);
     }
  }

  obf.close();
}


void
handle_output_by_qenv(T2_QEnv& qenv, void* stuff)
{
  // note that finalizeOutput() has already been called
  T2_Output *output = qenv.getOutput();
  t2_Assert (output != NULL);

  T2_OutputEnv& outenv = qenv.getOutputEnv();
  switch (outenv.getType()) {
     case t2_oUnknown:
		t2_err_ret(__FILE__, __LINE__,
			  "don't know what to do with output, do nothig");
		break;

     case t2_oBFile:
		t2_err_quit(__FILE__, __LINE__,
			  "this should have been taken care of earlier");
		break;

     case t2_oSocket:
		t2_Assert (stuff != NULL);
		handle_output_by_socket(qenv, *output, * ((T2_Socket *) stuff));
		break;

     default:
		t2_err_ret(__FILE__, __LINE__,
			  "unrecognized output handling type, do nothig");
  }

  qenv.deleteOutput();
}


void
handle_output_by_socket(T2_QEnv& qenv, T2_Output& output, T2_Socket& sock)
{
  if (!sock) {
     t2_err_ret(__FILE__, __LINE__, "socket to client is not connected");
     return;
  }

  // allocate output buffer
  size_t output_size = output.getOutputBufferSize(qenv.getQueryInfo());
  T2_OBuf obuf(output_size);
  T2_OutputBuffer outbuf(obuf);

  // get the set of data pointers to be written to a file
  T2_OutputDataPtrList ptrs;
  output.flushOutput(qenv.getQueryInfo(), outbuf, ptrs);

  const u_int np = ptrs.getNumberPointers();
  for (u_int i=0; i<np; i++) {
     size_t size = ptrs.getDataSize(i);
     if (size == 0)
        continue;

     ssize_t s = sock.bwrite(ptrs.getDataPointer(i), 1, size);
     if (s < 0 || (u_int) s != size) {
        t2_err_ret(__FILE__, __LINE__,
                   "write %d bytes to socket, but bwrite returns %d bytes",
		   size, s);
     }
  }
}
