
/* $Id: search-constructor-catalog.C,v 1.2 2000/01/17 04:37:22 chialin Exp $ */

/* search-constructor-catalog.C */


#include	<unistd.h>
#include	<sys/types.h>
#include	<regex.h>
#include	"t2_header.h"
#include	"t2_conscat.h"


//////////////////////////////////////////////////////////////////////////
//
// main program
//
//////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  if (argc < 4) {
     cout << "Usage: " << argv[0]
	  << " <cons catalog file> <flag = e/r> <pattern 1> [<pattern 2> ...]"
	  << endl
	  << "where" << endl
	  << "<flag> could be e (exact match) or r (regular expression)" << endl
	  << "<cons catalog file> is the output constructor catalog file" <<endl
	  << "<pattern 1> is a regular expression pattern used to search the catalog file"
	  << endl;
     return -1;
  }

  // try to open the constructor catalog file
  T2_ConsCatalogReader conscat(argv[1]);
  if (!conscat) {
      cout << "ERROR: cannot open '" << argv[1] << "'" << endl;
      return -1;
  }

  // check flag
  bool exact_match_flag;
  switch (argv[2][0]) {
     case 'E':
     case 'e':	// exact match
		exact_match_flag = true; break;
     case 'R':
     case 'r':	// regular expression
		exact_match_flag = false; break;
     default:
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: invalid flag '%c'", argv[2][0]);
		return -1;
  }

  // go through the given patterns and search through the catalog file
  for (int n=3; n<argc; n++) {
     T2_ConsCatalogEntry e;
     if (exact_match_flag == true) {
        cout << endl << "---- search for exact name= '" << argv[n] << "'" 
	     << endl;
	if (conscat.fetchExactMatch(e, argv[n]) == false)
	   cout << "no match" << endl;
	else
	   cout << e << endl;
     }
     else {
        if (conscat.fetchRegExpInit(argv[n]) == false) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: fetchRegExpInit fails for pattern '%s'", argv[n]);
	   continue;
        }
 
        cout << endl << "---- search against re pattern '" << argv[n] << "'" 
	     << endl;

        u_int cnt = 0;
        while (conscat.fetchRegExpName(e) == true) {
	   cout << "record " << ++cnt << endl
	        << e << endl;
        }
        cout << cnt << " records were found in total" << endl;
     }
  }

  conscat.close();
  return 0;
}

